!
! Module:boxchains
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: ******
! 
! Comments: TODO:
! 
! $Id: dss_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module boxchains

  use dal
  use errorHandling
  use dss

  use dss_aux		! for readRanges.
  use dss_ranges_aux
  use array_utils	! for smallestTrueIndex.
  use ssc_misc		! for orLists.

  implicit none 

  private 

  private calcBoxOverlap
  private dissectBox

  public constructBoxChain
  private copyBoxChain
  private calcBoxChainOverlap
  public dumpBoxChain
  public projectBoxChainToRangeChain
  public valueIsPassedByBoxChain

  public calcBoxChainOverlaps
  public calcNonOverlappingBoxChains


  type, public :: rangeBoxType
    type(RangeT) :: x, y
  end type rangeBoxType

  type, public :: boxChainType
    type(rangeBoxType), pointer :: boxes(:)
    integer(int32), pointer :: dssCompNums(:)
    character(20) :: xAxisName, yAxisName
  end type boxChainType


contains 

  !---------------------------------------------------------------------
  subroutine calcBoxOverlap(box1, box2, extraBoxes1&
  , extraBoxes2, thereisOverlap, box1HasBeenErased, box2HasBeenErased&
  , overlapBox)
  !---------------------------------------------------------------------
  ! The rangeBoxType comprises a RangeT variable in each of two directions,
  ! therefore defining a rectangular region. The present subroutine checks
  ! for overlap between the input pair of rectangles box1 and box2. The
  ! overlap region is clearly also of rangeBoxType since the overlap region
  ! between two rectangles is itself always a rectangle, but the remainders
  ! of the input rectangles are in general L-shaped and must be further
  ! subdivided to conform to the rangeBoxType. There is some freedom in the
  ! way this can be done. In the present subroutine the excess regions are
  ! divided first at values of the Y coordinate, as shown in the example
  ! below:
  !
  !	aaaaaaaaaaaaaaaaaaaaaaaaaa
  !	a                        a
  !	a                        a
  !	a              11111111111111111
  !	a              1         a     1
  !	a              1         a     1
  !	a              1         a     1
  !	a              1         a     1
  !	aaaaaaaaaaaaaaa1aaaaaaaaaa     1
  !	               1               1
  !	               1               1
  !	               1               1
  !	               11111111111111111
  !
  ! becomes
  !
  !	cccccccccccccccccccccccccc
  !	c                        c
  !	cccccccccccccccccccccccccc
  !	bbbbbbbbbbbbbbbAAAAAAAAAAA222222
  !	b             bA         A2    2
  !	b             bA         A2    2
  !	b             bA         A2    2
  !	b             bA         A2    2
  !	bbbbbbbbbbbbbbbAAAAAAAAAAA222222
  !	               33333333333333333
  !	               3               3
  !	               3               3
  !	               33333333333333333
  !
  ! Here 'b' and 'c' comprise the clipped return of input chain 'a', '2'
  ! and '3' comprise what remains of input '1', and 'A' represents the
  ! overlap.
  !
  ! NOTE! the returned pointers extraBoxes1 and extraBoxes2 should
  ! be deallocated after use.

    implicit none

    type(rangeBoxType), intent(inout) :: box1, box2
    type(rangeBoxType), pointer       :: extraBoxes1(:), extraBoxes2(:)
    logical(bool),      intent(out)   :: thereisOverlap,&
                                         box1HasBeenErased,&
                                         box2HasBeenErased
    type(rangeBoxType), intent(out)   :: overlapBox

    !-------------------------------------------------------------------
    ! Start

    box1HasBeenErased = .false. ! default
    box2HasBeenErased = .false. ! default

    ! Check first to see if there is any overlap at all.

!    if   (checkRangeOverlap(box1%x, box2%x) == BOVERLAPSA&
!    .and. checkRangeOverlap(box1%y, box2%y) == BOVERLAPSA) then
if (  checkRangeOverlap(box1%x, box2%x) .ne. BISTOOHIGH&
.and. checkRangeOverlap(box1%x, box2%x) .ne. BISTOOLOW&
.and. checkRangeOverlap(box1%y, box2%y) .ne. BISTOOHIGH&
.and. checkRangeOverlap(box1%y, box2%y) .ne. BISTOOLOW) then
      thereisOverlap = .true.
    else
      thereisOverlap = .false.
    end if

    if (thereisOverlap) then
!      overlapBox%x = andRangePair(box1%x, box2%x)
!      overlapBox%y = andRangePair(box1%y, box2%y)
call andRanges(box1%x, box2%x, overlapBox%x)
call andRanges(box1%y, box2%y, overlapBox%y)

      call dissectBox(overlapBox, box1, extraBoxes1, box1HasBeenErased)
      call dissectBox(overlapBox, box2, extraBoxes2, box2HasBeenErased)

    else
      allocate(extraBoxes1(0), extraBoxes2(0))
    end if

  end subroutine calcBoxOverlap

  !---------------------------------------------------------------------
  subroutine dissectBox(overlapBox, box, extraBoxes, boxHasBeenErased)
  !---------------------------------------------------------------------
  ! NOTE! the returned pointer extraBoxes should be deallocated after use.
  !
  ! This routine is designed to be called by calcBoxOverlap and is probably best left private. It IS NOT GUARANTEED TO WORK unless there is some overlap
  ! between the boxes (ie, overlapBox has sensible values). This should
  ! therefore be tested beforehand.

    implicit none

    type(rangeBoxType), intent(in)    :: overlapBox
    type(rangeBoxType), intent(inout) :: box
    type(rangeBoxType), pointer       :: extraBoxes(:)
    logical(bool),      intent(out)   :: boxHasBeenErased

    integer(int32) :: numXORedX, numXORedY, i
    type(RangeT), pointer :: xoredRangesX(:), xoredRangesY(:)

    !-------------------------------------------------------------------
    ! Start

    call xorRangePair(box%x, overlapBox%x, xoredRangesX)
    call xorRangePair(box%y, overlapBox%y, xoredRangesY)
    numXORedX = size(xoredRangesX)
    numXORedY = size(xoredRangesY)
    if (numXORedX < 1) then
      if (numXORedY < 1) then
        boxHasBeenErased = .true.
        allocate(extraBoxes(0))

      else ! numXORedY > 0
        box%y = copyRange(xoredRangesY(1))
        ! unnecessary to change box%x value.

        allocate(extraBoxes(numXORedY - 1))
        do i = 2, numXORedY
          extraBoxes(i - 1)%x = copyRange(box%x)
          extraBoxes(i - 1)%y = copyRange(xoredRangesY(i))
        end do
      end if

    else ! numXORedX > 0
      if (numXORedY < 1) then
        box%x = copyRange(xoredRangesX(1))
        ! unnecessary to change box%y value.

        allocate(extraBoxes(numXORedX - 1))
        do i = 2, numXORedX
          extraBoxes(i - 1)%x = copyRange(xoredRangesX(i))
          extraBoxes(i - 1)%y = copyRange(box%y)
        end do

      else ! numXORedY > 0
        box%y = copyRange(xoredRangesY(1))
        ! unnecessary to change box%x value.

        allocate(extraBoxes(numXORedX + numXORedY - 1))
        do i = 2, numXORedY
          extraBoxes(i - 1)%x = copyRange(box%x)
          extraBoxes(i - 1)%y = copyRange(xoredRangesY(i))
        end do
        do i = numXORedY + 1, numXORedY + numXORedX
          extraBoxes(i - 1)%x = copyRange(xoredRangesX(i - numXORedY))
          extraBoxes(i - 1)%y = copyRange(overlapBox%y)
        end do
      end if
    end if

    deallocate(xoredRangesX, xoredRangesY)

  end subroutine dissectBox

  !---------------------------------------------------------------------
  subroutine constructBoxChain(xRanges, yRanges, xAxisName, yAxisName&
  , dssCompNums, boxChain)
  !---------------------------------------------------------------------
  ! The type boxChainType consists of two vector pointers, one of type rangeBoxType and the other just an int32 pointer to store dss component numbers. It also contains two character variables that store the names of the x and y axes.
  !
  ! This subroutine takes 2 sets of ranges (supposedly orthogonal) and returns them as a single structure vector boxChain(). The two sets of ranges define a sort of checkerboard region in x-y space, ie made up of intersecting rectangles. If there are N xRanges and M yRanges then clearly there will be N*M intersection boxes. Each element of boxChain() contains the details of one of these boxes. The order of boxes in boxChain is not significant.
  !
  ! Oh yes, boxChain also points to another vector which contains the component numbers associated with the ranges. That is, both xRanges and yRanges should be specified by dss filters in all dss components listed in boxChain%dssCompNums().
  !
  ! NOTE that boxChain%boxes and %dssCompNums are pointers and so should be deallocated after use by the calling routine.

    implicit none

    type(RangeT),        intent(in)  :: xRanges(:), yRanges(:)
    integer(int32),      intent(in)  :: dssCompNums(:)
    type(boxChainType),  intent(out) :: boxChain
    character(*),        intent(in)  :: xAxisName, yAxisName

    integer(int16) :: xsize, ysize, numComps, boxNum, xi, yi
!, compNum
!, i

    !-------------------------------------------------------------------
    ! Start

    xsize = size(xRanges)
    ysize = size(yRanges)
    numComps = size(dssCompNums)
    allocate(boxChain%boxes(xsize * ysize), boxChain%dssCompNums(numComps))
    boxNum = 0
    do xi = 1, xsize
      do yi = 1, ysize
        boxNum = boxNum + 1
        boxChain%boxes(boxNum)%x = copyRange(xRanges(xi))
        boxChain%boxes(boxNum)%y = copyRange(yRanges(yi))
      end do
    end do

    boxChain%dssCompNums = dssCompNums

    boxChain%xAxisName = xAxisName
    boxChain%yAxisName = yAxisName

  end subroutine constructBoxChain

  !---------------------------------------------------------------------
  subroutine copyBoxChain(inChain, outChain)
  !---------------------------------------------------------------------
  ! Copies the pointer arrays in the input to the output and
  ! then deallocates the input.
  !
  ! NOTE that the pointers %boxes and %dssCompNums in outChain should be
  ! deallocated after use.

    implicit none

    type(boxChainType), intent(inout) :: inChain
    type(boxChainType), intent(out) :: outChain

    integer(int32) :: numBoxes, numComps

    !-------------------------------------------------------------------
    ! Start

    numBoxes = size(inChain%boxes)
    numComps = size(inChain%dssCompNums)
    allocate(outChain%boxes(numBoxes))
    allocate(outChain%dssCompNums(numComps))
    outChain%boxes = inChain%boxes
    outChain%dssCompNums = inChain%dssCompNums
    deallocate(inChain%boxes, inChain%dssCompNums)

!*** what about the axis names!

  end subroutine copyBoxChain

  !---------------------------------------------------------------------
  subroutine calcBoxChainOverlap(inChain1, inChain2, overlapChain)
  !---------------------------------------------------------------------
  ! NOTE! the returned pointers %boxes and %dssCompNums in inChain1, inChain2
  ! and overlapChain should be deallocated after use.
  !
  ! The input arguments are of type boxChainType and therefore contain two elements, 'boxes' and 'comps', both of them 1-D pointer arrays. The elements of 'boxes' describe rectangular regions which represent a combination of DSS range-type filters on two axes. The array 'comps' gives a set of DSS component numbers, each of which has a combination of filters on these two axes which is fully compatible with that stored in 'boxes' - that is, DSS components which have filters on the axes which, if ANDed with that in 'boxes', would return the identical filter as that in 'boxes'.
  !
  ! The purpose of the subroutine is to examine how the boxes in inChain1 overlap those of inChain2. Where there is any overlap, the chain of boxes describing the overlap regions is returned in the variable overlapChain. overlapChain%dssCompNums is in this case set to contain the union of inChain1%dssCompNums and inChain2%dssCompNums, since these overlap regions are consistent now with both sets of DSS components, under the definition given above. If there is no overlap, both overlapChain%boxes and overlapChain%dssCompNums are allocated to size = 0.
  !
  ! In case of some overlap occurring, the input chains are 'clipped' to exclude the overlap regions. This may entail some resizing of the 'boxes' element of either or both (in extreme cases, where one is entirely contained within the other for example, to 0). The way this is done is described in the subroutine calcBoxOverlap().

    implicit none

    type(boxChainType), intent(inout) :: inChain1, inChain2
    type(boxChainType), intent(out) :: overlapChain

    integer(int32) ::&
      numBoxes1, numBoxes2, guessAtMaxSize, maxSize1, maxSize2,&
      maxSizeOverlap, numClippedBoxes1, numClippedBoxes2,&
      numOverlapBoxes, boxNum1, boxNum2, box2loopLimit, numExtra,&
      start1, finis1, start2, finis2, i
    type(rangeBoxType), allocatable ::&
      tempClippedBoxes1(:), tempClippedBoxes2(:), tempOverlapBoxes(:)
    type(rangeBoxType), pointer ::&
      extraBoxFragments1(:), extraBoxFragments2(:)
    type(rangeBoxType) ::&
      overlapBox
    logical(bool) ::&
      thereWasOverlap
    logical(bool), allocatable ::&
      box1HasBeenErased(:), box2HasBeenErased(:), checkArray(:,:)

    !-------------------------------------------------------------------
    ! Start

    numBoxes1 = size(inChain1%boxes)
    numBoxes2 = size(inChain2%boxes)

    guessAtMaxSize = 5 * (numBoxes1 + numBoxes2)
    maxSize1 = guessAtMaxSize
    maxSize2 = guessAtMaxSize
    maxSizeOverlap = guessAtMaxSize
    allocate(tempClippedBoxes1(maxSize1), box1HasBeenErased(maxSize1))
    allocate(tempClippedBoxes2(maxSize2), box2HasBeenErased(maxSize2))
    allocate(tempOverlapBoxes(maxSizeOverlap))
    allocate(checkArray(maxSize1, maxSize2))

    tempClippedBoxes1(:numBoxes1) = inChain1%boxes
    tempClippedBoxes2(:numBoxes2) = inChain2%boxes
    checkArray = .false.
    numClippedBoxes1 = numBoxes1
    numClippedBoxes2 = numBoxes2
    numOverlapBoxes = 0
    box1HasBeenErased = .false.
    box2HasBeenErased = .false.

    boxNum1 = 1
    do while(boxNum1 <= numClippedBoxes1)
    if (box1HasBeenErased(boxNum1)) cycle

      box2loopLimit = numClippedBoxes2
      do boxNum2 = 1, box2loopLimit ! It is true that the number
        ! of tempClippedBoxes2 can increase during this loop, in
        ! exactly the same way as the number of tempClippedBoxes1
        ! can in the outer loop, so I could for the sake of
        ! consistency have made the inner loop exit the same type
        ! as that of the outer loop. However I have taken a
        ! slight shortcut because there is no point in testing
        ! any of the boxes2 added during the inner loop because
        ! they are all the fragments of boxes2 from which the
        ! overlap against the current box1 has already been clipped.

      if (box2HasBeenErased(boxNum2)) cycle

      if (checkArray(boxNum1, boxNum2)) cycle ! can't be any overlap with this
      ! combo, don't bother to check.

        call calcBoxOverlap(tempClippedBoxes1(boxNum1)&
          , tempClippedBoxes2(boxNum2), extraBoxFragments1&
          , extraBoxFragments2, thereWasOverlap, box1HasBeenErased(boxNum1)&
          , box2HasBeenErased(boxNum2), overlapBox)

        ! If there is clipping, the returned values of
        ! tempClippedBoxes(boxNum) are clipped as appropriate and
        ! any extra fragments returned in extraBoxFragments.

        checkArray(boxNum1, boxNum2) = .true.

        if (thereWasOverlap) then
          numOverlapBoxes = numOverlapBoxes + 1
!** test if bigger than maxSizeOverlap and increase if so

          tempOverlapBoxes(numOverlapBoxes) = overlapBox

          numExtra = size(extraBoxFragments1)
          if (numExtra > 0) then
            numClippedBoxes1 = numClippedBoxes1 + numExtra
!** test if bigger than maxSize1 and increase if so

            start1 = numClippedBoxes1 - numExtra + 1
            finis1 = numClippedBoxes1
            tempClippedBoxes1(start1:finis1) = extraBoxFragments1

            ! These new boxes1 are the children of the current box1.
            ! Copy the checks of the latter to the former:
            checkArray(start1:finis1, 1:numClippedBoxes2)&
              = spread(checkArray(boxNum1, 1:numClippedBoxes2), 1, numExtra)
          end if

          numExtra = size(extraBoxFragments2)
          if (numExtra > 0) then
            numClippedBoxes2 = numClippedBoxes2 + numExtra
!** test if bigger than maxSize2 and increase if so

            start2 = numClippedBoxes2 - numExtra + 1
            finis2 = numClippedBoxes2
            tempClippedBoxes2(start2:finis2) = extraBoxFragments2

            ! These new boxes2 are the children of the current box2.
            ! Copy the checks of the latter to the former:
            checkArray(1:numClippedBoxes1, start2:finis2)&
              = spread(checkArray(1:numClippedBoxes1, boxNum2), 2, numExtra)
          end if
        end if ! there was overlap.
      end do

      boxNum1 = boxNum1 + 1
    end do

    deallocate(extraBoxFragments1, extraBoxFragments2)
    deallocate(inChain1%boxes, inChain2%boxes)
!    allocate(inChain1%boxes(numClippedBoxes1))
!    allocate(inChain2%boxes(numClippedBoxes2))
allocate(inChain1%boxes(count(.not. box1HasBeenErased(:numClippedBoxes1))))
allocate(inChain2%boxes(count(.not. box2HasBeenErased(:numClippedBoxes2))))
    allocate(overlapChain%boxes(numOverlapBoxes))

!    inChain1%boxes = tempClippedBoxes1(:numClippedBoxes1)
!    inChain2%boxes = tempClippedBoxes2(:numClippedBoxes2)
i = 0
do boxNum1 = 1, numClippedBoxes1
if (box1HasBeenErased(boxNum1)) cycle
  i = i + 1
  inChain1%boxes(i) = tempClippedBoxes1(boxNum1)
end do
i = 0
do boxNum2 = 1, numClippedBoxes2
if (box2HasBeenErased(boxNum2)) cycle
  i = i + 1
  inChain2%boxes(i) = tempClippedBoxes2(boxNum2)
end do
    overlapChain%boxes = tempOverlapBoxes(:numOverlapBoxes)

    deallocate(tempClippedBoxes1, tempClippedBoxes2, tempOverlapBoxes)
    deallocate(box1HasBeenErased, box2HasBeenErased, checkArray)

    if (size(overlapChain%boxes) > 0) then ! some of the boxes overlapped.
      overlapChain%dssCompNums = orLists(inChain1%dssCompNums, inChain2%dssCompNums)
    end if

  end subroutine calcBoxChainOverlap

  !---------------------------------------------------------------------
  subroutine dumpBoxChain(boxChain)
  !---------------------------------------------------------------------

    implicit none

    type(boxChainType), intent(in) :: boxChain

    integer(int32) :: boxNum, compNum

    !-------------------------------------------------------------------
    ! Start

    do boxNum = 1, size(boxChain%boxes)
      print *, 'range box number =', boxNum

      print *, '  X axis name is ', trim(boxChain%xAxisName)
      select case(boxChain%boxes(boxNum)%x%lower%type)
      case(UNDEFINED)
        print *, '    Lower bound = UNDEFINED.'
      case(INCLUSIVE)
        print *, '    Lower bound = ', boxChain%boxes(boxNum)%x%lower%value&
          , ' (INCLUSIVE)'
      case(EXCLUSIVE)
        print *, '    Lower bound = ', boxChain%boxes(boxNum)%x%lower%value&
          , ' (EXCLUSIVE)'
      case default
        print *, '    Lower bound = ', boxChain%boxes(boxNum)%x%lower%value&
          , ' (type=???)'
      end select

      select case(boxChain%boxes(boxNum)%x%upper%type)
      case(UNDEFINED)
        print *, '    Upper bound = UNDEFINED.'
      case(INCLUSIVE)
        print *, '    Upper bound = ', boxChain%boxes(boxNum)%x%upper%value&
          , ' (INCLUSIVE)'
      case(EXCLUSIVE)
        print *, '    Upper bound = ', boxChain%boxes(boxNum)%x%upper%value&
          , ' (EXCLUSIVE)'
      case default
        print *, '    Upper bound = ', boxChain%boxes(boxNum)%x%upper%value&
          , ' (type=???)'
      end select

      print *, '  Y axis name is ', trim(boxChain%yAxisName)
      select case(boxChain%boxes(boxNum)%y%lower%type)
      case(UNDEFINED)
        print *, '    Lower bound = UNDEFINED.'
      case(INCLUSIVE)
        print *, '    Lower bound = ', boxChain%boxes(boxNum)%y%lower%value&
          , ' (INCLUSIVE)'
      case(EXCLUSIVE)
        print *, '    Lower bound = ', boxChain%boxes(boxNum)%y%lower%value&
          , ' (EXCLUSIVE)'
      case default
        print *, '    Lower bound = ', boxChain%boxes(boxNum)%y%lower%value&
          , ' (type=???)'
      end select

      select case(boxChain%boxes(boxNum)%y%upper%type)
      case(UNDEFINED)
        print *, '    Upper bound = UNDEFINED.'
      case(INCLUSIVE)
        print *, '    Upper bound = ', boxChain%boxes(boxNum)%y%upper%value&
          , ' (INCLUSIVE)'
      case(EXCLUSIVE)
        print *, '    Upper bound = ', boxChain%boxes(boxNum)%y%upper%value&
          , ' (EXCLUSIVE)'
      case default
        print *, '    Upper bound = ', boxChain%boxes(boxNum)%y%upper%value&
          , ' (type=???)'
      end select
    end do

    print *, 'Valid DSS component numbers are:'
    do compNum = 1, size(boxChain%dssCompNums)
      print *, '  ', boxChain%dssCompNums(compNum)
    end do

  end subroutine dumpBoxChain

  !---------------------------------------------------------------------
  function projectBoxChainToRangeChain(boxChain, projectToX) result(ranges)
  !---------------------------------------------------------------------
  ! Remember to deallocate ranges after use.

    implicit none

    type(boxChainType), intent(in) :: boxChain
    logical(bool),      intent(in) :: projectToX
    type(RangeT),       pointer    :: ranges(:)

    integer(int32) :: boxNum, numBoxes

    !-------------------------------------------------------------------
    ! Start

    numBoxes = size(boxChain%boxes)
    allocate(ranges(numBoxes))
    if (projectToX) then
      do boxNum = 1, numBoxes
        ranges(boxNum) = copyRange(boxChain%boxes(boxNum)%x)
      end do

    else ! projectToY:
      do boxNum = 1, numBoxes
        ranges(boxNum) = copyRange(boxChain%boxes(boxNum)%y)
      end do
    end if

    ! At this point, some of the ranges might overlap, and they are
    ! not necessarily in order. They have to be cleaned up.

    call correctRanges(ranges)

    ! Now they are ok.

  end function projectBoxChainToRangeChain

  !---------------------------------------------------------------------
  logical(bool) function valueIsPassedByBoxChain(xValue, yValue, boxChain)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: xValue, yValue
    type(boxChainType), intent(in) :: boxChain

    integer(int32) :: boxNum, numBoxes

    !-------------------------------------------------------------------
    ! Start

    valueIsPassedByBoxChain = .false. ! default

    numBoxes = size(boxChain%boxes)
    do boxNum = 1, numBoxes
      if   (valueIsWithinRange(xValue, boxChain%boxes(boxNum)%x)&
      .and. valueIsWithinRange(yValue, boxChain%boxes(boxNum)%y)) then
        valueIsPassedByBoxChain = .true.
  return
      end if
    end do

  end function valueIsPassedByBoxChain

  !---------------------------------------------------------------------
  subroutine calcBoxChainOverlaps(listOfBoxChains, newBoxChain)
  !---------------------------------------------------------------------
  ! The purpose of this subroutine is to examine how the boxes in the single chain 'newBoxChain' overlap those of all the box chains in 'listOfBoxChains'. The effect may be both to resize 'listOfBoxChains' and, where necessary, resize or alter its constituent chains, so as to (i) include new chains comprising the overlaps between all the input chains; (ii) the original chains but with all overlaps clipped or removed from their constituent boxes; (iii) a new chain comprising the chain input in 'newBoxChain', but with all overlap regions clipped from its boxes.
  !
  ! Note that the pointer arrays newBoxChain%boxes and newBoxChain%dssCompNums
  ! are deallocated in this routine.

    implicit none

    type(boxChainType), pointer :: listOfBoxChains(:)
    type(boxChainType), intent(inout) :: newBoxChain

    integer(int32) :: numInChains, numOutChains, inChainNum, outChainNum
    type(boxChainType) :: overlapChain
    type(boxChainType), pointer :: tempBoxChains(:)

    !-------------------------------------------------------------------
    ! Start

    numInChains = size(listOfBoxChains)
    numOutChains = 0

    do inChainNum = 1, numInChains
    if (size(newBoxChain%boxes) <= 0) exit ! no point in checking for
      ! more overlaps since newBoxChain is now entirely eaten away.

      call calcBoxChainOverlap(listOfBoxChains(inChainNum), newBoxChain&
        , overlapChain)

      if (size(listOfBoxChains(inChainNum)%boxes) > 0) then
        numOutChains = numOutChains + 1
        call copyBoxChain(listOfBoxChains(inChainNum)&
          , tempBoxChains(numOutChains))
      end if

      if (size(overlapChain%boxes) > 0) then ! there were some overlaps.
        numOutChains = numOutChains + 1
        call copyBoxChain(overlapChain, tempBoxChains(numOutChains))
      end if

      deallocate(listOfBoxChains(inChainNum)%boxes)
      deallocate(listOfBoxChains(inChainNum)%dssCompNums)
    end do

    if (size(newBoxChain%boxes) > 0) then
      numOutChains = numOutChains + 1
      call copyBoxChain(newBoxChain, tempBoxChains(numOutChains))
    end if

    allocate(listOfBoxChains(numOutChains))
    do outChainNum = 1, numOutChains
      call copyBoxChain(tempBoxChains(numOutChains)&
        , listOfBoxChains(numOutChains))
    end do

  end subroutine calcBoxChainOverlaps

  !---------------------------------------------------------------------
  subroutine calcNonOverlappingBoxChains(dssPointer, dssCompNums&
  , uniqueCombos, axisAName, axisBName, listOfBoxChains)
  !---------------------------------------------------------------------

!*** deallocate listOfBoxChains (and its components) after use.

    implicit none

    type(DssT),         intent(in) :: dssPointer
    integer(int32),     intent(in) :: dssCompNums(:)
    integer(int16),     intent(in) :: uniqueCombos(size(dssCompNums))
    character(*),       intent(in) :: axisAName, axisBName
    type(boxChainType), pointer    :: listOfBoxChains(:)

    integer(int16) :: minComboIndex, maxComboIndex, comboIndex, compIndex
    logical(bool), allocatable :: compMask(:)
    type(DScompT) :: dssComp
    type(RangeT), pointer :: rangesA(:), rangesB(:)
    integer(int32), allocatable :: dssCompNumsSubset(:)
    type(boxChainType) :: newBoxChain
    integer, parameter :: maxPermittedNumChains = 100

    !-------------------------------------------------------------------
    ! Start

    minComboIndex = minval(uniqueCombos)
    maxComboIndex = maxval(uniqueCombos)

    allocate(listOfBoxChains(minComboIndex), compMask(size(dssCompNums)))
    compMask = (uniqueCombos == minComboIndex)
    compIndex = smallestTrueIndex(compMask)
    ! smallestTrueIndex is in ssclib/array_utils.

    dssComp = dssComponent(dssPointer, dssCompNums(compIndex))
    call readRanges(dssComp, axisAName, rangesA)
    call readRanges(dssComp, axisBName, rangesB)

    allocate(dssCompNumsSubset(count(compMask)))
    dssCompNumsSubset = pack(dssCompNums, compMask)

    call constructBoxChain(rangesA, rangesB, axisAName, axisBName&
      , dssCompNumsSubset, listOfBoxChains(1))

    deallocate(dssCompNumsSubset)

    do comboIndex = minComboIndex + 1, maxComboIndex
    if (.not. any(uniqueCombos == comboIndex)) cycle

      compMask = (uniqueCombos == comboIndex)
      compIndex = smallestTrueIndex(compMask)
      ! smallestTrueIndex is in ssclib/array_utils.

      dssComp = dssComponent(dssPointer, dssCompNums(compIndex))
      ! The first dss component which has this filter combination.

      deallocate(rangesA, rangesB)
      call readRanges(dssComp, axisAName, rangesA)
      call readRanges(dssComp, axisBName, rangesB)

      allocate(dssCompNumsSubset(count(compMask)))
      dssCompNumsSubset = pack(dssCompNums, compMask)

      ! Load the component into newFiltChain:
      call constructBoxChain(rangesA, rangesB, axisAName, axisBName&
        , dssCompNumsSubset, newBoxChain)

      deallocate(dssCompNumsSubset)

      call calcBoxChainOverlaps(listOfBoxChains, newBoxChain)

if (size(listOfBoxChains) > maxPermittedNumChains) call error('bug_exp4','')

!*** test this loop with more complicated selections.
      deallocate(newBoxChain%boxes, newBoxChain%dssCompNums)
    end do
    deallocate(rangesA, rangesB)

  end subroutine calcNonOverlappingBoxChains

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module boxchains
