!
! Module: dal_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some wrapper routines for dealing with FITS files via the SAS DAL interface.
! 
! Comments: TODO:
!**** NOTE!! everything that accesses columns ought to be modified to check the dimensionality.
! 
! $Id: dal_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module dal_aux

  use dal
  use errorHandling

  use sort_mod	! used in sortTable, sortColumn.

  implicit none 

  private 

  public splitSetTabName
  public getRealKwd
  public getIntKwd
  public getStringKwd
  public setToNullArray_ims
  public isNullArray_ims
  public readArrayData
  public addOrOpenColumn
  public readColDataToFixed
  public readColDataToPtr
  public columnSetIsIdentical
  public minNonNullValue
  public maxNonNullValue
  public sortTable
  public sortColumn
  public getArraySize

  interface readArrayData
    module procedure readArrayDataName1dReal32, readArrayDataArray1dReal32
    module procedure readArrayDataName2dReal32, readArrayDataArray2dReal32
    module procedure readArrayDataName3dReal32, readArrayDataArray3dReal32
    module procedure readArrayDataName1dReal64, readArrayDataArray1dReal64
    module procedure readArrayDataName2dReal64, readArrayDataArray2dReal64
    module procedure readArrayDataName3dReal64, readArrayDataArray3dReal64
    module procedure readArrayDataName2dBool,   readArrayDataArray2dBool
    module procedure readArrayDataName2dInt8,  readArrayDataArray2dInt8
    module procedure readArrayDataName2dInt16,  readArrayDataArray2dInt16
    module procedure readArrayDataName2dInt32,  readArrayDataArray2dInt32
    module procedure readArrayDataSet1dReal32
    module procedure readArrayDataSet2dReal32
    module procedure readArrayDataSet3dReal32
    module procedure readArrayDataSet1dReal64
    module procedure readArrayDataSet2dReal64
    module procedure readArrayDataSet3dReal64
    module procedure readArrayDataSet2dBool
    module procedure readArrayDataSet2dInt8
    module procedure readArrayDataSet2dInt16
    module procedure readArrayDataSet2dInt32
  end interface

  interface readColDataToFixed
    module procedure readColToFixedCol2dReal32
    module procedure readColToFixedCol2dReal64
    module procedure readColToFixedNameReal32, readColToFixedColReal32
    module procedure readColToFixedNameReal64, readColToFixedColReal64
    module procedure readColToFixedNameInt8,   readColToFixedColInt8
    module procedure readColToFixedNameInt16,  readColToFixedColInt16
    module procedure readColToFixedNameInt32,  readColToFixedColInt32
    module procedure readColToFixedNameStr,    readColToFixedColStr
    module procedure readColToFixedNameBool,   readColToFixedColBool
  end interface
  interface readColDataToPtr
    module procedure readColToPtrName2dReal32
    module procedure readColToPtrName2dReal64
    module procedure readColToPtrNameReal32,   readColToPtrColReal32
    module procedure readColToPtrNameReal64,   readColToPtrColReal64
    module procedure readColToPtrNameInt8,     readColToPtrColInt8
    module procedure readColToPtrNameInt16,    readColToPtrColInt16
    module procedure readColToPtrNameInt32,    readColToPtrColInt32
  end interface

  interface minNonNullValue
    module procedure minNonNullValueSingle
    module procedure minNonNullValueDouble
  end interface

  interface maxNonNullValue
    module procedure maxNonNullValueSingle
    module procedure maxNonNullValueDouble
  end interface

  interface getArraySize
    module procedure getArraySizeSetName, getArraySizeSet
    module procedure getArraySizeArrayName, getArraySizeArrayNumber
    module procedure getArraySizeArray
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine splitSetTabName(setTabName, setName, tabName, noColonFound)
  !---------------------------------------------------------------------
  ! The sas param interface allows one to specify a dataset and table name
  ! in a single parameter, by using parameter type 'table'. The format of
  ! such a command-line argument is as in the followin example:
  !
  !	my_tab_parameter = fred.ds:BIGTAB
  !
  ! In practice it is easy inadvertently to leave off the colon + table
  ! name. In this case the dal does not offer any very helpful message.
  ! The aim of the present subroutine is (i) to allow people to leave off
  ! the tab name if they choose, in which case they get pointed to the
  ! first table in the dataset, and (ii) to issue a detailed and informatory
  ! warning in the case that it was necessary to specify the table name but
  ! the person forgot to do it.

    implicit none

    character(*),  intent(in)            :: setTabName
    character(*),  intent(out)           :: setName, tabName
    logical(bool), intent(out), optional :: noColonFound

    integer(int16) :: posOfColon
    type(DataSetT) :: set
    logical :: fileExists

    !-------------------------------------------------------------------
    ! Start

    posOfColon = index(setTabName, ':')
    if (posOfColon .le. 0) then ! no colon in the set-tab name. User
    ! has forgotten to include the table name.

      inquire(file=setTabName, exist=fileExists)

      if (.not. fileExists) then
call error('missingFile', "No colon+block name was found after the set &
&name. However, no dataset of that name exists, so I can't tell what block &
&of it you meant to access.")
      end if

      call warning('missingBlockName', "No colon+block name was &
        &found after the set name. I'm going to assume that you meant &
        &to access the first block in the set. If that's what you wanted, &
        &then there is no cause for alarm.")

      if (present(noColonFound)) noColonFound = .true.

      setName = setTabName
      set = dataSet(setName, READ)
      tabName = name(block(set, 0))
      call release(set)

    else
      if (present(noColonFound)) noColonFound = .false.

      setName = setTabName(:posOfColon - 1)
      tabName = setTabName(posOfColon + 1:)
    end if

  end subroutine splitSetTabName



!****** these below seem clunky - didn't I do it neater somewhere?????



  !---------------------------------------------------------------------
  subroutine getRealKwd(keyword, inDataSet, imageArray, value, status)
  !---------------------------------------------------------------------
  ! Gets FITS real-valued attribute from main dataset or primary array.
  !
  ! These routines were adapted from Clive Page's implot.

!** the real purpose of these is to cater for the silly situation in which we are in up to sas-6 in which eexpmap writes wcs kwds to the dataset header whereas evselect writes them to the array header.

    implicit none

    character*(*),  intent(in)            :: keyword
    type(DataSetT), intent(in)            :: inDataSet
    type(ArrayT),   intent(in)            :: imageArray
    real(double),   intent(out)           :: value
    integer,        intent(out), optional :: status

    integer :: localStatus
    logical :: fatalFail

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0
    value = 0.0d0 ! default

    if (present(status)) then
      fatalFail = .false.
    else
      fatalFail = .true.
    end if

    if(hasAttribute(inDataSet, keyword)) then
      if (fatalFail&
      .or. dataType(attribute(inDataSet, keyword)) .eq. REAL_ATTRIBUTE) then
        value = real64Attribute(inDataSet, keyword)
      else
        localStatus = -1
      end if
    else if (fatalFail .or. hasAttribute(imageArray, keyword)) then
      if (fatalFail&
      .or. dataType(attribute(imageArray, keyword)) .eq. REAL_ATTRIBUTE) then
        value = real64Attribute(imageArray, keyword)
      else
        localStatus = -2
      end if
    else
      localStatus = -3
    end if

    if (present(status)) status = localStatus

  end subroutine getRealKwd

  !---------------------------------------------------------------------
  subroutine getRealKwd_20031205(keyword, inDataSet, imageArray, value, status)
  !---------------------------------------------------------------------
  ! Gets FITS real32 or 64 attribute from main dataset or primary array.
  !
  ! These routines were adapted from Clive Page's implot.

    implicit none

    character*(*),  intent(in)            :: keyword
    type(DataSetT), intent(in)            :: inDataSet
    type(ArrayT),   intent(in)            :: imageArray
    real(double),   intent(out)           :: value
    integer,        intent(out), optional :: status

    integer :: localStatus
    logical :: fatalFail

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0
    value = 0.0d0 ! default

    if (present(status)) then
      fatalFail = .false.
    else
      fatalFail = .true.
    end if

    if(hasAttribute(inDataSet, keyword)) then
      if (dataType(attribute(inDataSet, keyword)) .eq. REAL32) then
        value = real32Attribute(inDataSet, keyword)
      else
        if (fatalFail&
        .or. dataType(attribute(inDataSet, keyword)) .eq. REAL64) then
          value = real64Attribute(inDataSet, keyword)
        else
          localStatus = -1
        end if
      end if
    else if (fatalFail .or. hasAttribute(imageArray, keyword)) then
      if (dataType(attribute(imageArray, keyword)) .eq. REAL32) then
        value = real32Attribute(imageArray, keyword)
      else
        if (fatalFail&
        .or. dataType(attribute(imageArray, keyword)) .eq. REAL64) then
          value = real64Attribute(imageArray, keyword)
        else
          localStatus = -2
        end if
      end if
    else
      localStatus = -3
    end if

    if (present(status)) status = localStatus

  end subroutine getRealKwd_20031205

  !---------------------------------------------------------------------
  subroutine getIntKwd(keyword, inDataSet, imageArray, value, status)
  !---------------------------------------------------------------------
  ! Gets FITS int8, 16 or 32 attribute from main dataset or primary array.
  !
  ! These routines were adapted from Clive Page's implot.

!** the real purpose of these is to cater for the silly situation in which we are in up to sas-6 in which eexpmap writes wcs kwds to the dataset header whereas evselect writes them to the array header.

    implicit none

    character*(*),  intent(in)            :: keyword
    integer(int32), intent(out)           :: value
    integer,        intent(out), optional :: status
    type(DataSetT), intent(in)            :: inDataSet
    type(ArrayT),   intent(in)            :: imageArray

    integer :: localStatus
    logical :: fatalFail

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0
    value = 0 ! default.

    if (present(status)) then
      fatalFail = .false.
    else
      fatalFail = .true.
    end if

    if(hasAttribute(inDataSet, keyword)) then
      if (fatalFail&
      .or. dataType(attribute(inDataSet, keyword)) .eq. INTEGER_ATTRIBUTE) then
        value = int32Attribute(inDataSet, keyword)
      else
        localStatus = -1
      end if
    else if (fatalFail .or. hasAttribute(imageArray, keyword)) then
      if (fatalFail&
      .or. dataType(attribute(imageArray, keyword)) .eq. INTEGER_ATTRIBUTE) then
        value = int32Attribute(imageArray, keyword)
      else
        localStatus = -2
      end if
    else
      localStatus = -3
    end if

    if (present(status)) status = localStatus

  end subroutine getIntKwd

  !---------------------------------------------------------------------
  subroutine getIntKwd_20031205(keyword, inDataSet, imageArray, value, status)
  !---------------------------------------------------------------------
  ! Gets FITS int8, 16 or 32 attribute from main dataset or primary array.
  !
  ! These routines were adapted from Clive Page's implot.

    implicit none

    character*(*),  intent(in)            :: keyword
    integer(int32), intent(out)           :: value
    integer,        intent(out), optional :: status
    type(DataSetT), intent(in)            :: inDataSet
    type(ArrayT),   intent(in)            :: imageArray

    integer :: localStatus
    logical :: fatalFail

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0
    value = 0 ! default.

    if (present(status)) then
      fatalFail = .false.
    else
      fatalFail = .true.
    end if

    if(hasAttribute(inDataSet, keyword)) then
      if (dataType(attribute(inDataSet, keyword)) .eq. INTEGER8) then
        value = int8Attribute(inDataSet, keyword)
      else if (dataType(attribute(inDataSet, keyword)) .eq. INTEGER16) then
        value = int16Attribute(inDataSet, keyword)
      else
        if (fatalFail&
        .or. dataType(attribute(inDataSet, keyword)) .eq. INTEGER32) then
          value = int32Attribute(inDataSet, keyword)
        else
          localStatus = -1
        end if
      end if
    else if (fatalFail .or. hasAttribute(imageArray, keyword)) then
      if (dataType(attribute(imageArray, keyword)) .eq. INTEGER8) then
        value = int8Attribute(imageArray, keyword)
      else if (dataType(attribute(imageArray, keyword)) .eq. INTEGER16) then
        value = int16Attribute(imageArray, keyword)
      else
        if (fatalFail&
        .or. dataType(attribute(imageArray, keyword)) .eq. INTEGER32) then
          value = int32Attribute(imageArray, keyword)
        else
          localStatus = -2
        end if
      end if
    else
      localStatus = -3
    end if

    if (present(status)) status = localStatus

  end subroutine getIntKwd_20031205

  !---------------------------------------------------------------------
  subroutine getStringKwd(keyword, inDataSet, imageArray, value, status)
  !---------------------------------------------------------------------
  ! Gets FITS string attribute from main dataset or primary array.
  !
  ! These routines were adapted from Clive Page's implot.

!****Here it is assumed that a keyword of any type can be accessed as string?? Should check this.

!** the real purpose of these is to cater for the silly situation in which we are in up to sas-6 in which eexpmap writes wcs kwds to the dataset header whereas evselect writes them to the array header.

    implicit none

    character*(*),  intent(in)            :: keyword
    character*(*),  intent(out)           :: value
    integer,        intent(out), optional :: status
    type(DataSetT), intent(in)            :: inDataSet
    type(ArrayT),   intent(in)            :: imageArray

    integer :: localStatus
    logical :: fatalFail

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0
    value = "??" ! default.

    if (present(status)) then
      fatalFail = .false.
    else
      fatalFail = .true.
    end if

    if(hasAttribute(inDataSet, keyword)) then
      value = stringAttribute(inDataSet, keyword)
    else if (fatalFail .or. hasAttribute(imageArray, keyword)) then
      value = stringAttribute(imageArray, keyword)
    else
      localStatus = -3
    end if

    if (present(status)) status = localStatus

  end subroutine getStringKwd

  !---------------------------------------------------------------------
  subroutine setToNullArray_ims(inArray, xi, yi)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT),   intent(in) :: inArray
    integer(int32), intent(in) :: xi, yi

    integer, pointer :: arrayDims(:)

    !-------------------------------------------------------------------
    ! Start

    arrayDims => dimensions(inArray)
    call setToNull(inArray, yi * arrayDims(0) + xi)

  end subroutine setToNullArray_ims

  !---------------------------------------------------------------------
  logical(bool) function isNullArray_ims(inArray, xi, yi)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT),   intent(in) :: inArray
    integer(int32), intent(in) :: xi, yi

    integer, pointer :: arrayDims(:)

    !-------------------------------------------------------------------
    ! Start

    arrayDims => dimensions(inArray)
    isNullArray_ims = isNull(inArray, yi * arrayDims(0) + xi)

  end function isNullArray_ims

  !---------------------------------------------------------------------
  subroutine readArrayDataName1dReal32(imageSetName, vector)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as
  ! 2- or 3-d, has no more than 1 dimension of size > 1
  !
  ! NOTE that you should deallocate vector() after use.

    implicit none

    character(*), intent(in) :: imageSetName
    real(single), pointer    :: vector(:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call readArrayDataArray1dReal32(inArray, vector)
    call release(set)

  end subroutine readArrayDataName1dReal32

  !---------------------------------------------------------------------
  subroutine readArrayDataName2dReal32(imageSetName, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    character(*), intent(in) :: imageSetName
    real(single), pointer    :: image(:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call readArrayDataArray2dReal32(inArray, image)
    call release(set)

  end subroutine readArrayDataName2dReal32

  !---------------------------------------------------------------------
  subroutine readArrayDataName3dReal32(imageSetName, cube)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! NOTE that you should deallocate cube() after use.

    implicit none

    character(*), intent(in) :: imageSetName
    real(single), pointer    :: cube(:,:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call readArrayDataArray3dReal32(inArray, cube)
    call release(set)

  end subroutine readArrayDataName3dReal32

  !---------------------------------------------------------------------
  subroutine readArrayDataName1dReal64(imageSetName, vector)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as
  ! 2- or 3-d, has no more than 1 dimension of size > 1
  !
  ! NOTE that you should deallocate vector() after use.

    implicit none

    character(*), intent(in) :: imageSetName
    real(double), pointer    :: vector(:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call readArrayDataArray1dReal64(inArray, vector)
    call release(set)

  end subroutine readArrayDataName1dReal64

  !---------------------------------------------------------------------
  subroutine readArrayDataName2dReal64(imageSetName, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    character(*), intent(in) :: imageSetName
    real(double), pointer    :: image(:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call readArrayDataArray2dReal64(inArray, image)
    call release(set)

  end subroutine readArrayDataName2dReal64

  !---------------------------------------------------------------------
  subroutine readArrayDataName3dReal64(imageSetName, cube)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! NOTE that you should deallocate cube() after use.

    implicit none

    character(*), intent(in) :: imageSetName
    real(double), pointer    :: cube(:,:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call readArrayDataArray3dReal64(inArray, cube)
    call release(set)

  end subroutine readArrayDataName3dReal64

  !---------------------------------------------------------------------
  subroutine readArrayDataName2dBool(imageSetName, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    character(*), intent(in) :: imageSetName
    logical(bool), pointer    :: image(:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call readArrayDataArray2dBool(inArray, image)
    call release(set)

  end subroutine readArrayDataName2dBool

  !---------------------------------------------------------------------
  subroutine readArrayDataName2dInt8(imageSetName, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    character(*),  intent(in) :: imageSetName
    integer(int8), pointer    :: image(:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call readArrayDataArray2dInt8(inArray, image)
    call release(set)

  end subroutine readArrayDataName2dInt8

  !---------------------------------------------------------------------
  subroutine readArrayDataName2dInt16(imageSetName, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    character(*),   intent(in) :: imageSetName
    integer(int16), pointer    :: image(:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call readArrayDataArray2dInt16(inArray, image)
    call release(set)

  end subroutine readArrayDataName2dInt16

  !---------------------------------------------------------------------
  subroutine readArrayDataName2dInt32(imageSetName, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    character(*),   intent(in) :: imageSetName
    integer(int32), pointer    :: image(:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call readArrayDataArray2dInt32(inArray, image)
    call release(set)

  end subroutine readArrayDataName2dInt32

  !---------------------------------------------------------------------
  subroutine readArrayDataSet1dReal32(imageSet, vector)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as
  ! 2- or 3-d, has no more than 1 dimension of size > 1
  !
  ! NOTE that you should deallocate vector() after use.

    implicit none

    type(DataSetT), intent(in) :: imageSet
    real(single),   pointer    :: vector(:)

    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    inArray = array(imageSet, 0, READ)
    call readArrayDataArray1dReal32(inArray, vector)

  end subroutine readArrayDataSet1dReal32

  !---------------------------------------------------------------------
  subroutine readArrayDataSet2dReal32(imageSet, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    type(DataSetT), intent(in) :: imageSet
    real(single),   pointer    :: image(:,:)

    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    inArray = array(imageSet, 0, READ)
    call readArrayDataArray2dReal32(inArray, image)

  end subroutine readArrayDataSet2dReal32

  !---------------------------------------------------------------------
  subroutine readArrayDataSet3dReal32(imageSet, cube)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! NOTE that you should deallocate cube() after use.

    implicit none

    type(DataSetT), intent(in) :: imageSet
    real(single),   pointer    :: cube(:,:,:)

    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    inArray = array(imageSet, 0, READ)
    call readArrayDataArray3dReal32(inArray, cube)

  end subroutine readArrayDataSet3dReal32

  !---------------------------------------------------------------------
  subroutine readArrayDataSet1dReal64(imageSet, vector)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as
  ! 2- or 3-d, has no more than 1 dimension of size > 1
  !
  ! NOTE that you should deallocate vector() after use.

    implicit none

    type(DataSetT), intent(in) :: imageSet
    real(double),   pointer    :: vector(:)

    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    inArray = array(imageSet, 0, READ)
    call readArrayDataArray1dReal64(inArray, vector)

  end subroutine readArrayDataSet1dReal64

  !---------------------------------------------------------------------
  subroutine readArrayDataSet2dReal64(imageSet, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    type(DataSetT), intent(in) :: imageSet
    real(double),   pointer    :: image(:,:)

    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    inArray = array(imageSet, 0, READ)
    call readArrayDataArray2dReal64(inArray, image)

  end subroutine readArrayDataSet2dReal64

  !---------------------------------------------------------------------
  subroutine readArrayDataSet3dReal64(imageSet, cube)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! NOTE that you should deallocate cube() after use.

    implicit none

    type(DataSetT), intent(in) :: imageSet
    real(double),   pointer    :: cube(:,:,:)

    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    inArray = array(imageSet, 0, READ)
    call readArrayDataArray3dReal64(inArray, cube)

  end subroutine readArrayDataSet3dReal64

  !---------------------------------------------------------------------
  subroutine readArrayDataSet2dBool(imageSet, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    type(DataSetT), intent(in) :: imageSet
    logical(bool),  pointer    :: image(:,:)

    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    inArray = array(imageSet, 0, READ)
    call readArrayDataArray2dBool(inArray, image)

  end subroutine readArrayDataSet2dBool

  !---------------------------------------------------------------------
  subroutine readArrayDataSet2dInt8(imageSet, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    type(DataSetT), intent(in) :: imageSet
    integer(int8),  pointer    :: image(:,:)

    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    inArray = array(imageSet, 0, READ)
    call readArrayDataArray2dInt8(inArray, image)

  end subroutine readArrayDataSet2dInt8

  !---------------------------------------------------------------------
  subroutine readArrayDataSet2dInt16(imageSet, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    type(DataSetT), intent(in) :: imageSet
    integer(int16), pointer    :: image(:,:)

    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    inArray = array(imageSet, 0, READ)
    call readArrayDataArray2dInt16(inArray, image)

  end subroutine readArrayDataSet2dInt16

  !---------------------------------------------------------------------
  subroutine readArrayDataSet2dInt32(imageSet, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    type(DataSetT), intent(in) :: imageSet
    integer(int32), pointer    :: image(:,:)

    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    inArray = array(imageSet, 0, READ)
    call readArrayDataArray2dInt32(inArray, image)

  end subroutine readArrayDataSet2dInt32

  !---------------------------------------------------------------------
  subroutine readArrayDataArray1dReal32(inArray, vector)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as
  ! 2- or 3-d, has no more than 1 dimension of size > 1
  !
  ! NOTE that you should deallocate vector() after use.

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(single), pointer    :: vector(:)

    integer(int32), pointer :: dimVector(:)
    real(single),   pointer ::     image(:,:)
    real(single),   pointer ::      cube(:,:,:)
    integer :: numDimensions, arrayShape(3), biggestDimension(1), vectorShape(1)

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)
    arrayShape = 0 ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      call readArrayDataArray2dReal32(inArray, image)
      arrayShape(:2) = shape(image)
if (count(arrayShape .gt. 1) .gt. 1) call error('badImageShape','')
      biggestDimension = maxloc(arrayShape)
      vectorShape(1) = size(image, biggestDimension(1))
      allocate(vector(vectorShape(1)))
      vector = reshape(image, vectorShape)
      deallocate(image)

    case(3)
      call readArrayDataArray3dReal32(inArray, cube)
      arrayShape(:3) = shape(cube)
if (count(arrayShape .gt. 1) .gt. 1) call error('badCubeShape','')
      biggestDimension = maxloc(arrayShape)
      vectorShape(1) = size(cube, biggestDimension(1))
      allocate(vector(vectorShape(1)))
      vector = reshape(cube, vectorShape)
      deallocate(cube)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine readArrayDataArray1dReal32

  !---------------------------------------------------------------------
  subroutine readArrayDataArray2dReal32(inArray, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(single), pointer    :: image(:,:)

    integer(int32), pointer :: dimVector(:)
    real(single),   pointer ::      cube(:,:,:)
    integer ::&
      numDimensions, arrayShape(3), biggestDimensions(2, 1), imageShape(2)
    logical :: mask(3)

    integer(int8),  pointer :: arrayDataInt8(  :,:)
    integer(int16), pointer :: arrayDataInt16( :,:)
    integer(int32), pointer :: arrayDataInt32( :,:)
    real(single),   pointer :: arrayDataReal32(:,:)
    real(double),   pointer :: arrayDataReal64(:,:)

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)
    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(image(dimVector(0), dimVector(1)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        image = real(arrayDataInt8, single)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        image = real(arrayDataInt16, single)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        image = real(arrayDataInt32, single)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        image = arrayDataReal32
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        image = real(arrayDataReal64, single)
      end if

    case(3)
      call readArrayDataArray3dReal32(inArray, cube)
      arrayShape(:3) = shape(cube)
if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
      biggestDimensions(1,:) = maxloc(arrayShape)
      mask(biggestDimensions(1,1)) = .false.
      biggestDimensions(2,:) = maxloc(arrayShape, mask)
      imageShape(1) = size(cube, biggestDimensions(1,1))
      imageShape(2) = size(cube, biggestDimensions(2,1))
      allocate(image(imageShape(1), imageShape(2)))
      image = reshape(cube, imageShape)
      deallocate(cube)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine readArrayDataArray2dReal32

  !---------------------------------------------------------------------
  subroutine readArrayDataArray3dReal32(inArray, cube)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! NOTE that you should deallocate cube() after use.

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(single), pointer    :: cube(:,:,:)

    integer(int32), pointer :: dimVector(:)
    integer :: numDimensions

    integer(int8),  pointer :: arrayDataInt8(  :,:,:)
    integer(int16), pointer :: arrayDataInt16( :,:,:)
    integer(int32), pointer :: arrayDataInt32( :,:,:)
    real(single),   pointer :: arrayDataReal32(:,:,:)
    real(double),   pointer :: arrayDataReal64(:,:,:)

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)

    select case(numDimensions)
    case(1, 2)
call error('arrayRankTooSmall', '')
    case(3)
      allocate(cube(dimVector(0), dimVector(1), dimVector(2)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array3Data(inArray)
        cube = real(arrayDataInt8, single)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array3Data(inArray)
        cube = real(arrayDataInt16, single)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array3Data(inArray)
        cube = real(arrayDataInt32, single)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array3Data(inArray)
        cube = arrayDataReal32
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array3Data(inArray)
        cube = real(arrayDataReal64, single)
      end if

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine readArrayDataArray3dReal32

  !---------------------------------------------------------------------
  subroutine readArrayDataArray1dReal64(inArray, vector)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as
  ! 2- or 3-d, has no more than 1 dimension of size > 1
  !
  ! NOTE that you should deallocate vector() after use.

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(double), pointer    :: vector(:)

    integer(int32), pointer :: dimVector(:)
    real(double),   pointer ::     image(:,:)
    real(double),   pointer ::      cube(:,:,:)
    integer :: numDimensions, arrayShape(3), biggestDimension(1), vectorShape(1)

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)
    arrayShape = 0 ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      call readArrayDataArray2dReal64(inArray, image)
      arrayShape(:2) = shape(image)
if (count(arrayShape .gt. 1) .gt. 1) call error('badImageShape','')
      biggestDimension = maxloc(arrayShape)
      vectorShape(1) = size(image, biggestDimension(1))
      allocate(vector(vectorShape(1)))
      vector = reshape(image, vectorShape)
      deallocate(image)

    case(3)
      call readArrayDataArray3dReal64(inArray, cube)
      arrayShape(:3) = shape(cube)
if (count(arrayShape .gt. 1) .gt. 1) call error('badCubeShape','')
      biggestDimension = maxloc(arrayShape)
      vectorShape(1) = size(cube, biggestDimension(1))
      allocate(vector(vectorShape(1)))
      vector = reshape(cube, vectorShape)
      deallocate(cube)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine readArrayDataArray1dReal64

  !---------------------------------------------------------------------
  subroutine readArrayDataArray2dReal64(inArray, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(double), dimension(:,:), pointer :: image

    integer(int32), pointer :: dimVector(:)
    real(double),   pointer ::      cube(:,:,:)
    integer ::&
      numDimensions, arrayShape(3), biggestDimensions(2, 1), imageShape(2)
    logical :: mask(3)

    integer(int8),  pointer :: arrayDataInt8(  :,:)
    integer(int16), pointer :: arrayDataInt16( :,:)
    integer(int32), pointer :: arrayDataInt32( :,:)
    real(single),   pointer :: arrayDataReal32(:,:)
    real(double),   pointer :: arrayDataReal64(:,:)

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)
    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(image(dimVector(0), dimVector(1)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        image = real(arrayDataInt8, double)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        image = real(arrayDataInt16, double)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        image = real(arrayDataInt32, double)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        image = real(arrayDataReal32, double)
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        image = arrayDataReal64
      end if

    case(3)
      call readArrayDataArray3dReal64(inArray, cube)
      arrayShape(:3) = shape(cube)
if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
      biggestDimensions(1,:) = maxloc(arrayShape)
      mask(biggestDimensions(1,1)) = .false.
      biggestDimensions(2,:) = maxloc(arrayShape, mask)
      imageShape(1) = size(cube, biggestDimensions(1,1))
      imageShape(2) = size(cube, biggestDimensions(2,1))
      allocate(image(imageShape(1), imageShape(2)))
      image = reshape(cube, imageShape)
      deallocate(cube)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine readArrayDataArray2dReal64

  !---------------------------------------------------------------------
  subroutine readArrayDataArray3dReal64(inArray, cube)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! NOTE that you should deallocate cube() after use.

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(double), dimension(:,:,:), pointer :: cube

    integer(int32), pointer :: dimVector(:)
    integer :: numDimensions

    integer(int8),  pointer :: arrayDataInt8(  :,:,:)
    integer(int16), pointer :: arrayDataInt16( :,:,:)
    integer(int32), pointer :: arrayDataInt32( :,:,:)
    real(single),   pointer :: arrayDataReal32(:,:,:)
    real(double),   pointer :: arrayDataReal64(:,:,:)

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)

    select case(numDimensions)
    case(1, 2)
call error('arrayRankTooSmall', '')
    case(3)
      allocate(cube(dimVector(0), dimVector(1), dimVector(2)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array3Data(inArray)
        cube = real(arrayDataInt8, double)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array3Data(inArray)
        cube = real(arrayDataInt16, double)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array3Data(inArray)
        cube = real(arrayDataInt32, double)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array3Data(inArray)
        cube = real(arrayDataReal32, double)
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array3Data(inArray)
        cube = arrayDataReal64
      end if

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine readArrayDataArray3dReal64

  !---------------------------------------------------------------------
  subroutine readArrayDataArray2dBool(inArray, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    type(ArrayT), intent(in) :: inArray
    logical(bool), dimension(:,:), pointer :: image

    integer(int32), pointer :: dimVector(:)
!    real(double),   pointer ::      cube(:,:,:)
    integer ::&
      numDimensions, arrayShape(3)!, biggestDimensions(2, 1), imageShape(2)
    logical :: mask(3)

    integer(int8),  pointer :: arrayDataInt8(  :,:)
    integer(int16), pointer :: arrayDataInt16( :,:)
    integer(int32), pointer :: arrayDataInt32( :,:)
    real(single),   pointer :: arrayDataReal32(:,:)
    real(double),   pointer :: arrayDataReal64(:,:)

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)
    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(image(dimVector(0), dimVector(1)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        image = (arrayDataInt8 > 0)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        image = (arrayDataInt16 > 0)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        image = (arrayDataInt32 > 0)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        image = (arrayDataReal32 > 0)
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        image = (arrayDataReal64 > 0)
      end if

    case(3)
!      call readArrayDataArray3dBool(inArray, cube)
!      arrayShape(:3) = shape(cube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(image(imageShape(1), imageShape(2)))
!      image = reshape(cube, imageShape)
!      deallocate(cube)
!******
call error('notYetSupported','')

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine readArrayDataArray2dBool

  !---------------------------------------------------------------------
  subroutine readArrayDataArray2dInt8(inArray, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    type(ArrayT), intent(in) :: inArray
    integer(int8), dimension(:,:), pointer :: image

    integer(int32), pointer :: dimVector(:)
!    real(double),   pointer ::      cube(:,:,:)
    integer ::&
      numDimensions, arrayShape(3)!, biggestDimensions(2, 1), imageShape(2)
    logical :: mask(3)

    integer(int8),  pointer :: arrayDataInt8(  :,:)
    integer(int16), pointer :: arrayDataInt16( :,:)
    integer(int32), pointer :: arrayDataInt32( :,:)
    real(single),   pointer :: arrayDataReal32(:,:)
    real(double),   pointer :: arrayDataReal64(:,:)

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)
    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(image(dimVector(0), dimVector(1)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        image = arrayDataInt8
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        image = int(arrayDataInt16, int8)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        image = int(arrayDataInt32, int8)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        image = nint(arrayDataReal32, int8)
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        image = nint(arrayDataReal64, int8)
      end if

    case(3)
!      call readArrayDataArray3dBool(inArray, cube)
!      arrayShape(:3) = shape(cube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(image(imageShape(1), imageShape(2)))
!      image = reshape(cube, imageShape)
!      deallocate(cube)
!******
call error('notYetSupported','')

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine readArrayDataArray2dInt8

  !---------------------------------------------------------------------
  subroutine readArrayDataArray2dInt16(inArray, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    type(ArrayT), intent(in) :: inArray
    integer(int16), dimension(:,:), pointer :: image

    integer(int32), pointer :: dimVector(:)
!    real(double),   pointer ::      cube(:,:,:)
    integer ::&
      numDimensions, arrayShape(3)!, biggestDimensions(2, 1), imageShape(2)
    logical :: mask(3)

    integer(int8),  pointer :: arrayDataInt8(  :,:)
    integer(int16), pointer :: arrayDataInt16( :,:)
    integer(int32), pointer :: arrayDataInt32( :,:)
    real(single),   pointer :: arrayDataReal32(:,:)
    real(double),   pointer :: arrayDataReal64(:,:)

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)
    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(image(dimVector(0), dimVector(1)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        image = int(arrayDataInt8, int16)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        image = arrayDataInt16
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        image = int(arrayDataInt32, int16)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        image = nint(arrayDataReal32, int16)
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        image = nint(arrayDataReal64, int16)
      end if

    case(3)
!      call readArrayDataArray3dBool(inArray, cube)
!      arrayShape(:3) = shape(cube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(image(imageShape(1), imageShape(2)))
!      image = reshape(cube, imageShape)
!      deallocate(cube)
!******
call error('notYetSupported','')

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine readArrayDataArray2dInt16

  !---------------------------------------------------------------------
  subroutine readArrayDataArray2dInt32(inArray, image)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from an array of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later.
  !
  ! The routine can also handle an array which, although specified as 3-d,
  ! has at least 1 dimension of size == 1. In this case the order of the
  ! dimensions is altered: the 1st dimension of returned image is the
  ! largest dimension of the cube; the 2nd is the second-largest.
  !
  ! NOTE that you should deallocate image() after use.

    implicit none

    type(ArrayT), intent(in) :: inArray
    integer(int32), dimension(:,:), pointer :: image

    integer(int32), pointer :: dimVector(:)
!    real(double),   pointer ::      cube(:,:,:)
    integer ::&
      numDimensions, arrayShape(3)!, biggestDimensions(2, 1), imageShape(2)
    logical :: mask(3)

    integer(int8),  pointer :: arrayDataInt8(  :,:)
    integer(int16), pointer :: arrayDataInt16( :,:)
    integer(int32), pointer :: arrayDataInt32( :,:)
    real(single),   pointer :: arrayDataReal32(:,:)
    real(double),   pointer :: arrayDataReal64(:,:)

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)
    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(image(dimVector(0), dimVector(1)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        image = int(arrayDataInt8, int32)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        image = int(arrayDataInt16, int32)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        image = arrayDataInt32
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        image = nint(arrayDataReal32, int32)
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        image = nint(arrayDataReal64, int32)
      end if

    case(3)
!      call readArrayDataArray3dBool(inArray, cube)
!      arrayShape(:3) = shape(cube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(image(imageShape(1), imageShape(2)))
!      image = reshape(cube, imageShape)
!      deallocate(cube)
!******
call error('notYetSupported','')

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine readArrayDataArray2dInt32

  !---------------------------------------------------------------------
  function addOrOpenColumn(tab, colName, inDataType, units, comment)&
  result(col)
  !---------------------------------------------------------------------

    implicit none

    type(TableT),   intent(in)           :: tab
    character(*),   intent(in)           :: colName
    integer(int32), intent(in), optional :: inDataType
    character(*),   intent(in), optional :: units, comment
    type(ColumnT) :: col

    integer(int32) :: localDataType
    character(255) :: localUnits, localComment

    !-------------------------------------------------------------------
    ! Start

    if (present(inDataType)) then
      localDataType = inDataType
    else
      localDataType = REAL32
    end if

    if (present(units)) then
      localUnits = units
    else
      localUnits = ''
    end if

    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if

    if (hasColumn(tab, colName)) then
      col = column(tab, colName)

      if (present(inDataType) .and. dataType(col) .ne. localDataType) then
        call release(col)
        call deleteColumn(tab, colName)
        if (.not. present(inDataType)) call warning('noDataType',"The &
          &specified column doesn't exist in the file, so I am adding it. &
          &You didn't specify a data type so I'm making it REAL32. Hope &
          &that's ok!")

        col = addColumn(tab, colName, localDataType, localUnits&
          , comment=localComment)
      end if

    else
      if (.not. present(inDataType)) call warning('noDataType',"The specified &
        &column doesn't exist in the file, so I am adding it. You didn't &
        &specify a data type so I'm making it REAL32. Hope that's ok!")

      col = addColumn(tab, colName, localDataType, localUnits&
        , comment=localComment)
    end if

  end function addOrOpenColumn

  !---------------------------------------------------------------------
  subroutine readColToPtrName2dReal32(tab, colName, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real32 vector colData.
  !
  ! NOTE that colData should be deallocated after you're finished with it.

    implicit none 

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: colName
    real(single), pointer    :: colData(:,:)

    type(ColumnT) :: col
    integer, pointer :: dimVector(:)

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    dimVector => dimensions(col)
    if (size(dimVector) > 1) then
call error('badColDimension','ssclib:')
    end if

    if (size(dimVector) == 0) then
      allocate(colData(0:0, 0:numberOfRows(col) - 1))
      call readColToFixedColReal32(col, colData(0,:))

    else
      allocate(colData(0:dimVector(0)-1, 0:numberOfRows(col) - 1))
      call readColToFixedCol2dReal32(col, colData)
    end if
    call release(col)

  end subroutine readColToPtrName2dReal32

  !---------------------------------------------------------------------
  subroutine readColToPtrName2dReal64(tab, colName, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real64 vector colData.
  !
  ! NOTE that colData should be deallocated after you're finished with it.

    implicit none 

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: colName
    real(double), pointer    :: colData(:,:)

    type(ColumnT) :: col
    integer, pointer :: dimVector(:)

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    dimVector => dimensions(col)
    if (size(dimVector) > 1) then
call error('badColDimension','ssclib:')
    end if

    if (size(dimVector) == 0) then
      allocate(colData(0:0, 0:numberOfRows(col) - 1))
      call readColToFixedColReal64(col, colData(0,:))

    else
      allocate(colData(0:dimVector(0)-1, 0:numberOfRows(col) - 1))
      call readColToFixedCol2dReal64(col, colData)
    end if

  end subroutine readColToPtrName2dReal64

  !---------------------------------------------------------------------
  subroutine readColToFixedNameReal32(tab, colName, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real32 vector colData.

    implicit none 

    type(TableT), intent(in)  :: tab
    character(*), intent(in)  :: colName
    real(single), intent(out) :: colData(:)

    type(ColumnT) :: col

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    call readColToFixedColReal32(col, colData)
    call release(col)

  end subroutine readColToFixedNameReal32

  !---------------------------------------------------------------------
  subroutine readColToFixedNameReal64(tab, colName, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real64 vector colData.

    implicit none 

    type(TableT), intent(in)  :: tab
    character(*), intent(in)  :: colName
    real(double), intent(out) :: colData(:)

    type(ColumnT) :: col

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    call readColToFixedColReal64(col, colData)
    call release(col)

  end subroutine readColToFixedNameReal64

  !---------------------------------------------------------------------
  subroutine readColToFixedNameInt8(tab, colName, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int8 vector colData.

    implicit none 

    type(TableT),  intent(in)  :: tab
    character(*),  intent(in)  :: colName
    integer(int8), intent(out) :: colData(:)

    type(ColumnT) :: col

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    call readColToFixedColInt8(col, colData)
    call release(col)

  end subroutine readColToFixedNameInt8

  !---------------------------------------------------------------------
  subroutine readColToFixedNameInt16(tab, colName, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int16 vector colData.

    implicit none 

    type(TableT),   intent(in)  :: tab
    character(*),   intent(in)  :: colName
    integer(int16), intent(out) :: colData(:)

    type(ColumnT) :: col

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    call readColToFixedColInt16(col, colData)
    call release(col)

  end subroutine readColToFixedNameInt16

  !---------------------------------------------------------------------
  subroutine readColToFixedNameInt32(tab, colName, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int32 vector colData.

    implicit none 

    type(TableT),   intent(in)  :: tab
    character(*),   intent(in)  :: colName
    integer(int32), intent(out) :: colData(:)

    type(ColumnT) :: col

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    call readColToFixedColInt32(col, colData)
    call release(col)

  end subroutine readColToFixedNameInt32

  !---------------------------------------------------------------------
  subroutine readColToFixedNameStr(tab, colName, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! type string.


!*** extend later to convert other data types to string???


    implicit none 

    type(TableT),  intent(in)  :: tab
    character(*),  intent(in)  :: colName
    character(*),  intent(out) :: colData(:)

    type(ColumnT) :: col

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    call readColToFixedColStr(col, colData)
    call release(col)

  end subroutine readColToFixedNameStr

  !---------------------------------------------------------------------
  subroutine readColToFixedNameBool(tab, colName, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32, real64 and bool without needing
  ! to know which it is. The data is returned in a logical vector colData.

    implicit none 

    type(TableT),  intent(in)  :: tab
    character(*),  intent(in)  :: colName
    logical(bool), intent(out) :: colData(:)

    type(ColumnT) :: col

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    call readColToFixedColBool(col, colData)
    call release(col)

  end subroutine readColToFixedNameBool

  !---------------------------------------------------------------------
  subroutine readColToPtrNameReal32(tab, colName, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real32 vector colData.
  !
  ! NOTE that colData should be deallocated after you're finished with it.

    implicit none 

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: colName
    real(single), pointer    :: colData(:)

    type(ColumnT) :: col

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    allocate(colData(0:numberOfRows(col) - 1))
    call readColToFixedColReal32(col, colData)
    call release(col)

  end subroutine readColToPtrNameReal32

  !---------------------------------------------------------------------
  subroutine readColToPtrNameReal64(tab, colName, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real64 vector colData.
  !
  ! NOTE that colData should be deallocated after you're finished with it.

    implicit none 

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: colName
    real(double), pointer    :: colData(:)

    type(ColumnT) :: col

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    allocate(colData(0:numberOfRows(col) - 1))
    call readColToFixedColReal64(col, colData)
    call release(col)

  end subroutine readColToPtrNameReal64

  !---------------------------------------------------------------------
  subroutine readColToPtrNameInt8(tab, colName, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int8 vector colData.
  !
  ! NOTE that colData should be deallocated after you're finished with it.

    implicit none 

    type(TableT),  intent(in) :: tab
    character(*),  intent(in) :: colName
    integer(int8), pointer    :: colData(:)

    type(ColumnT) :: col

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    allocate(colData(0:numberOfRows(col) - 1))
    call readColToFixedColInt8(col, colData)
    call release(col)

  end subroutine readColToPtrNameInt8

  !---------------------------------------------------------------------
  subroutine readColToPtrNameInt16(tab, colName, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int16 vector colData.
  !
  ! NOTE that colData should be deallocated after you're finished with it.

    implicit none 

    type(TableT),   intent(in) :: tab
    character(*),   intent(in) :: colName
    integer(int16), pointer    :: colData(:)

    type(ColumnT) :: col

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    allocate(colData(0:numberOfRows(col) - 1))
    call readColToFixedColInt16(col, colData)
    call release(col)

  end subroutine readColToPtrNameInt16

  !---------------------------------------------------------------------
  subroutine readColToPtrNameInt32(tab, colName, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int32 vector colData.
  !
  ! NOTE that colData should be deallocated after you're finished with it.

    implicit none 

    type(TableT),   intent(in) :: tab
    character(*),   intent(in) :: colName
    integer(int32), pointer    :: colData(:)

    type(ColumnT) :: col

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    allocate(colData(0:numberOfRows(col) - 1))
    call readColToFixedColInt32(col, colData)
    call release(col)

  end subroutine readColToPtrNameInt32

  !---------------------------------------------------------------------
  subroutine readColToPtrColReal32(col, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real32 vector colData.
  !
  ! NOTE that colData should be deallocated after you're finished with it.

    implicit none 

    type(ColumnT), intent(in) :: col
    real(single),  pointer    :: colData(:)

    !-------------------------------------------------------------------
    ! Start:

    allocate(colData(0:numberOfRows(col) - 1))
    call readColToFixedColReal32(col, colData)

  end subroutine readColToPtrColReal32

  !---------------------------------------------------------------------
  subroutine readColToPtrColReal64(col, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real64 vector colData.
  !
  ! NOTE that colData should be deallocated after you're finished with it.

    implicit none 

    type(ColumnT), intent(in) :: col
    real(double),  pointer    :: colData(:)

    !-------------------------------------------------------------------
    ! Start:

    allocate(colData(0:numberOfRows(col) - 1))
    call readColToFixedColReal64(col, colData)

  end subroutine readColToPtrColReal64

  !---------------------------------------------------------------------
  subroutine readColToPtrColInt8(col, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int8 vector colData.
  !
  ! NOTE that colData should be deallocated after you're finished with it.

    implicit none 

    type(ColumnT), intent(in) :: col
    integer(int8), pointer    :: colData(:)

    !-------------------------------------------------------------------
    ! Start:

    allocate(colData(0:numberOfRows(col) - 1))
    call readColToFixedColInt8(col, colData)

  end subroutine readColToPtrColInt8

  !---------------------------------------------------------------------
  subroutine readColToPtrColInt16(col, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int16 vector colData.
  !
  ! NOTE that colData should be deallocated after you're finished with it.

    implicit none 

    type(ColumnT),  intent(in) :: col
    integer(int16), pointer    :: colData(:)

    !-------------------------------------------------------------------
    ! Start:

    allocate(colData(0:numberOfRows(col) - 1))
    call readColToFixedColInt16(col, colData)

  end subroutine readColToPtrColInt16

  !---------------------------------------------------------------------
  subroutine readColToPtrColInt32(col, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int32 vector colData.
  !
  ! NOTE that colData should be deallocated after you're finished with it.

    implicit none 

    type(ColumnT),  intent(in) :: col
    integer(int32), pointer    :: colData(:)

    !-------------------------------------------------------------------
    ! Start:

    allocate(colData(0:numberOfRows(col) - 1))
    call readColToFixedColInt32(col, colData)

  end subroutine readColToPtrColInt32

  !---------------------------------------------------------------------
  subroutine readColToFixedCol2dReal32(col, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real32 vector colData.
  !
  ! The second index of the returned array represents the row number.

    implicit none 

    type(ColumnT), intent(in)  :: col
    real(single),  intent(out) :: colData(:,:)

    integer(int8),  pointer :: vectorColDataInt8(:,:)
    integer(int16), pointer :: vectorColDataInt16(:,:)
    integer(int32), pointer :: vectorColDataInt32(:,:)
    real(single),   pointer :: vectorColDataReal32(:,:)
    real(double),   pointer :: vectorColDataReal64(:,:)

    !-------------------------------------------------------------------
    ! Start:

    select case(dataType(col))
    case (INTEGER8)
      vectorColDataInt8 => int8VectorData(col)
      colData = real(vectorColDataInt8, single)
    case (INTEGER16)
      vectorColDataInt16 => int16VectorData(col)
      colData = real(vectorColDataInt16, single)
    case (INTEGER32)
      vectorColDataInt32 => int32VectorData(col)
      colData = real(vectorColDataInt32, single)
    case (REAL32)
      vectorColDataReal32 => real32VectorData(col)
      colData = vectorColDataReal32
    case default ! this gives a DAL error if data type is none of the above.
      vectorColDataReal64 => real64VectorData(col)
      colData = real(vectorColDataReal64, single)
    end select

  end subroutine readColToFixedCol2dReal32

  !---------------------------------------------------------------------
  subroutine readColToFixedCol2dReal64(col, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real64 vector colData.
  !
  ! The second index of the returned array represents the row number.

    implicit none 

    type(ColumnT), intent(in)  :: col
    real(double),  intent(out) :: colData(:,:)

    integer(int8),  pointer :: vectorColDataInt8(:,:)
    integer(int16), pointer :: vectorColDataInt16(:,:)
    integer(int32), pointer :: vectorColDataInt32(:,:)
    real(single),   pointer :: vectorColDataReal32(:,:)
    real(double),   pointer :: vectorColDataReal64(:,:)

    !-------------------------------------------------------------------
    ! Start:

    select case(dataType(col))
    case (INTEGER8)
      vectorColDataInt8 => int8VectorData(col)
      colData = real(vectorColDataInt8, double)
    case (INTEGER16)
      vectorColDataInt16 => int16VectorData(col)
      colData = real(vectorColDataInt16, double)
    case (INTEGER32)
      vectorColDataInt32 => int32VectorData(col)
      colData = real(vectorColDataInt32, double)
    case (REAL32)
      vectorColDataReal32 => real32VectorData(col)
      colData = real(vectorColDataReal32, double)
    case default ! this gives a DAL error if data type is none of the above.
      vectorColDataReal64 => real64VectorData(col)
      colData = vectorColDataReal64
    end select

  end subroutine readColToFixedCol2dReal64

  !---------------------------------------------------------------------
  subroutine readColToFixedColReal32(col, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real32 vector colData.

    implicit none 

    type(ColumnT), intent(in)  :: col
    real(single),  intent(out) :: colData(:)

    integer, pointer :: colDims(:)
    integer(int8),  pointer :: colDataInt8(:), colDataInt8Vector(:,:)
    integer(int16), pointer :: colDataInt16(:), colDataInt16Vector(:,:)
    integer(int32), pointer :: colDataInt32(:), colDataInt32Vector(:,:)
    real(single),   pointer :: colDataReal32(:), colDataReal32Vector(:,:)
    real(double),   pointer :: colDataReal64(:), colDataReal64Vector(:,:)

    !-------------------------------------------------------------------
    ! Start:

    colDims => dimensions(col)
    if (size(colDims) == 1) then
      if (colDims(0) == 1) then
        select case(dataType(col))
        case (INTEGER8)
          colDataInt8 => int8Data(col)
          colData = real(colDataInt8, single)
        case (INTEGER16)
          colDataInt16 => int16Data(col)
          colData = real(colDataInt16, single)
        case (INTEGER32)
          colDataInt32 => int32Data(col)
          colData = real(colDataInt32, single)
        case (REAL32)
          colDataReal32 => real32Data(col)
          colData = colDataReal32
        case default ! this gives a DAL error if data type is none of the above.
          colDataReal64 => real64Data(col)
          colData = real(colDataReal64, single)
        end select

      else if (colDims(0) > 1) then
        call warning('colIsVector',"ssclib:")

        select case(dataType(col))
        case (INTEGER8)
          colDataInt8Vector => int8VectorData(col)
          colData = real(colDataInt8Vector(0,:), single)
        case (INTEGER16)
          colDataInt16Vector => int16VectorData(col)
          colData = real(colDataInt16Vector(0,:), single)
        case (INTEGER32)
          colDataInt32Vector => int32VectorData(col)
          colData = real(colDataInt32Vector(0,:), single)
        case (REAL32)
          colDataReal32Vector => real32VectorData(col)
          colData = colDataReal32Vector(0,:)
        case default ! this gives a DAL error if data type is none of the above.
          colDataReal64Vector => real64VectorData(col)
          colData = real(colDataReal64Vector(0,:), single)
        end select
      end if
    else
call error('badColDimension',"ssclib:")
    end if

  end subroutine readColToFixedColReal32

  !---------------------------------------------------------------------
  subroutine readColToFixedColReal64(col, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real64 vector colData.

    implicit none 

    type(ColumnT), intent(in)  :: col
    real(double),  intent(out) :: colData(:)

    integer, pointer :: colDims(:)
    integer(int8),  pointer :: colDataInt8(:), colDataInt8Vector(:,:)
    integer(int16), pointer :: colDataInt16(:), colDataInt16Vector(:,:)
    integer(int32), pointer :: colDataInt32(:), colDataInt32Vector(:,:)
    real(single),   pointer :: colDataReal32(:), colDataReal32Vector(:,:)
    real(double),   pointer :: colDataReal64(:), colDataReal64Vector(:,:)

    !-------------------------------------------------------------------
    ! Start:

!print *, '  readColToFixedColReal64'
    colDims => dimensions(col)
!print *, '  size(colDims), colDims(0):', size(colDims), colDims(0)
    if (size(colDims) == 1) then
      if (colDims(0) == 1) then
!print *, '  111'
        select case(dataType(col))
        case (INTEGER8)
          colDataInt8 => int8Data(col)
          colData = real(colDataInt8, double)
        case (INTEGER16)
          colDataInt16 => int16Data(col)
          colData = real(colDataInt16, double)
        case (INTEGER32)
          colDataInt32 => int32Data(col)
          colData = real(colDataInt32, double)
        case (REAL32)
          colDataReal32 => real32Data(col)
          colData = real(colDataReal32, double)
        case default ! this gives a DAL error if data type is none of the above.
          colDataReal64 => real64Data(col)
          colData = colDataReal64
        end select

      else if (colDims(0) > 1) then
!print *, '  222'
        call warning('badColDimension',"ssclib:")

        select case(dataType(col))
        case (INTEGER8)
          colDataInt8Vector => int8VectorData(col)
          colData = real(colDataInt8Vector(0,:), double)
        case (INTEGER16)
          colDataInt16Vector => int16VectorData(col)
          colData = real(colDataInt16Vector(0,:), double)
        case (INTEGER32)
          colDataInt32Vector => int32VectorData(col)
          colData = real(colDataInt32Vector(0,:), double)
        case (REAL32)
          colDataReal32Vector => real32VectorData(col)
          colData = real(colDataReal32Vector(0,:), double)
        case default ! this gives a DAL error if data type is none of the above.
          colDataReal64Vector => real64VectorData(col)
          colData = colDataReal64Vector(0,:)
        end select
      end if
    else
call error('badColDimension',"ssclib:")
    end if

  end subroutine readColToFixedColReal64

  !---------------------------------------------------------------------
  subroutine readColToFixedColInt8(col, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int8 vector colData.

    implicit none 

    type(ColumnT), intent(in)  :: col
    integer(int8), intent(out) :: colData(:)

    integer(int8),  pointer :: colDataInt8(:)
    integer(int16), pointer :: colDataInt16(:)
    integer(int32), pointer :: colDataInt32(:)
    real(single),   pointer :: colDataReal32(:)
    real(double),   pointer :: colDataReal64(:)

    !-------------------------------------------------------------------
    ! Start:

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = colDataInt8
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int8)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = int(colDataInt32, int8)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int8)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int8)
    end select

  end subroutine readColToFixedColInt8

  !---------------------------------------------------------------------
  subroutine readColToFixedColInt16(col, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int16 vector colData.

    implicit none 

    type(ColumnT),  intent(in)  :: col
    integer(int16), intent(out) :: colData(:)

    integer(int8),  pointer :: colDataInt8(:)
    integer(int16), pointer :: colDataInt16(:)
    integer(int32), pointer :: colDataInt32(:)
    real(single),   pointer :: colDataReal32(:)
    real(double),   pointer :: colDataReal64(:)

    !-------------------------------------------------------------------
    ! Start:

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = int(colDataInt8, int16)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int16)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = colDataInt32
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int16)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int16)
    end select

  end subroutine readColToFixedColInt16

  !---------------------------------------------------------------------
  subroutine readColToFixedColInt32(col, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int32 vector colData.

    implicit none 

    type(ColumnT),  intent(in)  :: col
    integer(int32), intent(out) :: colData(:)

    integer(int8),  pointer :: colDataInt8(:)
    integer(int16), pointer :: colDataInt16(:)
    integer(int32), pointer :: colDataInt32(:)
    real(single),   pointer :: colDataReal32(:)
    real(double),   pointer :: colDataReal64(:)

    !-------------------------------------------------------------------
    ! Start:

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = int(colDataInt8, int32)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int32)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = colDataInt32
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int32)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int32)
    end select

  end subroutine readColToFixedColInt32

  !---------------------------------------------------------------------
  subroutine readColToFixedColStr(col, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! type string.


!*** extend later to convert other data types to string???


    implicit none 

    type(ColumnT), intent(in)  :: col
    character(*),  intent(out) :: colData(:)

    integer :: row

    !-------------------------------------------------------------------
    ! Start:

    colData = ''
    do row = 0, numberOfRows(col) - 1
      colData(row+1) = stringCell(col, row)
    end do

  end subroutine readColToFixedColStr

  !---------------------------------------------------------------------
  subroutine readColToFixedColBool(col, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32, real64 and bool without needing
  ! to know which it is. The data is returned in a logical vector colData.

    implicit none 

    type(ColumnT), intent(in)  :: col
    logical(bool), intent(out) :: colData(:)

    integer(int8),  pointer :: colDataInt8(:)
    integer(int16), pointer :: colDataInt16(:)
    integer(int32), pointer :: colDataInt32(:)
    real(single),   pointer :: colDataReal32(:)
    real(double),   pointer :: colDataReal64(:)
    logical(bool),  pointer :: colDataBool(:)

    !-------------------------------------------------------------------
    ! Start:

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      where(colDataInt8 > 0) colData = .true.
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      where(colDataInt16 > 0) colData = .true.
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      where(colDataInt32 > 0) colData = .true.
    case (REAL32)
      colDataReal32 => real32Data(col)
      where(colDataReal32 > 0.0) colData = .true.
    case (REAL64)
      colDataReal64 => real64Data(col)
      where(colDataReal64 > 0.0d0) colData = .true.
    case default ! this gives a DAL error if data type is none of the above.
      colDataBool => boolData(col)
      colData = colDataBool
    end select

  end subroutine readColToFixedColBool

  !---------------------------------------------------------------------
  logical(bool) function columnSetIsIdentical(tableA, tableB)
  !---------------------------------------------------------------------
  ! Note this does not compare the column VALUES, just their occurrence,
  ! dimensions and data type.
  !
  ! NOTE that this function allocates some column and dimensions() pointers.
  ! These should of course be cleared by releasing the tables or even
  ! the datasets at some stage.

!** This seems only to be used by filtersAreIdentical in dss_aux at present.


    implicit none

    type(TableT), intent(in) :: tableA, tableB

    integer :: numColsA, numColsB, colNum
    type(ColumnT) :: colA, colB
    character(255) :: colName
    integer, pointer :: dimsA(:), dimsB(:)

    !-------------------------------------------------------------------
    ! Start

    columnSetIsIdentical = .false. ! default

    numColsA = numberOfColumns(tableA)
    numColsB = numberOfColumns(tableB)
  if (numColsA .ne. numColsB) return

    do colNum = 1, numColsA
      colA = column(tableA, colNum - 1)
      colName = name(colA)
  if (.not. hasColumn(tableB, colName)) return
  if (dataType(colA) .ne. dataType(column(tableB, colName))) return
      colB = column(tableB, colName)
      dimsA => dimensions(colA)
      dimsB => dimensions(colB)
  if (size(dimsA) .ne. size(dimsB)) return
  if (any(dimsA .ne. dimsB)) return
    end do

    columnSetIsIdentical = .true.

  end function columnSetIsIdentical

  !---------------------------------------------------------------------
  subroutine minNonNullValueSingle(tab, colName, minValue, allRowsNull)
  !---------------------------------------------------------------------

    implicit none

    type(TableT),  intent(in)  :: tab
    character(*),  intent(in)  :: colName
    real(single),  intent(out) :: minValue
    logical(bool), intent(out) :: allRowsNull

    type(ColumnT) :: col
    integer(int32) :: firstNonNullRow, row
    real(single), pointer :: colData(:)

    !-------------------------------------------------------------------
    ! Start

    minValue = 0.0

    col = column(tab, colName)
    allRowsNull = .true. ! default
    firstNonNullRow = -1
    do row = 0, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        allRowsNull = .false.
        firstNonNullRow = row
    exit
      end if
    end do

    if (allRowsNull) then
      call release(col)
  return
    end if

    colData => real32Data(col)
    minValue = colData(firstNonNullRow)
    do row = firstNonNullRow + 1, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        minValue = min(minValue, colData(row))
      end if
    end do

    call release(col)

  end subroutine minNonNullValueSingle

  !---------------------------------------------------------------------
  subroutine minNonNullValueDouble(tab, colName, minValue, allRowsNull)
  !---------------------------------------------------------------------

    implicit none

    type(TableT),  intent(in)  :: tab
    character(*),  intent(in)  :: colName
    real(double),  intent(out) :: minValue
    logical(bool), intent(out) :: allRowsNull

    type(ColumnT) :: col
    integer(int32) :: firstNonNullRow, row
    real(double), pointer :: colData(:)

    !-------------------------------------------------------------------
    ! Start

    minValue = 0.0

    col = column(tab, colName)
    allRowsNull = .true. ! default
    firstNonNullRow = -1
    do row = 0, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        allRowsNull = .false.
        firstNonNullRow = row
    exit
      end if
    end do

    if (allRowsNull) then
      call release(col)
  return
    end if

    colData => real64Data(col)
    minValue = colData(firstNonNullRow)
    do row = firstNonNullRow + 1, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        minValue = min(minValue, colData(row))
      end if
    end do

    call release(col)

  end subroutine minNonNullValueDouble

  !---------------------------------------------------------------------
  subroutine maxNonNullValueSingle(tab, colName, maxValue, allRowsNull)
  !---------------------------------------------------------------------
!**** NOTE!! everything that accesses columns ought to be similarly modified to check the dimensionality.

    implicit none

    type(TableT),  intent(in)  :: tab
    character(*),  intent(in)  :: colName
    real(single),  intent(out) :: maxValue
    logical(bool), intent(out) :: allRowsNull

    type(ColumnT) :: col
    integer(int32) :: firstNonNullRow, row
    real(single), pointer :: colData(:), colData2d(:,:)
    integer, pointer :: colDims(:)

    !-------------------------------------------------------------------
    ! Start

    maxValue = 0.0

    col = column(tab, colName)
    allRowsNull = .true. ! default
    firstNonNullRow = -1

    colDims => dimensions(col)
    if (size(colDims) == 1) then
      do row = 0, numberOfRows(tab) - 1
        if (.not. isNull(col, row)) then
          allRowsNull = .false.
          firstNonNullRow = row
      exit
        end if
      end do

      if (allRowsNull) then
        call release(col)
  return
      end if

      colData => real32Data(col)
      maxValue = colData(firstNonNullRow)
      do row = firstNonNullRow + 1, numberOfRows(tab) - 1
        if (.not. isNull(col, row)) then
          maxValue = max(maxValue, colData(row))
        end if
      end do

    else if (size(colDims) == 2) then
      do row = 0, numberOfRows(tab) - 1
        if (.not. isNull(col, row, 0)) then
          allRowsNull = .false.
          firstNonNullRow = row
      exit
        end if
      end do

      if (allRowsNull) then
        call release(col)
  return
      end if

      colData2d => real32VectorData(col)
      maxValue = colData2d(0,firstNonNullRow)
      do row = firstNonNullRow + 1, numberOfRows(tab) - 1
        if (.not. isNull(col, row, 0)) then
          maxValue = max(maxValue, colData2d(0,row))
        end if
      end do

    else
call error('tooManyColDims',"")
    end if

    call release(col)

  end subroutine maxNonNullValueSingle

  !---------------------------------------------------------------------
  subroutine maxNonNullValueSingle_20031208(tab, colName, maxValue, allRowsNull)
  !---------------------------------------------------------------------

    implicit none

    type(TableT),  intent(in)  :: tab
    character(*),  intent(in)  :: colName
    real(single),  intent(out) :: maxValue
    logical(bool), intent(out) :: allRowsNull

    type(ColumnT) :: col
    integer(int32) :: firstNonNullRow, row
    real(single), pointer :: colData(:)

    !-------------------------------------------------------------------
    ! Start

    maxValue = 0.0

    col = column(tab, colName)
    allRowsNull = .true. ! default
    firstNonNullRow = -1
    do row = 0, numberOfRows(col) - 1
!*** why not numberOfRows(tab)?
      if (.not. isNull(col, row)) then
        allRowsNull = .false.
        firstNonNullRow = row
    exit
      end if
    end do

    if (allRowsNull) then
      call release(col)
  return
    end if

    colData => real32Data(col)
    maxValue = colData(firstNonNullRow)
    do row = firstNonNullRow + 1, numberOfRows(col) - 1
!*** why not numberOfRows(tab)?
      if (.not. isNull(col, row)) then
        maxValue = max(maxValue, colData(row))
      end if
    end do

    call release(col)

  end subroutine maxNonNullValueSingle_20031208

  !---------------------------------------------------------------------
  subroutine maxNonNullValueDouble(tab, colName, maxValue, allRowsNull)
  !---------------------------------------------------------------------

    implicit none

    type(TableT),  intent(in)  :: tab
    character(*),  intent(in)  :: colName
    real(double),  intent(out) :: maxValue
    logical(bool), intent(out) :: allRowsNull

    type(ColumnT) :: col
    integer(int32) :: firstNonNullRow, row
    real(double), pointer :: colData(:)

    !-------------------------------------------------------------------
    ! Start

    maxValue = 0.0

    col = column(tab, colName)
    allRowsNull = .true. ! default
    firstNonNullRow = -1
    do row = 0, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        allRowsNull = .false.
        firstNonNullRow = row
    exit
      end if
    end do

    if (allRowsNull) then
      call release(col)
  return
    end if

    colData => real64Data(col)
    maxValue = colData(firstNonNullRow)
    do row = firstNonNullRow + 1, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        maxValue = max(maxValue, colData(row))
      end if
    end do

    call release(col)

  end subroutine maxNonNullValueDouble

  !---------------------------------------------------------------------
  subroutine sortTable(tab, sortColName)
  !---------------------------------------------------------------------
!*** expand this later so can evaluate a column expression before sorting it???

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: sortColName

    integer(int32) :: numRows, i
    type(ColumnT) :: sortCol
    real(single), pointer :: sortColData(:)
    integer, allocatable :: index(:)

    !-------------------------------------------------------------------
    ! Start

    numRows = numberOfRows(tab)
    if (numRows <= 0) then
      call warning('noRowsInTable','')
  return
    end if

    sortCol = column(tab, sortColName)
    sortColData => real32Data(sortCol)
    allocate(index(numRows))
    call quickSort(sortColData, index)
    call deleteColumn(tab, sortColName)

    do i = 0, numberOfColumns(tab) - 1
      call sortColumn(tab, i, index)
    end do

  end subroutine sortTable

  !---------------------------------------------------------------------
  subroutine sortColumn(tab, colNum, index)
  !---------------------------------------------------------------------
!*** overload it so that (i) index becomes optional (sort if not present) and (ii) can alternately give col name.

    implicit none

    type(TableT), intent(in) :: tab
    integer(int32), intent(in) :: colNum, index(:)

    integer(int32) :: numRows, row, stringLength
    type(ColumnT) :: col
    integer(int8),    pointer :: int8Ptr(:)
    integer(int16),   pointer :: int16Ptr(:)
    integer(int32),   pointer :: int32Ptr(:)
    real(single),     pointer :: real32Ptr(:)
    real(double),     pointer :: real64Ptr(:)
    logical(bool),    pointer :: boolPtr(:)
    integer(int8),    allocatable :: tempInt8Vector(:)
    integer(int16),   allocatable :: tempInt16Vector(:)
    integer(int32),   allocatable :: tempInt32Vector(:)
    real(single),     allocatable :: tempReal32Vector(:)
    real(double),     allocatable :: tempReal64Vector(:)
    logical(bool),    allocatable :: tempBoolVector(:)
    integer(int16), parameter :: tempStringLength = 255
    character(tempStringLength), allocatable :: tempStringVector(:)

    !-------------------------------------------------------------------
    ! Start

    numRows = numberOfRows(tab)
    if (size(index) .ne. numRows)&
call error('noRowsInTable','')

    if (numRows <= 0) then
      call warning('','')
  return
    end if

    col = column(tab, colNum)
    select case(dataType(col))
    case(INTEGER8)
      int8Ptr => int8Data(col)
      allocate(tempInt8Vector(numRows))
      do row = 0, numRows - 1
        tempInt8Vector(row+1) = int8Ptr(index(row+1)-1)
      end do
      int8Ptr = tempInt8Vector

    case(INTEGER16)
      int16Ptr => int16Data(col)
      allocate(tempInt16Vector(numRows))
      do row = 0, numRows - 1
        tempInt16Vector(row+1) = int16Ptr(index(row+1)-1)
      end do
      int16Ptr = tempInt16Vector

    case(INTEGER32)
      int32Ptr => int32Data(col)
      allocate(tempInt32Vector(numRows))
      do row = 0, numRows - 1
        tempInt32Vector(row+1) = int32Ptr(index(row+1)-1)
      end do
      int32Ptr = tempInt32Vector

    case(REAL32)
      real32Ptr => real32Data(col)
      allocate(tempReal32Vector(numRows))
      do row = 0, numRows - 1
        tempReal32Vector(row+1) = real32Ptr(index(row+1)-1)
      end do
      real32Ptr = tempReal32Vector

    case(REAL64)
      real64Ptr => real64Data(col)
      allocate(tempReal64Vector(numRows))
      do row = 0, numRows - 1
        tempReal64Vector(row+1) = real64Ptr(index(row+1)-1)
      end do
      real64Ptr = tempReal64Vector

    case(BOOLEAN)
      boolPtr => boolData(col)
      allocate(tempBoolVector(numRows))
      do row = 0, numRows - 1
        tempBoolVector(row+1) = boolPtr(index(row+1)-1)
      end do
      boolPtr = tempBoolVector

    case(STRING)
      stringLength = len(stringCell(col, 0))
      if (stringLength > tempStringLength) then
        call warning('','')
        stringLength = tempStringLength
      end if

      allocate(tempStringVector(numRows))
      tempStringVector = ''
      do row = 0, numRows - 1
        tempStringVector(row+1) = stringCell(col, index(row+1)-1)
      end do
      do row = 0, numRows - 1
        call setStringCell(col, row&
          , tempStringVector(row+1)(:stringLength))
      end do

    case default
call error('','')
    end select

    call release(col)

  end subroutine sortColumn

  !---------------------------------------------------------------------
  subroutine getArraySizeSetName(setName, xSize, ySize)
  !---------------------------------------------------------------------
  ! Assumes the array of interest is the 0th in the dataset.

    implicit none

    character(*),   intent(in)  :: setName
    integer(int16), intent(out) :: xSize, ySize

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, 0, READ)
    call getArraySizeArray(inArray, xSize, ySize)
    call release(set)

  end subroutine getArraySizeSetName

  !---------------------------------------------------------------------
  subroutine getArraySizeSet(set, xSize, ySize)
  !---------------------------------------------------------------------
  ! Assumes the array of interest is the 0th in the dataset.

    implicit none

    type(DataSetT), intent(in)  :: set
    integer(int16), intent(out) :: xSize, ySize

    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    inArray = array(set, 0, READ)
    call getArraySizeArray(inArray, xSize, ySize)
    call release(inArray)

  end subroutine getArraySizeSet

  !---------------------------------------------------------------------
  subroutine getArraySizeArrayName(setName, arrayName, xSize, ySize)
  !---------------------------------------------------------------------

    implicit none

    character(*),   intent(in)  :: setName, arrayName
    integer(int16), intent(out) :: xSize, ySize

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, arrayName, READ)
    call getArraySizeArray(inArray, xSize, ySize)
    call release(set)

  end subroutine getArraySizeArrayName

  !---------------------------------------------------------------------
  subroutine getArraySizeArrayNumber(setName, arrayNumber, xSize, ySize)
  !---------------------------------------------------------------------

    implicit none

    character(*),   intent(in)  :: setName
    integer(int32), intent(in)  :: arrayNumber
    integer(int16), intent(out) :: xSize, ySize

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, arrayNumber, READ)
    call getArraySizeArray(inArray, xSize, ySize)
    call release(set)

  end subroutine getArraySizeArrayNumber

  !---------------------------------------------------------------------
  subroutine getArraySizeArray(inArray, xSize, ySize)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT),   intent(in)  :: inArray
    integer(int16), intent(out) :: xSize, ySize

    integer, pointer :: dims(:)

    !-------------------------------------------------------------------
    ! Start

    dims => dimensions(inArray)
if (size(dims) .ne. 2) call error('',"")

    xSize = dims(0)
    ySize = dims(1)

    ! No need to deallocate dims since it is associated with the pointer dimensions, which (presumably) is deallocated when the dataset is released.

  end subroutine getArraySizeArray

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module dal_aux
