!
! Module: detection_stats
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: *****.
! 
! Comments: TODO:
!*********** Look through ALL software for unnecessary pointer allocations!
!*********** Look through ALL software and declare all ptrs => null()!
! 
! $Id: detection_stats.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module detection_stats

  use types

  use constants
  use math_utils	! for incompleteGammaP, incompleteGammaQ, erfc
  use minimizations	! for riddersMethod

  implicit none 

  private 

  public eddington
  public inverseEddington
  public chi2Prob
  public poissonProb
  public integratedPoissonProb
  public minDetPoissonCounts
  private yScalar
  private yVector
  public integratedGaussProb
  public calcChi2HalfDegFree
  public sensyDegradnFromGaussSmooth

  interface poissonProb
    module procedure poissonProbSingle, poissonProbInt32
  end interface

  interface minDetPoissonCounts
    module procedure minDetPoissonCountsScalar, minDetPoissonCountsVector
  end interface

  integer, public :: numBands = 0 ! default value.
  real, allocatable, public :: globalSrcCountRatios(:), globalBkgCounts(:)
  real, public :: targetLikelihood = 0.0, targetProb = 0.0 ! default values.
  real, public :: globalBkgCount = 0.0 ! default value.

  real :: detSrcGlobal, avBkgGlobal ! Used in the Newton's Method solution
  integer :: minDetableCountsGlobal ! in inverseEddington().

contains 

  !---------------------------------------------------------------------
  subroutine eddington(avBkg, avSrc, minLikelihood, detFrac, eddingtonBias&
  , status)
  !---------------------------------------------------------------------
  ! In an image field comprising a known background <B> plus added sources
  ! there is a probability P(<B>), associated with the detection in a pixel
  ! of B counts, of B arising purely through Poisson sampling of the
  ! background. It is therefore common practice to calculate this
  ! probability across the image and accept or retain only those source
  ! detections for which P falls below a specified threshhold P_min. Hence
  ! <B> and P_min (actually supplied to the subroutine as a minimum
  ! likelihood L_min = -ln(P_min)) together define a minimum number of
  ! counts C_min; counts above this value shall be registered as a source
  ! detection in that pixel.
  !
  ! Suppose the flux in a given pixel derives from <B> plus a source flux
  ! which contributes <S> counts on average to the pixel. The probability
  ! of detecting C counts in the pixel will follow the Poisson distribution
  !
  !	        (<B>+<S>)**C exp(-<B>-<S>)
  !	P(C) = ---------------------------.
  !	                   C!
  !
  ! If <B>+<S> is not much greater than C_min, not only will the fraction
  ! of sources of average counts <S> which succeed in being detected be
  ! significantly depressed, but of those sources of this value which are
  ! detected, the average count rate calculated will be distorted to higher
  ! values. This last effect is related to Eddington bias.
  !
  ! The detection fraction f may be calculated from
  !
  !	                   __inf
  !	f(<B>,<S>,C_min) = >_      (<B>+<S>)**i exp(-<B>-<S>) / i!
  !	                   i=C_min
  !
  !	                 = integratedPoissonProb(<B>+<S>, C_min),
  !
  ! whereas the average detected count <S>_det is given by
  !
  !	           __inf
  !	           >_      i * (<B>+<S>)**i exp(-<B>-<S>) / i!
  !	           i=C_min
  !	<S>_det = --------------------------------------------- - <B>
  !	           __inf
  !	           >_      (<B>+<S>)**i exp(-<B>-<S>) / i!
  !	           i=C_min
  !
  !
  !	           (<B>+<S>) * integratedPoissonProb(<B>+<S>, C_min - 1)
  !	        = ------------------------------------------------ - <B>.
  !	                  integratedPoissonProb(<B>+<S>, C_min)
  !
  ! Also:
  !	           __inf     i * (<B>+<S>)**i
  !	           >_      -------------------
  !	           i=C_min         i!
  !	<S>_det = ------------------------------ - <B>
  !	             __inf    (<B>+<S>)**i
  !	             >_      --------------
  !	             i=C_min       i!
  !
  ! The subroutine returns both these numbers (actually eddingtonBias
  ! = <S>_det / <S> is returned instead of <S>_det).
  !
  ! Note: since a-priori knowledge of <S> is required it probably only
  ! makes sense to call this routine in conjunction with simulations of
  ! background + source data.
!*** NOTE! This hasn't been tested yet. I ran up these equations in about 5 min so no guarantees!
  !.....................................................................

    implicit none

    real(single), intent(in)            :: avBkg, avSrc, minLikelihood
    real(single), intent(out)           :: detFrac, eddingtonBias
    integer,      intent(out), optional :: status

    integer :: foreignStatus
    real(single) :: minDetableCountsGlobalReal, uncert

    !-------------------------------------------------------------------
    ! Start

    if (avBkg <= 0.0) then
      if (present(status)) status = -1
  return
    end if

    if (avSrc <= 0.0) then
      if (present(status)) status = -2
  return
    end if

    if (minLikelihood <= 0.0) then
      if (present(status)) status = -3
  return
    end if

    ! Calculate C_min:
    call minDetPoissonCounts(avBkg, minLikelihood&
      , minDetableCountsGlobalReal, uncert, status=foreignStatus)

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus - 4
  return
    end if

    minDetableCountsGlobal = ceiling(minDetableCountsGlobalReal)
    if (minDetableCountsGlobal <= 0) then
      if (present(status)) status = -4
  return
    end if

    detFrac = integratedPoissonProb(avBkg + avSrc, minDetableCountsGlobal)

    eddingtonBias = ((avBkg + avSrc) * integratedPoissonProb(avBkg + avSrc&
      , minDetableCountsGlobal - 1) / integratedPoissonProb(avBkg + avSrc&
      , minDetableCountsGlobal) - avBkg) / avSrc

    if (present(status)) status = 0

  end subroutine eddington

  !---------------------------------------------------------------------
  subroutine inverseEddington(avBkg, detSrc, detSrcUncert, minLikelihood&
  , avSrc, avSrcUncert, status)
  !---------------------------------------------------------------------
  ! In an image field comprising a known background <B> plus added sources
  ! there is a probability P(<B>), associated with the detection in a pixel
  ! of B counts, of B arising purely through Poisson sampling of the
  ! background. It is therefore common practice to calculate this
  ! probability across the image and accept or retain only those source
  ! detections for which P falls below a specified threshhold P_min. Hence
  ! <B> and P_min (actually supplied to the subroutine as a minimum
  ! likelihood L_min = -ln(P_min)) together define a minimum number of
  ! counts C_min; counts above this value shall be registered as a source
  ! detection in that pixel.
  !
  ! Suppose the flux in a given pixel derives from <B> plus a source flux
  ! which contributes <S> counts on average to the pixel. The probability
  ! of detecting C counts in the pixel will follow the Poisson distribution
  !
  !	        (<B>+<S>)**C exp(-<B>-<S>)
  !	P(C) = ---------------------------.
  !	                   C!
  !
  ! If <B>+<S> is not much greater than C_min, not only will the fraction
  ! of sources of average counts <S> which succeed in being detected be
  ! significantly depressed, but of those sources of this value which are
  ! detected, the average count rate calculated will be distorted to higher
  ! values. This last effect is related to Eddington bias.
  !
  ! The average detected count <S>_det is given by
  !
  !	           __inf
  !	           >_      i * (<B>+<S>)**i exp(-<B>-<S>) / i!
  !	           i=C_min
  !	<S>_det = --------------------------------------------- - <B>
  !	             __inf
  !	             >_      (<B>+<S>)**i exp(-<B>-<S>) / i!
  !	             i=C_min
  !
  !
  !	           __inf     i * (<B>+<S>)**i
  !	           >_      -------------------
  !	           i=C_min         i!
  !	        = ------------------------------ - <B>
  !	             __inf    (<B>+<S>)**i
  !	             >_      --------------
  !	             i=C_min       i!
  !
  ! The intention of the present subroutine is to invert this relation
  ! using Newton's method.
  !
!*** NOTE! This hasn't been tested yet. I ran up these equations in about 5 min so no guarantees!
  !.....................................................................

    implicit none

    real(single), intent(in)            :: avBkg, detSrc, detSrcUncert,&
                                           minLikelihood
    real(single), intent(out)           :: avSrc, avSrcUncert
    integer,      intent(out), optional :: status

    integer :: foreignStatus, localStatus!, i
    real(single) ::&
      minDetableCountsGlobalReal, IPP0, IPP1, minAvDetSrc, startAvSrc0,&
      startAvSrc1, dAvDetSrc, uncert!, dummy1, dummy2, dummy3

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0

    if (avBkg <= 0.0) then
      if (present(status)) status = -1
  return
    end if

    if (minLikelihood <= 0.0) then
      if (present(status)) status = -2
  return
    end if

    ! Calculate C_min:
    call minDetPoissonCounts(avBkg, minLikelihood&
      , minDetableCountsGlobalReal, uncert, status=foreignStatus)

!print *, ''
!print *, 'avBkg, minLikelihood, minDetableCountsGlobal:', avBkg, minLikelihood, minDetableCountsGlobalReal

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus - 5
  return
    end if

    minDetableCountsGlobal = ceiling(minDetableCountsGlobalReal)
    if (minDetableCountsGlobal <= 0) then
      if (present(status)) status = -3
  return
    end if

    ! Calc minAvDetSrc, which is avDetSrc for avSrc = 0:
    IPP0 = integratedPoissonProb(avBkg, minDetableCountsGlobal)
    IPP1 = integratedPoissonProb(avBkg, minDetableCountsGlobal-1)
    minAvDetSrc = avBkg * IPP1 / IPP0 - avBkg
    if (detSrc <= minAvDetSrc) then
      if (present(status)) status = -4
  return
    end if

    avBkgGlobal  = avBkg
    detSrcGlobal = detSrc
    startAvSrc0 = 0.0
    startAvSrc1 = detSrc

!dummy1 = y(0.0) + detSrc
!print *, 'avBkg, IPP0, IPP1:', avBkg, IPP0, IPP1
!print *, 'minAvDetSrc, y+detSrc, detSrc:', minAvDetSrc, dummy1, detSrc

!do i = 1, 100
!  avSrc = (i / 100.0) * 2.0 * minAvDetSrc
!  dummy1 = y(avSrc-0.1)
!  dummy2 = y(avSrc+0.1)
!  dummy3 = dy(avSrc)
!  print *, 'x, y, dy, dy:', avSrc, (dummy2 + dummy1) / 2.0, (dummy2 - dummy1) / 0.2, dummy3
!end do

    call hybridNewtonsMethod(startAvSrc0, startAvSrc1, y, dy, avSrc&
      , status=foreignStatus)

!print *, 'avSrc:', avSrc

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus - 5
  return
    else if (foreignStatus > 0) then
      localStatus = foreignStatus
    end if

    dAvDetSrc = dy(avSrc)
    if (dAvDetSrc <= 0.0) then
      if (present(status)) status = foreignStatus - 5
  return
    end if
    avSrcUncert = detSrcUncert / dAvDetSrc

    if (present(status)) status = localStatus

  end subroutine inverseEddington

  !---------------------------------------------------------------------
  real function y(avSrc)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: avSrc
    real :: avTotal, IPP0, IPP1

    !-------------------------------------------------------------------
    ! Start

    avTotal = avSrc + avBkgGlobal
    IPP0 = integratedPoissonProb(avTotal, minDetableCountsGlobal)
    IPP1 = integratedPoissonProb(avTotal, minDetableCountsGlobal-1)
    y = avTotal * IPP1 / IPP0 - (detSrcGlobal + avBkgGlobal)
!print *, 'avTotal, IPP0, IPP1, y:'&
!, avTotal, IPP0, IPP1, avTotal * IPP1 / IPP0 - avBkgGlobal

  end function y

  !---------------------------------------------------------------------
  real function dy(avSrc)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: avSrc
    real :: x, IPP0, IPP1, IPP2!, fred

    !-------------------------------------------------------------------
    ! Start

    x = avSrc + avBkgGlobal
    IPP0 = integratedPoissonProb(x, minDetableCountsGlobal)
    IPP1 = integratedPoissonProb(x, minDetableCountsGlobal-1)
    IPP2 = integratedPoissonProb(x, minDetableCountsGlobal-2)
    dy = (IPP1 + x * (IPP2 - IPP1 - IPP1 * (IPP1 - IPP0) / IPP0)) / IPP0
!fred = (IPP2 - IPP1 - IPP1 * (IPP1 - IPP0) / IPP0
!dy = (IPP1 + x * fred) / IPP0

  end function dy

  !---------------------------------------------------------------------
  real function dy_20030728(avSrc)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: avSrc
    real :: x, IPP0, IPP1, IPP2

    !-------------------------------------------------------------------
    ! Start

    x = avSrc + avBkgGlobal
    IPP0 = integratedPoissonProb(x, minDetableCountsGlobal)
    IPP1 = integratedPoissonProb(x, minDetableCountsGlobal-1)
    IPP2 = integratedPoissonProb(x, minDetableCountsGlobal-2)
    dy_20030728 = ((IPP1 + x * IPP2) / IPP0 - x * IPP1 * IPP1) / IPP0

  end function dy_20030728

  !---------------------------------------------------------------------
  subroutine chi2Prob(chi2, degFree, probability, status)
  !---------------------------------------------------------------------
  ! This subroutine gives the probability that a given chi2/degFree is
  ! 'natural' - ie it gives the probability of the null hypothesis. Chi2
  ! (chi squared) is defined for normally distributed data as follows:
  !
  !	                     (y_i - e_i)^2
  !	chi^2 = sum[i=1, N] ---------------
  !	                      sigma_e_i^2
  !
  ! where y_i is the measured or data value, e_i is the model or expected
  ! value and sigma_e_i is the uncertainty in the latter.

    implicit none

    real(single), intent(in)            :: chi2
    integer,      intent(in)            :: degFree
    real(single), intent(out)           :: probability
    integer,      intent(out), optional :: status

    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start

    probability = 0.0

    if (chi2 .le. 0.0) then
      if (present(status)) status = -1
  return
    end if
    if (degFree .le. 0) then
      if (present(status)) status = -2
  return
    end if

    call incompleteGammaQ(degFree / 2.0, chi2 / 2.0, probability, foreignStatus)
    ! in math_utils

    if (present(status)) status = foreignStatus

  end subroutine chi2Prob

  !---------------------------------------------------------------------
  real function poissonProbSingle(av, realI)
  !---------------------------------------------------------------------
  ! Returns the Poisson probability of the occurrence of i events when
  ! the ensemble average = av. This probability distribution is given by
  !
  !	        av^i exp(-av)
  !	P(i) = ---------------
  !	             i!
  !
  ! This routine was adapted from material in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in) :: av, realI

    !-------------------------------------------------------------------
    ! Start

    poissonProbSingle = exp(realI * log(av) - av - lnGamma(realI + 1.0))

  end function poissonProbSingle

  !---------------------------------------------------------------------
  real function poissonProbInt32(av, i)
  !---------------------------------------------------------------------
  ! Returns the Poisson probability of the occurrence of i events when
  ! the ensemble average = av. This probability distribution is given by
  !
  !	        av^i exp(-av)
  !	P(i) = ---------------
  !	             i!
  !
  ! This routine was adapted from material in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single),   intent(in) :: av
    integer(int32), intent(in) :: i

    logical :: firstEntry = .true.
    integer, save :: threshhold
    real(single) :: logHugeInt, nOld, nNew

    !-------------------------------------------------------------------
    ! Start

    poissonProbInt32 = 0.0

  if (av <= 0) return
  if (i < 0) return

    if (firstEntry) then ! calculate the threshhold value corresponding to
    ! the max integer size of the present platform:

      firstEntry = .false.

      ! Invert Stirling's formula for factorial(n) by iteration:
      logHugeInt = log(real(huge(i)))
      nOld = logHugeInt
      do
        nNew = (logHugeInt - 0.5 * log(TwoPi * nOld)) / (log(nOld) - 1.0)
      if (abs(nNew - nOld) < 1.0) exit
        nOld = nNew
      end do

      threshhold = floor(nNew - 1) ! the minus 1 just for a bit of elbow room.
    end if

    if (i < threshhold) then
      poissonProbInt32 = exp(i * log(av) - av) / factorial(i)

    else
      poissonProbInt32 = exp(i * log(av) - av - lnGamma(i + 1.0))
    end if

  end function poissonProbInt32

  !---------------------------------------------------------------------
  real(single) function integratedPoissonProb(av, i)
  !---------------------------------------------------------------------
  ! Returns the probability that an integer random variable which obeys
  ! a Poisson distribution about the average 'av' will EQUAL OR EXCEED i.

    implicit none

    real(single), intent(in) :: av
    integer, intent(in) :: i

    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start

    if (av < 0.0) then
      integratedPoissonProb = 0.0
  return
    end if

    if (i <= 0) then
      integratedPoissonProb = 1.0
  return
    end if

    call incompleteGammaP(real(i), av, integratedPoissonProb, foreignStatus)
    ! in math_utils

    if (foreignStatus .ne. 0) integratedPoissonProb = 0.0

  end function integratedPoissonProb

  !---------------------------------------------------------------------
  subroutine integratedPoissonProb_20030516(testY, meanY, probability, status)
  !---------------------------------------------------------------------
  ! Returns the probability that a Poisson-distributed (integer) variable
  ! y will exceed testY.

    implicit none

    integer, intent(in) :: testY
    real(single), intent(in) :: meanY
    real(single), intent(out) :: probability
    integer, intent(out) :: status

    !-------------------------------------------------------------------
    ! Start

    status = 0
    probability = 0.0
    if (testY .lt. 0) then
      status = -1
  return
    end if
    if (meanY .lt. 0.0) then
      status = -2
  return
    end if

    call incompleteGammaP(real(testY + 1), meanY, probability, status)
    ! in math_utils

  end subroutine integratedPoissonProb_20030516

  !---------------------------------------------------------------------
  subroutine minDetPoissonCountsScalar(bkgCount, likelihoodCutoff&
  , detectableSrcCount, detectableSrcCountUncert, status)
  !---------------------------------------------------------------------
  ! Given a probability distribution P of event counts, any sample value c is associated with a certain probability
  ! P_int (and therefore likelihood L = -ln(P_int)) of it _not_ being due
  ! to chance. This probability is obtained by summing or integrating the
  ! probability distribution from j=c to j=infinity, which sum happens,
  ! for the Poisson probability distribution assumed here, to be equal to
  ! the incomplete Gamma function P(c,a). (Note that this is valid whether
  ! c is integer or not.)
  !
  ! The present subroutine inverts this relation to return that value of
  ! c (=counts) which is associated with the specified pair (L,a). (Ideally
  ! there would be some nifty series for this inversion but I haven't come
  ! across one. Failing that I would have liked to perform this inversion
  ! by use of Newton's method but I can't find an expression for
  ! dP_int/dc. Ridders' method is the best I can come up with at present.)
  !
  ! Note: This routine is ONLY valid if the background can itself be
  ! assumed to be a known value (ie, of insignificant uncertainty). If
  ! the background has had to be estimated then the stats get more
  ! complicated and simple Poisson statistics no longer apply.
  !.....................................................................

    implicit none

    real(single), intent(in)            :: bkgCount, likelihoodCutoff
    real(single), intent(out)           :: detectableSrcCount
    real(single), intent(out)           :: detectableSrcCountUncert
    integer,      intent(out), optional :: status

    real(single) ::&
      prob, oldTotalCount, totalCount, totalCount0, totalCount1,&
      convergenceCriterion
    integer :: localStatus, foreignStatus

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '  minDetPoissonCountsScalar: aaa'
    localStatus = 0
    detectableSrcCount  = 0.0 ! default
    detectableSrcCountUncert = 0.0 ! default

    ! Check for wrong and special arguments:

    if (likelihoodCutoff <= 0.0) then
      if (present(status)) status = -1
  return
    end if
    if (bkgCount <= 0.0) then
      if (present(status)) status = -2
  return
    end if

    !...................................................................
    ! Set up globals:

    targetLikelihood = likelihoodCutoff
    targetProb = exp(-likelihoodCutoff)
    globalBkgCount = bkgCount

    !...................................................................
    ! Start with counts = 1.0 and multiply or divide it by 2 until the
    ! solution is bracketted:

    oldTotalCount = 1.0
    call incompleteGammaP(oldTotalCount, bkgCount, prob) ! in math_utils
    if (prob == targetProb) then
      detectableSrcCount = oldTotalCount - bkgCount
  return

    else if (prob < targetProb) then ! decrease the counts:
      do
        totalCount = oldTotalCount / 2.0
        call incompleteGammaP(totalCount, bkgCount, prob) ! in math_utils
      if (prob >= targetProb) exit
        oldTotalCount = totalCount
      end do

      if (prob == targetProb) then
        detectableSrcCount = totalCount - bkgCount
  return
      end if

      totalCount0 = totalCount
      totalCount1 = oldTotalCount

    else ! increase the counts:
      do
        totalCount = oldTotalCount * 2.0
        call incompleteGammaP(totalCount, bkgCount, prob) ! in math_utils
      if (prob <= targetProb) exit
        oldTotalCount = totalCount
      end do

      if (prob == targetProb) then
        detectableSrcCount = totalCount - bkgCount
  return
      end if

      totalCount0 = totalCount
      totalCount1 = oldTotalCount
    end if

    !...................................................................
    ! Do Ridders'-method inversion:

    convergenceCriterion = (totalCount0 + totalCount1) / 2000.0

    call riddersMethod(totalCount0, totalCount1, yScalar, totalCount&
      , maxNumIterations=30, convergenceCriterion=convergenceCriterion&
      , dx=detectableSrcCountUncert, status=foreignStatus)
    ! in minimizations

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus - 2
!      detectableSrcCount = 0.0
  return
    else if (foreignStatus > 0) then
      localStatus = foreignStatus
    end if

    detectableSrcCount = totalCount - bkgCount

    if (present(status)) status = localStatus
if (testFlag) print *, '  minDetPoissonCountsScalar: zzz'

  end subroutine minDetPoissonCountsScalar

  !---------------------------------------------------------------------
  real(single) function yScalar(totalCount)
!real function yScalar(totalCount)
  !---------------------------------------------------------------------
  ! Note that globalBkgCount and targetLikelihood are globals and should be given sensible values before the function is called. There is no protection in this function for undefined or incorrect values.

    implicit none

    real(single), intent(in) :: totalCount
!real, intent(in) :: totalCount

    real(single) :: prob

    !-------------------------------------------------------------------
    ! Start

    call incompleteGammaP(totalCount, globalBkgCount, prob) ! in math_utils
    yScalar = targetLikelihood + log(prob)

  end function yScalar

  !---------------------------------------------------------------------
  subroutine minDetPoissonCountsVector(bkgCounts, likelihoodCutoff&
  , detectableSrcCounts, detectableSrcCountsUncert, srcCountRatios, status)
  !---------------------------------------------------------------------
  ! Given a probability distribution P of event counts having an average
  ! value a, any sample value c is associated with a certain probability
  ! P_int (and therefore likelihood L = -ln(P_int)) of it _not_ being due
  ! to chance. This probability is obtained by summing or integrating the
  ! probability distribution from j=c to j=infinity, which sum happens,
  ! for the Poisson probability distribution assumed here, to be equal to
  ! the incomplete Gamma function P(c,a). (Note that this is valid whether
  ! c is integer or not.)
  !
  ! The present subroutine inverts this relation to return that value of
  ! c (=counts) which is associated with the specified pair (L,a). (Ideally
  ! there would be some nifty series for this inversion but I haven't come
  ! across one. Failing that I would have liked to perform this inversion
  ! by use of Newton's method but I can't find an expression for
  ! dP_int/dc. Ridders' method is the best I can come up with at present.)
  !
  ! The present subroutine is adapted to the situation in which the same
  ! piece of sky is observed in several energy bands. A single nett
  ! likelihood value is calculated from the sum of the likelihoods L_i for the
  ! N different bands. The sum over individual band likelihoods has a
  ! probability distribution similar to that of chi^2 divided by 2, where
  ! chi^2 a number of degrees of freedom roughly equal to the number of
  ! bands N times 2. The nett likelihood that the detected counts in all
  ! bands are _not_ produced just by background can therefore be shown
  ! to equal -ln(Q[f{N}, sum{L_i,i=1,N}]) where Q is the complement
  ! incomplete gamma function and f(x) is an approximately linear function
  ! of x of slope 1. Monte Carlo studies show that f(2) ~ 2, f(5) ~ 4,
  ! f(10) ~ 8 and so forth; however eboxdetect at the present time
  ! assumes that f(N) = N, so that assumption has been built into the
  ! present subroutine as well.
  !
  ! In this case of many-band detection it is only possible, given a cutoff
  ! value of the summed likelihood, to return the associated band-specific
  ! source count values if the ratios of the distribution of source counts
  ! across the bands are known a priori. In effect, one must know or assume
  ! the source spectrum - all the present subroutine can do is normalize this
  ! spectrum to the cutoff value of summed likelihood.
  !
  ! Notes:
  !	(1) This routine is ONLY valid if the background can itself be
  ! assumed to be a known value (ie, of insignificant uncertainty). If
  ! the background has had to be estimated then the stats get more
  ! complicated and simple Poisson statistics no longer apply.
  !	(2) The subroutine cannot be called twice from the same program
  ! with a different value of numBands unless this is set to zero and the
  ! vectors globalSrcCountRatios and globalBkgCounts deallocated in between
  ! calls.
  !.....................................................................

    implicit none

    real(single), intent(in)            :: srcCountRatios(:),&
                                           bkgCounts(size(srcCountRatios)),&
                                           likelihoodCutoff
    real(single), intent(out)           :: detectableSrcCounts(&
                                             size(srcCountRatios))
    real(single), intent(out)           :: detectableSrcCountsUncert(&
                                             size(srcCountRatios))
    integer,      intent(out), optional :: status

    real(single) ::&
      prob, totalProb,&
      oldCommonSrcCount, commonSrcCount, commonSrcCount0, commonSrcCount1,&
      commonSrcCountUncert, convergenceCriterion, lSum
    integer :: band, localStatus, foreignStatus
    logical :: firstTime = .true. ! preset variables are saved

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '  minDetPoissonCountsVector: aaa'

    if (firstTime) then
      numBands = size(srcCountRatios)
    else if (numBands .ne. size(srcCountRatios)) then
call error('',"")
    end if

    if (size(srcCountRatios) == 1) then
      call minDetPoissonCountsScalar(bkgCounts(1), likelihoodCutoff&
        , detectableSrcCounts(1), detectableSrcCountsUncert(1), localStatus)

      if (present(status)) status = localStatus
  return
    end if

    localStatus = 0
    if (present(status)) status = localStatus
    detectableSrcCounts       = 0.0 ! default
    detectableSrcCountsUncert = 0.0 ! default

    ! Check for wrong and special arguments:

    if (likelihoodCutoff <= 0.0) then
      if (present(status)) status = -1
  return
    end if
    if (size(srcCountRatios) <= 0) then
      if (present(status)) status = -2
  return
    end if
    if (any(srcCountRatios <= 0.0)) then
      if (present(status)) status = -3
  return
    end if
    if (any(bkgCounts <= 0.0)) then
      if (present(status)) status = -4
  return
    end if
if (testFlag) print *, '  bbb'

    ! Check that >0 source counts returns a value of likelihood which
    ! is greater than the target one.

if (testFlag) print *, '  numBands=', numBands, '; should='&
, size(srcCountRatios)
    if (numBands > 1) then
      lSum = 0.0
      do band = 1, numBands
        call incompleteGammaP(bkgCounts(band), bkgCounts(band), prob)
        ! in math_utils

        lSum = lSum - log(prob)
      end do
      call incompleteGammaQ(real(numBands, single), lSum, totalProb)
      ! in math_utils

      if (totalProb < targetProb) then
        if (present(status)) status = -5
  return
      end if

    else ! only 1 band.
      call incompleteGammaP(bkgCounts(1), bkgCounts(1), prob)
      ! in math_utils

      if (prob < targetProb) then
        if (present(status)) status = -5
if (testFlag) print *, '  prob, targetProb:', prob, targetProb
  return
      end if
    end if

    !...................................................................
    ! Set up globals:

    targetLikelihood = likelihoodCutoff
    targetProb = exp(-likelihoodCutoff)
!    numBands = size(srcCountRatios)
    if (firstTime) then ! then globalSrcCountRatios and globalBkgCounts
    ! have not yet been allocated. Do so:
      firstTime = .false.
      allocate(globalSrcCountRatios(numBands), globalBkgCounts(numBands))
    end if
    globalSrcCountRatios = srcCountRatios
    globalBkgCounts = bkgCounts
if (testFlag) print *, '  ccc'

    !...................................................................
    ! Start with src-only count = 1.0 and multiply or divide it by 2 until the
    ! solution is bracketted:

    oldCommonSrcCount = 1.0
    lSum = 0.0
    do band = 1, numBands
      call incompleteGammaP(oldCommonSrcCount * srcCountRatios(band)&
        + bkgCounts(band), bkgCounts(band), prob)
      ! in math_utils

      lSum = lSum - log(prob)
    end do
    call incompleteGammaQ(real(numBands, single), lSum, totalProb)
    ! in math_utils
if (testFlag) print *, '  ddd'

    if (totalProb == targetProb) then
      detectableSrcCounts = oldCommonSrcCount * srcCountRatios
  return

    else if (totalProb < targetProb) then ! decrease the counts:
if (testFlag) print *, '  totalProb < targetProb'
      do
        commonSrcCount = oldCommonSrcCount / 2.0
        lSum = 0.0
        do band = 1, numBands
          call incompleteGammaP(commonSrcCount * srcCountRatios(band)&
            + bkgCounts(band), bkgCounts(band), prob)
          ! in math_utils

          lSum = lSum - log(prob)
        end do
        call incompleteGammaQ(real(numBands, single), lSum, totalProb)
        ! in math_utils

      if (totalProb >= targetProb) exit
        oldCommonSrcCount = commonSrcCount
      end do

      if (totalProb == targetProb) then
        detectableSrcCounts = commonSrcCount * srcCountRatios
  return
      end if

      commonSrcCount0 = commonSrcCount
      commonSrcCount1 = oldCommonSrcCount

    else ! increase the counts:
if (testFlag) print *, '  totalProb >= targetProb'
      do
        commonSrcCount = oldCommonSrcCount * 2.0
        lSum = 0.0
        do band = 1, numBands
          call incompleteGammaP(commonSrcCount * srcCountRatios(band)&
            + bkgCounts(band), bkgCounts(band), prob)
          ! in math_utils

          lSum = lSum - log(prob)
        end do
        call incompleteGammaQ(real(numBands, single), lSum, totalProb)
        ! in math_utils

      if (totalProb <= targetProb) exit
        oldCommonSrcCount = commonSrcCount
      end do

      if (totalProb == targetProb) then
        detectableSrcCounts = commonSrcCount * srcCountRatios
  return
      end if

      commonSrcCount0 = commonSrcCount
      commonSrcCount1 = oldCommonSrcCount
    end if
if (testFlag) print *, '  eee'

    !...................................................................
    ! Do Ridders'-method inversion:

    convergenceCriterion = (commonSrcCount0 + commonSrcCount1) / 2000.0

    call riddersMethod(commonSrcCount0, commonSrcCount1, yVector&
      , commonSrcCount, maxNumIterations=30&
      , convergenceCriterion=convergenceCriterion&
      , dx=commonSrcCountUncert, status=foreignStatus)
    ! in minimizations
if (testFlag) print *, '  fff'

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus - 5
      detectableSrcCounts = 0.0
  return
    else if (foreignStatus > 0) then
      localStatus = foreignStatus
    end if

    detectableSrcCounts       = commonSrcCount       * srcCountRatios
    detectableSrcCountsUncert = commonSrcCountUncert * srcCountRatios

    if (present(status)) status = localStatus
if (testFlag) print *, '  minDetPoissonCountsVector: zzz'

  end subroutine minDetPoissonCountsVector

  !---------------------------------------------------------------------
  real(single) function yVector(commonSrcCount)
!real function yVector(commonSrcCount)
  !---------------------------------------------------------------------
  ! Note that globalBkgCounts, globalSrcCountRatios and targetLikelihood are globals and should be given sensible values before the function is called. There is no protection in this function for undefined or incorrect values.

    implicit none

    real(single), intent(in) :: commonSrcCount
!real, intent(in) :: commonSrcCount

    real(single) :: totalProb, prob, lSum
    integer :: band

    !-------------------------------------------------------------------
    ! Start

!    totalProb = 1.0
lSum = 0.0
    do band = 1, numBands
      call incompleteGammaP(commonSrcCount * globalSrcCountRatios(band)&
        + globalBkgCounts(band), globalBkgCounts(band), prob)
      ! in math_utils
!      totalProb = totalProb * prob
lSum = lSum - log(prob)
    end do
call incompleteGammaQ(real(numBands, single), lSum, totalProb)
! in math_utils
    yVector = targetLikelihood + log(totalProb)

  end function yVector

  !---------------------------------------------------------------------
  function integratedGaussProb(testY, meanY, sigma) result(probability)
  !---------------------------------------------------------------------
  ! Returns the probability that a gaussian-distributed variable y will
  ! depart from the mean by greater than abs(testY - meanY). For values
  ! obeying a gaussian distribution of given mean and
  ! standard deviation s, the probability of a given y value (or greater)
  ! occurring by chance is 1 - erf(abs(testY - meanY)/s/sqrt(2))

    implicit none

    real(single), intent(in) :: testY, meanY, sigma
    real(single) :: probability

    !-------------------------------------------------------------------
    ! Start

    probability = erfc(real(abs(testY - meanY) / sigma / RootTwo))
    ! in math_utils

  end function integratedGaussProb

  !---------------------------------------------------------------------
  function calcChi2HalfDegFree(numLikelihoods) result(f)
  !---------------------------------------------------------------------
  ! A sum over likelihoods seems to have a probability distribution similar to that of a chi^2 distribution with 2f degrees of freedom. The factor f is approximately equal to the number of likelihoods in the sum but not quite. However for the time being (until more accurate information is available, that is) we will assume that the equality holds exactly.

    implicit none

    integer(int16), intent(in) :: numLikelihoods
    real(single)               :: f

    !-------------------------------------------------------------------
    ! Start

!!!    f = real(numLikelihoods, single)
!*** just for the time being, until we manage to calculate or estimate the true relationship, which appears to be not quite linear.

    if (numLikelihoods > 1) then
      f = 1.0 + (numLikelihoods - 1.0) * 8.1
    else
      f = real(numLikelihoods, single)
    end if

  end function calcChi2HalfDegFree

  !---------------------------------------------------------------------
  subroutine sensyDegradnFromGaussSmooth(expectedBkgCount, likelihoodCutoff&
  , gaussWidth, minDetSrcCountsBefore, minDetSrcCountsAfter, deltaLikelihood&
  , status)
  !---------------------------------------------------------------------
  ! Suppose we have an image of random values, each following the Poisson distribution, whose parent image is a single source (1 pixel wide) upon a uniform background. Suppose further that we know (or can estimate) the expectation value <B> of the background. The likelihood that the total counts C=<B>+S at the 'source' pixel (x,y) are due to chance fluctuations of the background is given by
  !
  !	L = -ln(P[C{x,y},<B>])
  !
  ! where P is the incomplete Gamma function. Suppose now we smooth the input image by a gaussian, such that the new value B_1 for the counts at pixel (x,y) is given by
  !
  !	            __ __
  !	            \  \  exp(-[{x-i}^2+{y-j}^2]/2/sigma/sigma) * C_0(i,j)
  !	            /_ /_
  !	            i  j
  !	B_1(x,y) = --------------------------------------------------------.
  !	                __ __
  !	                \  \  exp(-[{x-i}^2+{y-j}^2]/2/sigma/sigma)
  !	                /_ /_
  !	                i  j
  !
  ! If we approximate C_0(i,j) as equal to <B> for all (i,j) not equal to (x,y) then this can be written as
  !
  !	                              S_0
  !	B_1 ~ <B_0> + -------------------------------------.
  !	               __ __
  !	               \  \  exp(-(i^2+j^2)/2/sigma/sigma)
  !	               /_ /_
  !	               i  j
  !
  ! If we use this smoothed image as a estimated background map for the detection of the source, clearly there will be some degradation in detection sensitivity, depending on three factors: the 'real' background <B_0>, the value of S_0, and the degree of smoothing. It is the function of the present subroutine to calculate the amount of this degradation.
  !
  ! The first step in the algorithm is to calculate S_0. That value is calculated which would just be detectable at the given values of expectedBkgCount and likelihoodCutoff. This number is returned in minDetSrcCountsBefore.
  !
  ! Next we calculate B_1. This value is then used to calculate two final numbers: the new likelihood L_new is calculated from
  !
  !	L_new = -ln(P[C{x,y},<B_1>])
  !
  ! and a new minimum detectable count is obtained by inverting
  !
  !	L = -ln(P[C_new{x,y},<B_1>]).
  !
  ! L minus L_new is returned in deltaLikelihood and S_new = C_new-<B_1> as minDetSrcCountsAfter.
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)            :: expectedBkgCount,&
                                           likelihoodCutoff,&
                                           gaussWidth
    real(single), intent(out)           :: minDetSrcCountsBefore,&
                                           minDetSrcCountsAfter,&
                                           deltaLikelihood
    integer,      intent(out), optional :: status

    real(single) ::&
      detectableSrcCountUncert, totalCount, bkgAfterSmoothing, prob
    integer :: foreignStatus, halfSize, i
    real(single), allocatable :: gaussArray(:,:), gaussVector(:)

    !-------------------------------------------------------------------
    ! Start

    status = 0

    call minDetPoissonCountsScalar(expectedBkgCount, likelihoodCutoff&
      , minDetSrcCountsBefore, detectableSrcCountUncert, foreignStatus)

!*** check foreignStatus

    totalCount = minDetSrcCountsBefore + expectedBkgCount

    halfSize = floor(5.0 * gaussWidth)
    allocate(gaussArray(-halfSize:halfSize,-halfSize:halfSize))
    allocate(gaussVector(-halfSize:halfSize))
    gaussVector = (/(i/gaussWidth, i=-halfSize,halfSize)/)
    gaussVector = gaussVector * gaussVector
    gaussVector = exp(-gaussVector/2.0)
    gaussArray = spread(gaussVector,2,2*halfSize+1)&
      * spread(gaussVector,1,2*halfSize+1)

    bkgAfterSmoothing = minDetSrcCountsBefore / sum(gaussArray)

    call incompleteGammaP(totalCount, bkgAfterSmoothing, prob) ! in math_utils
    deltaLikelihood = likelihoodCutoff + log(prob)

    call minDetPoissonCountsScalar(bkgAfterSmoothing, likelihoodCutoff&
      , minDetSrcCountsAfter, detectableSrcCountUncert, foreignStatus)

!*** check foreignStatus

  end subroutine sensyDegradnFromGaussSmooth

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module detection_stats
