!
! Module:dss_ranges_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains auxiliary routines to do with the data subspace (DSS).
! 
! Comments: TODO:
!	- change status values to error calls.
! 
! $Id: dss_ranges_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module dss_ranges_aux

  use types
  use errorHandling
  use dss

  use sort_mod		! for quickSortStatic
  use ssc_misc		! for fracVal.
  use array_utils	! for smallestTrueIndex, largestTrueIndex.

  implicit none 

  private 

  public initializeBoundT
  public copyBound
  public sortBounds
  public swapBounds
  private swapRangeBounds
  public dumpBounds

  public initializeRangeT
  public copyRange
  public checkRangeOverlap
  public xorRangePair
  public rangeToStr
  public rangeTypeToStr
  public valueIsWithinRange

  public rangesAreWellFormed
  public rangesPairIsIdentical
  public copyRanges
  public correctRanges
  public sortRanges
  public integrateRanges
  public dumpRanges

  public andRanges
  public orRanges

  interface dumpBounds
    module procedure dumpBoundsScalar, dumpBoundsVector
  end interface

  interface valueIsWithinRange
    module procedure valueIsWithinRangeSingle, valueIsWithinRangeInt32
    module procedure valueIsWithinRangeDouble
  end interface

  interface integrateRanges
    module procedure integrateRangesFnScalar, integrateRangesFnVector
    module procedure integrateRangesScalar,   integrateRangesVector
    module procedure integrateRangesSingleScalar
  end interface

  interface dumpRanges
    module procedure dumpRangesScalar, dumpRangesVector
  end interface

  interface andRanges
    module procedure andRangesScalVec,    andRangesScalScalVec
    module procedure andRangesScalVecVec, andRangesVecScalVec
    module procedure andRangesVecVec,     andRangesVecVecVec
    module procedure andRangesScalScal,   andRangesScalScalScal
  end interface

  interface orRanges
    module procedure orRangesScalVec,    orRangesScalScalVec
    module procedure orRangesScalVecVec, orRangesVecScalVec
    module procedure orRangesVecVec,     orRangesVecVecVec
    module procedure orRangesScalScal,   orRangesScalScalScal
  end interface


  integer, public, parameter ::&
    UNDECIDED  = -1,&
    BISTOOLOW  =  1,&
    BISTOOHIGH =  2,&
    BOVERLAPSA =  3,&
    BCONTAINSA =  4,&
    BISWITHINA =  5,&
    BEQUALSA   =  6

contains 

  !---------------------------------------------------------------------
  subroutine initializeBoundT(bound)
  !---------------------------------------------------------------------

    implicit none

    type(BoundT), intent(out) :: bound

    !-------------------------------------------------------------------
    ! Start

    bound%type  = INCLUSIVE
    bound%value = 0.0d0

  end subroutine initializeBoundT

  !---------------------------------------------------------------------
  function copyBound(inBound) result(outBound)
  !---------------------------------------------------------------------
  ! Copies over all values between two variables of dss type BoundT. This is called as in the following example:
  !
  !	outBound = copyBound(inBound)

    implicit none

    type(BoundT), intent(in) :: inBound
    type(BoundT) :: outBound

    !-------------------------------------------------------------------
    ! Start

    outBound%type  = inBound%type
    outBound%value = inBound%value

  end function copyBound

  !---------------------------------------------------------------------
  subroutine initializeRangeT(range)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(out) :: range

    !-------------------------------------------------------------------
    ! Start

    call initializeBoundT(range%lower)
    call initializeBoundT(range%upper)

  end subroutine initializeRangeT

  !---------------------------------------------------------------------
  function copyRange(inRange) result(outRange)
  !---------------------------------------------------------------------
  ! Copies over all values between two variables of dss type RangeT. This is called as in the following example:
  !
  !	outRange = copyRange(inRange)

    implicit none

    type(RangeT), intent(in) :: inRange
    type(RangeT) :: outRange

    !-------------------------------------------------------------------
    ! Start

    outRange%lower%type  = inRange%lower%type
    outRange%lower%value = inRange%lower%value
    outRange%upper%type  = inRange%upper%type
    outRange%upper%value = inRange%upper%value

  end function copyRange

  !---------------------------------------------------------------------
  subroutine swapRangeBounds(range)
  !---------------------------------------------------------------------
  ! No reason to make this public. It is mainly used by correctRanges.

    implicit none

    type(RangeT), intent(inout) :: range

    type(BoundT) :: tempLo

    !-------------------------------------------------------------------
    ! Start

    tempLo      = range%lower
    range%lower = range%upper
    range%upper = tempLo

  end subroutine swapRangeBounds

  !---------------------------------------------------------------------
  subroutine dumpBoundsScalar(bound, numIndentSpacesArg)
  !---------------------------------------------------------------------

    implicit none

    type(BoundT), intent(in)           :: bound
    integer,      intent(in), optional :: numIndentSpacesArg

    integer :: numIndentSpaces = 0

    !-------------------------------------------------------------------
    ! Start

    if (present(numIndentSpacesArg)) numIndentSpaces = numIndentSpacesArg

    select case(bound%type)
    case(UNDEFINED)
      print *, repeat(' ', numIndentSpaces)&
        //'Bound type = UNDEFINED.'
    case(INCLUSIVE)
      print *, repeat(' ', numIndentSpaces)&
        //'Bound = ', bound%value, ' (INCLUSIVE)'
    case(EXCLUSIVE)
      print *, repeat(' ', numIndentSpaces)&
        //'Bound = ', bound%value, ' (EXCLUSIVE)'
    case default
      print *, repeat(' ', numIndentSpaces)&
        //'Bound = ', bound%value, ' (type=???)'
    end select

  end subroutine dumpBoundsScalar

  !---------------------------------------------------------------------
  subroutine dumpBoundsScalar_20040506(bound, indentStrArg)
  !---------------------------------------------------------------------

    implicit none

    type(BoundT), intent(in)           :: bound
    character(*), intent(in), optional :: indentStrArg

    integer :: numIndentSpaces = 0

    !-------------------------------------------------------------------
    ! Start

    if (present(indentStrArg)) numIndentSpaces = len(indentStrArg)

    select case(bound%type)
    case(UNDEFINED)
      print *, repeat(' ', numIndentSpaces)&
        //'Bound type = UNDEFINED.'
    case(INCLUSIVE)
      print *, repeat(' ', numIndentSpaces)&
        //'Bound = ', bound%value, ' (INCLUSIVE)'
    case(EXCLUSIVE)
      print *, repeat(' ', numIndentSpaces)&
        //'Bound = ', bound%value, ' (EXCLUSIVE)'
    case default
      print *, repeat(' ', numIndentSpaces)&
        //'Bound = ', bound%value, ' (type=???)'
    end select

  end subroutine dumpBoundsScalar_20040506

  !---------------------------------------------------------------------
  subroutine dumpBoundsVector(bounds, numIndentSpacesArg)
  !---------------------------------------------------------------------

    implicit none

    type(BoundT), intent(in)           :: bounds(:)
    integer,      intent(in), optional :: numIndentSpacesArg

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    do i = 1, size(bounds)
      call dumpBoundsScalar(bounds(i), numIndentSpacesArg)
      print *, ''
    end do

  end subroutine dumpBoundsVector

  !---------------------------------------------------------------------
  subroutine dumpBoundsVector_20040506(bounds, indentStrArg)
  !---------------------------------------------------------------------

    implicit none

    type(BoundT), intent(in)           :: bounds(:)
    character(*), intent(in), optional :: indentStrArg

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    do i = 1, size(bounds)
      call dumpBoundsScalar_20040506(bounds(i), indentStrArg)
      print *, ''
    end do

  end subroutine dumpBoundsVector_20040506


  !---------------------------------------------------------------------
  function checkRangeOverlap(rangeA, rangeB) result(status)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in) :: rangeA, rangeB
    integer :: status

    integer :: loStatus, hiStatus
    integer, parameter ::&
      ALO_LT_BLO = 1,&
      ALO_GT_BLO = 2,&
      ALO_EQ_BLO = 3,&
      AHI_LT_BHI = 4,&
      AHI_GT_BHI = 5,&
      AHI_EQ_BHI = 6

    !-------------------------------------------------------------------
    ! Start

    status = UNDECIDED

    ! Dispose first of the cases where at least one of the bounds is UNDEFINED:

    if (rangeA%lower%type == UNDEFINED) then
      if (rangeB%lower%type == UNDEFINED) then
        if (rangeA%upper%type == UNDEFINED) then
          if (rangeB%upper%type == UNDEFINED) then
! A:	<------>
! B:	<------>
            status = BEQUALSA
          else
! A:	<------>
! B:	<-----?
            status = BISWITHINA
          end if
        else
          if (rangeB%upper%type == UNDEFINED) then
! A:	<-----?
! B:	<------>
            status = BCONTAINSA
          else
! A:	<-----?
! B:	<-----?
            if (rangeA%upper%value < rangeB%upper%value) then
              status = BCONTAINSA
            else if (rangeA%upper%value == rangeB%upper%value) then
              if (rangeA%upper%type == INCLUSIVE) then
                if (rangeB%upper%type == INCLUSIVE) then
                  status = BEQUALSA
                else ! rangeB%upper%type == EXCLUSIVE
                  status = BISWITHINA
                end if
              else ! rangeA%upper%type == EXCLUSIVE
                if (rangeB%upper%type == INCLUSIVE) then
                  status = BCONTAINSA
                else ! rangeB%upper%type == EXCLUSIVE
                  status = BEQUALSA
                end if
              end if
            else ! rangeA%upper%value > rangeB%upper%value
              status = BISWITHINA
            end if
          end if
        end if
      else
        if (rangeA%upper%type == UNDEFINED) then
! A:	<------>
! B:	 ?----?>
          status = BISWITHINA
        else
          if (rangeB%upper%type == UNDEFINED) then
! A:	<-----?
! B:	 ?----->
            status = BOVERLAPSA
          else
            if (rangeA%upper%value < rangeB%upper%value) then
! A:	<-----?
! B:	 ?-----?
              status = BOVERLAPSA
            else if (rangeA%upper%value == rangeB%upper%value) then
              if (rangeA%upper%type == INCLUSIVE) then
! A:	<-----I
! B:	 ?----?
                status = BISWITHINA
              else ! rangeA%upper%type == EXCLUSIVE
                if (rangeB%upper%type == INCLUSIVE) then
! A:	<-----O
! B:	 ?----I
                  status = BOVERLAPSA
                else ! rangeB%upper%type == EXCLUSIVE
! A:	<-----O
! B:	 ?----O
                  status = BISWITHINA
                end if
              end if
            else ! rangeA%upper%value > rangeB%upper%value
! A:	<------?
! B:	 ?----?
              status = BISWITHINA
            end if
          end if
        end if
      end if
    else
      if (rangeB%lower%type == UNDEFINED) then
        if (rangeA%upper%type == UNDEFINED) then
          if (rangeB%upper%type == UNDEFINED) then
! A:	 ?----->
! B:	<------>
            status = BCONTAINSA
          else
! A:	 ?----->
! B:	<-----?
            status = BOVERLAPSA
          end if
        else
          if (rangeB%upper%type == UNDEFINED) then
! A:	 ?----?
! B:	<------>
            status = BCONTAINSA
          else
            if (rangeA%upper%value < rangeB%upper%value) then
! A:	 ?----?
! B:	<------?
              status = BCONTAINSA
            else if (rangeA%upper%value == rangeB%upper%value) then
              if (rangeA%upper%type == INCLUSIVE) then
                if (rangeB%upper%type == INCLUSIVE) then
! A:	 ?----I
! B:	<-----I
                  status = BCONTAINSA
                else ! rangeB%upper%type == EXCLUSIVE
! A:	 ?----I
! B:	<-----O
                  status = BOVERLAPSA
                end if
              else ! rangeA%upper%type == EXCLUSIVE
                if (rangeB%upper%type == INCLUSIVE) then
! A:	 ?----O
! B:	<-----I
                  status = BCONTAINSA
                else ! rangeB%upper%type == EXCLUSIVE
! A:	 ?----O
! B:	<-----O
                  status = BCONTAINSA
                end if
              end if
            else ! rangeA%upper%value > rangeB%upper%value
! A:	 ?-----?
! B:	<-----?
              status = BOVERLAPSA
            end if
          end if
        end if
      else
        if (rangeA%upper%type == UNDEFINED) then
          if (rangeB%upper%type == UNDEFINED) then
            if (rangeA%lower%value < rangeB%lower%value) then
! A:	 ?----->
! B:	  ?---->
              status = BISWITHINA
            else if (rangeA%lower%value == rangeB%lower%value) then
              if (rangeA%lower%type == INCLUSIVE) then
                if (rangeB%lower%type == INCLUSIVE) then
! A:	 I----->
! B:	 I----->
                  status = BEQUALSA
                else ! rangeB%lower%type == EXCLUSIVE
! A:	 I----->
! B:	 O----->
                  status = BISWITHINA
                end if
              else ! rangeA%lower%type == EXCLUSIVE
                if (rangeB%lower%type == INCLUSIVE) then
! A:	 O----->
! B:	 I----->
                  status = BCONTAINSA
                else ! rangeB%lower%type == EXCLUSIVE
! A:	 O----->
! B:	 O----->
                  status = BEQUALSA
                end if
              end if
            else ! rangeA%lower%value > rangeB%lower%value
! A:	  ?---->
! B:	 ?----->
              status = BCONTAINSA
            end if
          else
            if (rangeA%lower%value < rangeB%lower%value) then
! A:	 ?----->
! B:	  ?---?
              status = BISWITHINA
            else if (rangeA%lower%value == rangeB%lower%value) then
              if (rangeA%lower%type == INCLUSIVE) then
! A:	 I----->
! B:	 ?----?
                status = BISWITHINA
              else ! rangeA%lower%type == EXCLUSIVE
                if (rangeB%lower%type == INCLUSIVE) then
! A:	 O----->
! B:	 I----?
                  status = BOVERLAPSA
                else ! rangeB%lower%type == EXCLUSIVE
! A:	 O----->
! B:	 O----?
                  status = BISWITHINA
                end if
              end if
            else ! rangeA%lower%value > rangeB%lower%value
! A:	  ?---->
! B:	 ?----?
              status = BOVERLAPSA
            end if
          end if
        else
          if (rangeB%upper%type == UNDEFINED) then
            if (rangeA%lower%value < rangeB%lower%value) then
! A:	 ?-----?
! B:	  ?----->
              status = BOVERLAPSA
            else if (rangeA%lower%value == rangeB%lower%value) then
              if (rangeA%lower%type == INCLUSIVE) then
                if (rangeB%lower%type == INCLUSIVE) then
! A:	 I-----?
! B:	 I------>
                  status = BCONTAINSA
                else ! rangeB%lower%type == EXCLUSIVE
! A:	 I-----?
! B:	 O------>
                  status = BOVERLAPSA
                end if
              else ! rangeA%lower%type == EXCLUSIVE
! A:	 O-----?
! B:	 ?------>
                status = BCONTAINSA
              end if
            else ! rangeA%lower%value > rangeB%lower%value
! A:	  ?----?
! B:	 ?------>
              status = BCONTAINSA
            end if
          else
! A:	 ?-----?
! B:	 ?-----?
! Leave UNDECIDED to avoid too deep nested ifs.
          end if
        end if
      end if
    end if


  if (status .ne. UNDECIDED) return

    if (rangeA%upper%value < rangeB%lower%value) then
      status = BISTOOHIGH
    else if (rangeA%upper%value == rangeB%lower%value) then
      if   (rangeA%upper%type == INCLUSIVE&
      .and. rangeB%lower%type == INCLUSIVE) then
        status = BOVERLAPSA
      else
        status = BISTOOHIGH
      end if
    end if

  if (status .ne. UNDECIDED) return

    if (rangeA%lower%value > rangeB%upper%value) then
      status = BISTOOLOW
    else if (rangeA%lower%value == rangeB%upper%value) then
      if   (rangeA%lower%type == INCLUSIVE&
      .and. rangeB%upper%type == INCLUSIVE) then
        status = BOVERLAPSA
      else
        status = BISTOOLOW
      end if
    end if


  if (status .ne. UNDECIDED) return

    ! From here, BISTOOLOW and BISTOOHIGH are no longer possible.


    if (rangeA%lower%type == INCLUSIVE) then
      if (rangeB%lower%type == INCLUSIVE) then
        if (rangeA%lower%value > rangeB%lower%value) then
          loStatus = ALO_GT_BLO
        else if (rangeA%lower%value < rangeB%lower%value) then
          loStatus = ALO_LT_BLO
        else ! rangeA%lower%value == rangeB%lower%value
          loStatus = ALO_EQ_BLO
        end if
      else ! rangeB%lower%type == EXCLUSIVE
        if (rangeA%lower%value > rangeB%lower%value) then
          loStatus = ALO_GT_BLO
        else
          loStatus = ALO_LT_BLO
        end if
      end if
    else ! rangeA%lower%type == EXCLUSIVE
      if (rangeB%lower%type == INCLUSIVE) then
        if (rangeA%lower%value >= rangeB%lower%value) then
          loStatus = ALO_GT_BLO
        else
          loStatus = ALO_LT_BLO
        end if
      else ! rangeB%lower%type == EXCLUSIVE
        if (rangeA%lower%value > rangeB%lower%value) then
          loStatus = ALO_GT_BLO
        else if (rangeA%lower%value < rangeB%lower%value) then
          loStatus = ALO_LT_BLO
        else ! rangeA%lower%value == rangeB%lower%value
          loStatus = ALO_EQ_BLO
        end if
      end if
    end if


    if (rangeA%upper%type == INCLUSIVE) then
      if (rangeB%upper%type == INCLUSIVE) then
        if (rangeA%upper%value > rangeB%upper%value) then
          hiStatus = AHI_GT_BHI
        else if (rangeA%upper%value < rangeB%upper%value) then
          hiStatus = AHI_LT_BHI
        else ! rangeA%upper%value == rangeB%upper%value
          hiStatus = AHI_EQ_BHI
        end if
      else ! EXCLUSIVE
        if (rangeA%upper%value >= rangeB%upper%value) then
          hiStatus = AHI_GT_BHI
        else
          hiStatus = AHI_LT_BHI
        end if
      end if
    else ! rangeA%upper%type == EXCLUSIVE
      if (rangeB%upper%type == INCLUSIVE) then
        if (rangeA%upper%value > rangeB%upper%value) then
          hiStatus = AHI_GT_BHI
        else
          hiStatus = AHI_LT_BHI
        end if
      else ! EXCLUSIVE
        if (rangeA%upper%value > rangeB%upper%value) then
          hiStatus = AHI_GT_BHI
        else if (rangeA%upper%value < rangeB%upper%value) then
          hiStatus = AHI_LT_BHI
        else ! rangeA%upper%value == rangeB%upper%value
          hiStatus = AHI_EQ_BHI
        end if
      end if
    end if


    select case(loStatus)
    case(ALO_LT_BLO)
      select case(hiStatus)
      case(AHI_LT_BHI)
        status = BOVERLAPSA
      case(AHI_GT_BHI)
        status = BISWITHINA
      case default ! AHI_EQ_BHI
        status = BISWITHINA
      end select
    case(ALO_GT_BLO)
      select case(hiStatus)
      case(AHI_LT_BHI)
        status = BCONTAINSA
      case(AHI_GT_BHI)
        status = BOVERLAPSA
      case default ! AHI_EQ_BHI
        status = BCONTAINSA
      end select
    case default ! ALO_EQ_BLO
      select case(hiStatus)
      case(AHI_LT_BHI)
        status = BCONTAINSA
      case(AHI_GT_BHI)
        status = BISWITHINA
      case default ! AHI_EQ_BHI
        status = BEQUALSA
      end select
    end select

  end function checkRangeOverlap

  !---------------------------------------------------------------------
  subroutine xorRangePair(range, rangeSubset, xoredRanges)
  !---------------------------------------------------------------------
  ! This returns those fragments (if any) of range that are _not_ part
  ! of rangeSubset.
  !
  ! The pointer xoredRanges should be deallocated after use.

!*** maybe make it more general, so that it returns xor of both input ranges??

    implicit none

    type(RangeT), intent(in) :: range, rangeSubset
    type(RangeT), pointer :: xoredRanges(:)

    logical :: noOverlap, lowerEndRemnantExists, upperEndRemnantExists
    integer(int8) :: numFragments, fragmentNum

    !-------------------------------------------------------------------
    ! Start

    ! Have to accommodate the trivial case in which there is no overlap
    ! at all between range and rangeSubset:

    noOverlap = .false. ! default
    if (range%lower%type .ne. UNDEFINED&
    .and. rangeSubset%upper%type .ne. UNDEFINED) then
      if (range%lower%type == INCLUSIVE) then
        if  ((rangeSubset%upper%type == INCLUSIVE&
        .and. rangeSubset%upper%value <  range%lower%value)&
        .or. (rangeSubset%upper%type == EXCLUSIVE&
        .and. rangeSubset%upper%value <= range%lower%value))&
          noOverlap = .true.

      else ! range%lower%type == EXCLUSIVE
        if (rangeSubset%upper%value <= range%lower%value)&
          noOverlap = .true.

      end if

    else if (range%upper%type .ne. UNDEFINED&
    .and. rangeSubset%lower%type .ne. UNDEFINED) then
      if (range%upper%type == INCLUSIVE) then
        if  ((rangeSubset%lower%type == INCLUSIVE&
        .and. rangeSubset%lower%value >  range%upper%value)&
        .or. (rangeSubset%lower%type == EXCLUSIVE&
        .and. rangeSubset%lower%value >= range%upper%value))&
          noOverlap = .true.

      else ! range%upper%type == EXCLUSIVE
        if (rangeSubset%lower%value >= range%upper%value)&
          noOverlap = .true.

      end if
    end if

    if (noOverlap) then
      allocate(xoredRanges(1))
      xoredRanges(1) = copyRange(range)
  return
    end if

    ! If there is overlap, the number of xoredRanges can be 0, 1 or 2, depending on where the rangeSubset lies and the types of boundaries each range has at each end. Another way to think of it is that there can either be or not be a remnant at either end of range after rangeSubset has been removed. The subroutine has been written to check each end in turn and see if there is a remnant there.

    lowerEndRemnantExists = .false. ! default
    select case (range%lower%type)
    case(UNDEFINED)
      if(rangeSubset%lower%type .ne. UNDEFINED)&
        lowerEndRemnantExists = .true.

    case(INCLUSIVE)
      if  ((rangeSubset%lower%type == INCLUSIVE&
      .and. rangeSubset%lower%value >  range%lower%value)&
      .or. (rangeSubset%lower%type == EXCLUSIVE&
      .and. rangeSubset%lower%value >= range%lower%value))&
        lowerEndRemnantExists = .true.

    case(EXCLUSIVE)
      if   (rangeSubset%lower%type .ne. UNDEFINED&
      .and. rangeSubset%lower%value > range%lower%value)&
        lowerEndRemnantExists = .true.

    case default
call error('dss_ranges_aux_bug0',"Shouldn't ever happen, contact the code developer.")
    end select

    upperEndRemnantExists = .false. ! default
    select case (range%upper%type)
    case(UNDEFINED)
      if(rangeSubset%upper%type .ne. UNDEFINED)&
        upperEndRemnantExists = .true.

    case(INCLUSIVE)
      if  ((rangeSubset%upper%type == INCLUSIVE&
      .and. rangeSubset%upper%value <  range%upper%value)&
      .or. (rangeSubset%upper%type == EXCLUSIVE&
      .and. rangeSubset%upper%value <= range%upper%value))&
        upperEndRemnantExists = .true.

    case(EXCLUSIVE)
      if   (rangeSubset%upper%type .ne. UNDEFINED&
      .and. rangeSubset%upper%value < range%upper%value)&
        upperEndRemnantExists = .true.

    case default
call error('dss_ranges_aux_bug1',"Shouldn't ever happen, contact the code developer.")
    end select

    numFragments = 0
    if (lowerEndRemnantExists) numFragments = numFragments + 1
    if (upperEndRemnantExists) numFragments = numFragments + 1
    allocate(xoredRanges(numFragments))
 
    ! Now load up xoredRanges:

    fragmentNum = 0
    if (lowerEndRemnantExists) then
      fragmentNum = fragmentNum + 1
      xoredRanges(fragmentNum)%lower%type  = range%lower%type
      xoredRanges(fragmentNum)%lower%value = range%lower%value
      xoredRanges(fragmentNum)%upper%value = rangeSubset%lower%value

      if (rangeSubset%lower%type == INCLUSIVE) then
        xoredRanges(fragmentNum)%upper%value = EXCLUSIVE

      else ! rangeSubset%lower%type == EXCLUSIVE. UNDEFINED is ruled out
        ! in the test for lowerEndRemnantExists.

        xoredRanges(fragmentNum)%upper%value = INCLUSIVE
      end if
    end if
    
    if (upperEndRemnantExists) then
      fragmentNum = fragmentNum + 1
      xoredRanges(fragmentNum)%upper%type  = range%upper%type
      xoredRanges(fragmentNum)%upper%value = range%upper%value
      xoredRanges(fragmentNum)%lower%value = rangeSubset%upper%value

      if (rangeSubset%upper%type == INCLUSIVE) then
        xoredRanges(fragmentNum)%lower%value = EXCLUSIVE

      else ! rangeSubset%upper%type == EXCLUSIVE. UNDEFINED is ruled out
        ! in the test for upperEndRemnantExists.

        xoredRanges(fragmentNum)%lower%value = INCLUSIVE
      end if
    end if

  end subroutine xorRangePair

  !---------------------------------------------------------------------
  subroutine rangeToStr(range, rangeLoValStr, rangeHiValStr)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in)  :: range
    character(*), intent(out) :: rangeLoValStr, rangeHiValStr

    !-------------------------------------------------------------------
    ! Start

    rangeLoValStr = ''
    if (range%lower%type == UNDEFINED) then
if (len(rangeLoValStr) < 4) call error('',"")
      rangeLoValStr = '-inf'
    else
if (range%lower%value > 999999) call error('',"")
if (range%lower%value < -99999) call error('',"")
if (len(rangeLoValStr) < 9) call error('',"")
      write(rangeLoValStr, '(f9.1)') range%lower%value
    end if
    rangeLoValStr = adjustl(rangeLoValStr)

    rangeHiValStr = ''
    if (range%upper%type == UNDEFINED) then
if (len(rangeHiValStr) < 3) call error('',"")
      rangeHiValStr = 'inf'
    else
if (range%upper%value > 999999) call error('',"")
if (range%upper%value < -99999) call error('',"")
if (len(rangeHiValStr) < 9) call error('',"")
      write(rangeHiValStr, '(f9.1)') range%upper%value
    end if
    rangeHiValStr = adjustl(rangeHiValStr)

  end subroutine rangeToStr

  !---------------------------------------------------------------------
  function rangeTypeToStr(rangeType) result(rangeTypeStr)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: rangeType
    character(9) :: rangeTypeStr

    !-------------------------------------------------------------------
    ! Start

    select case(rangeType)
    case(UNDEFINED)
      rangeTypeStr = 'UNDEFINED'
    case(INCLUSIVE)
      rangeTypeStr = 'INCLUSIVE'
    case(EXCLUSIVE)
      rangeTypeStr = 'EXCLUSIVE'
    case default
      rangeTypeStr = 'UNKNOWN  '
    end select

  end function rangeTypeToStr

  !---------------------------------------------------------------------
  logical(bool) function valueIsWithinRangeSingle(value, range)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: value
    type(RangeT), intent(in) :: range

    !-------------------------------------------------------------------
    ! Start

    valueIsWithinRangeSingle = .false. ! default

    if   (range%lower%type .eq. UNDEFINED&
    .or. (range%lower%type .eq. INCLUSIVE .and. range%lower%value <= value)&
    .or. (range%lower%type .eq. EXCLUSIVE .and. range%lower%value < value))&
    then
      if   (range%upper%type .eq. UNDEFINED&
      .or. (range%upper%type .eq. INCLUSIVE .and. range%upper%value >= value)&
      .or. (range%upper%type .eq. EXCLUSIVE .and. range%upper%value > value))&
      then
        valueIsWithinRangeSingle = .true.
      end if
    end if

  end function valueIsWithinRangeSingle

  !---------------------------------------------------------------------
  logical(bool) function valueIsWithinRangeDouble(value, range)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: value
    type(RangeT), intent(in) :: range

    !-------------------------------------------------------------------
    ! Start

    valueIsWithinRangeDouble = .false. ! default

    if   (range%lower%type .eq. UNDEFINED&
    .or. (range%lower%type .eq. INCLUSIVE .and. range%lower%value <= value)&
    .or. (range%lower%type .eq. EXCLUSIVE .and. range%lower%value < value))&
    then
      if   (range%upper%type .eq. UNDEFINED&
      .or. (range%upper%type .eq. INCLUSIVE .and. range%upper%value >= value)&
      .or. (range%upper%type .eq. EXCLUSIVE .and. range%upper%value > value))&
      then
        valueIsWithinRangeDouble = .true.
      end if
    end if

  end function valueIsWithinRangeDouble

  !---------------------------------------------------------------------
  logical(bool) function valueIsWithinRangeInt32(value, range)
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in) :: value
    type(RangeT), intent(in) :: range

    !-------------------------------------------------------------------
    ! Start

    valueIsWithinRangeInt32 = .false. ! default

    if   (range%lower%type .eq. UNDEFINED&
    .or. (range%lower%type .eq. INCLUSIVE .and. range%lower%value <= value)&
    .or. (range%lower%type .eq. EXCLUSIVE .and. range%lower%value < value))&
    then
      if   (range%upper%type .eq. UNDEFINED&
      .or. (range%upper%type .eq. INCLUSIVE .and. range%upper%value >= value)&
      .or. (range%upper%type .eq. EXCLUSIVE .and. range%upper%value > value))&
      then
        valueIsWithinRangeInt32 = .true.
      end if
    end if

  end function valueIsWithinRangeInt32

  !---------------------------------------------------------------------
  logical(bool) function rangesAreWellFormed(ranges)
  !---------------------------------------------------------------------
  ! This checks that the range vector (of size n) obeys the following
  ! conditions:
  !	(i) Only ranges(1)%lower%type and ranges(n)%upper%type are
  ! permitted to be UNDEFINED.
  !	(ii) For all ranges for which neither lower or upper type
  ! = UNDEFINED (ie, for all internal ranges), upper%value > lower%value
  ! unless both types are INCLUSIVE, in which case upper%value may equal
  ! lower%value. This means that each range must encompass some non-zero
  ! set of real numbers. 
  !	(iii) For all ranges, ranges(i)%upper%value <
  ! ranges(i+1)%lower%value unless both ranges(i)%upper%type and
  ! ranges(i+1)%lower%type are EXCLUSIVE, in which case ranges(i)%upper%value
  ! may equal ranges(i+1)%lower%value. This means that the gap between two
  ! adjacent ranges must encompass some non-zero set of real numbers.
  !
  ! Note that any ranges pointer can be brought into valid condition by
  ! passing it through the subroutine correctRanges().

    implicit none

    type(RangeT), intent(in) :: ranges(:)

    integer :: numRanges

    !-------------------------------------------------------------------
    ! Start

    numRanges = size(ranges)
    if (numRanges <= 0) then
      rangesAreWellFormed = .true.
  return
    end if

    rangesAreWellFormed = .false. ! default

    if (numRanges > 1) then
  if (any(ranges(2:           )%lower%type == UNDEFINED)) return
  if (any(ranges( :numRanges-1)%upper%type == UNDEFINED)) return

  if (any(ranges(2:)%upper%value > ranges(:numRanges-1)%lower%value)) return

      if (any(ranges(2:)%upper%value >= ranges(:numRanges-1)%lower%value&
      .and. ranges(2:           )%upper%type == INCLUSIVE&
      .and. ranges( :numRanges-1)%lower%type == INCLUSIVE))&
  return
    end if

    if (any(ranges%upper%value < ranges%lower%value&
    .and. ranges%upper%type .ne. UNDEFINED&
    .and. ranges%lower%type .ne. UNDEFINED))&
  return

    if (any(ranges%upper%value <= ranges%lower%value&
    .and. ranges%upper%type == EXCLUSIVE&
    .and. ranges%lower%type == EXCLUSIVE))&
  return

  rangesAreWellFormed = .true. ! passed all the tests.

  end function rangesAreWellFormed

  !---------------------------------------------------------------------
  logical(bool) function rangesPairIsIdentical(rangesA, rangesB)
  !---------------------------------------------------------------------
!***** NOTE should compare real values with an 'almost-equal'.

    implicit none

    type(RangeT), intent(in) :: rangesA(:), rangesB(:)

    integer :: numRanges, i

    !-------------------------------------------------------------------
    ! Start

    rangesPairIsIdentical = .false. ! default

  if (size(rangesA) .ne. size(rangesB)) return
    numRanges = size(rangesA)
    do i = 1, numRanges

      if (rangesA(i)%lower%type .ne. rangesB(i)%lower%type&
      .or. rangesA(i)%upper%type .ne. rangesB(i)%upper%type)&
  return

      if (rangesA(i)%lower%type .ne. UNDEFINED&
      .and. rangesA(i)%lower%value .ne. rangesB(i)%lower%value)&
  return

      if (rangesA(i)%upper%type .ne. UNDEFINED&
      .and. rangesA(i)%upper%value .ne. rangesB(i)%upper%value)&
  return
    end do

    rangesPairIsIdentical = .true.

  end function rangesPairIsIdentical

  !---------------------------------------------------------------------
  subroutine copyRanges(inRanges, outRanges)
  !---------------------------------------------------------------------
  ! NOTE! You should deallocate outRanges() after use.

    implicit none

    type(RangeT), intent(in) :: inRanges(:)
    type(RangeT), pointer    :: outRanges(:)

    integer :: numRanges, rangeNum

    !-------------------------------------------------------------------
    ! Start

    numRanges = size(inRanges)
    allocate(outRanges(numRanges))
    do rangeNum = 1, numRanges
      outRanges(rangeNum) = copyRange(inRanges(rangeNum))
    end do

  end subroutine copyRanges

  !---------------------------------------------------------------------
  subroutine correctRanges(ranges)
  !---------------------------------------------------------------------
!*** change to a function??

    implicit none

    type(RangeT), pointer :: ranges(:)

    type(RangeT), allocatable :: localRanges(:)
    integer :: numRanges, i, locVec(1), start, finis, rangeNum

    !-------------------------------------------------------------------
    ! Start

    numRanges = size(ranges)
  if (numRanges .le. 0) return

    allocate(localRanges(numRanges))

    ! Swap any bounds which have the wrong internal order:
    do i = 1, numRanges
      if (ranges(i)%lower%type .ne. UNDEFINED&
      .and. ranges(i)%upper%type .ne. UNDEFINED&
      .and. ranges(i)%lower%value .gt. ranges(i)%upper%value) then
        call swapRangeBounds(ranges(i))
      end if
    end do

    ! Now sort them in order first of the lower bound, then the upper:
    call sortRanges(ranges)

    ! Find the last element of each array that has lo%type = UNDEFINED:
    if (any(ranges%lower%type .eq. UNDEFINED)) then
      locVec = maxloc(ranges%lower%value, (ranges%lower%type .eq. UNDEFINED))
      start = locVec(1)
    else
      start = 1
    end if

    ! Find the first element of each array that has hi%type = UNDEFINED:
    if (any(ranges%upper%type .eq. UNDEFINED)) then
      locVec = minloc(ranges%upper%value, (ranges%upper%type .eq. UNDEFINED))
      finis = locVec(1)
    else
      finis = 1
    end if

    ! Finally, transfer them to localRanges, amalgamating (ORing) overlapping
    ! members:
    rangeNum = 1
    localRanges(1) = ranges(start)
    do i = start + 1, finis
!      if (checkRangeOverlap(localRanges(rangeNum), ranges(i)) .eq. BOVERLAPSA)&
if (  checkRangeOverlap(localRanges(rangeNum), ranges(i)) .ne. BISTOOHIGH&
.and. checkRangeOverlap(localRanges(rangeNum), ranges(i)) .ne. BISTOOLOW)&
      then
!        localRanges(rangeNum) = orRangePair(localRanges(rangeNum), ranges(i))
call orRanges(ranges(i), localRanges(rangeNum))
      else
        rangeNum = rangeNum + 1
        localRanges(rangeNum) = ranges(i)
      end if
    end do

    deallocate(ranges)
    allocate(ranges(rangeNum))
    ranges = localRanges

  end subroutine correctRanges

  !---------------------------------------------------------------------
  subroutine sortRanges(ranges)
  !---------------------------------------------------------------------
!***** explanation of these next three?

  ! This is at present a private subroutine and only called by correctRanges().


    implicit none

    type(RangeT), intent(inout) :: ranges(:)

    integer :: i, numRanges, start, finis

    !-------------------------------------------------------------------
    ! Start

    numRanges = size(ranges)

    call sortBounds(ranges%lower, 'low')

    start = 1
    do while(start .le. numRanges)
      if (ranges(start)%lower%type .eq. UNDEFINED) then
        start = start + 1
    cycle
      end if

      finis = start
      do i = start + 1, numRanges
        if (ranges(i)%lower%value .ne. ranges(start)%lower%value&
        .or. ranges(i)%lower%type .ne. ranges(start)%lower%type) then
          finis = i - 1
        end if
      end do

      if (finis .gt. start) then
        call sortBounds(ranges(start:finis)%upper, 'hi')
      end if

      start = finis + 1
    end do

  end subroutine sortRanges

  !---------------------------------------------------------------------
  function integrateRangesFnScalar(yFn, xRange, dxArg, numSamplesArg)&
  result(approxIntegral)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in)           :: xRange
    real(single), intent(in), optional :: dxArg
    integer,      intent(in), optional :: numSamplesArg
    real(single)                       :: approxIntegral

    interface
      function yFn(x)
        use types

        real(single), intent(in) :: x
        real(single) :: yFn
      end function yFn
    end interface

    type(RangeT) :: xRanges(1)

    !-------------------------------------------------------------------
    ! Start

    xRanges(1) = copyRange(xRange)
    approxIntegral = integrateRangesFnVector(yFn, xRanges, dxArg, numSamplesArg)

  end function integrateRangesFnScalar

  !---------------------------------------------------------------------
  function integrateRangesFnVector(yFn, xRanges, dxArg, numSamplesArg)&
  result(approxIntegral)
  !---------------------------------------------------------------------
  ! This function returns a trapezoidal approximation to the integral of yFn(x) within the supplied ranges.
  !
  ! If neither dx nor numSamples is supplied, numSamples is assumed to be equal to 10000.
  !
  ! If both dx and numSamples are supplied, numSamples is ignored.
!**** warning in this case?
  !
  ! Note that, in the case that numSamples is supplied but dx not, the actual number of x-axis samples used may be slightly different to numSamples.
  !
  ! If the upper and/or lower integration bounds are UNDEFINED, the integration in these unbounded end ranges is carried out with a change of variable, from x to u = x/(1+x) for x>=0 and to u = x/(1-x) for x<0. In other words, for UNDEFINED lower bound, the integral calculated is
  !
  !	     a
  !	    /                     1
  !	I = | du yFn[u/(1+u)] ---------,  a < 0;
  !	    /                  (1+u)^2
  !	   -1
  !
  ! for UNDEFINED upper bound, the integral calculated is
  !
  !	     1
  !	    /                     1
  !	I = | du yFn[u/(1-u)] ---------,  b >= 0.
  !	    /                  (1-u)^2
  !	   b
  !
  ! The size of du is set = dx.
!*** discuss convergence and fact that yFn(-inf) and yFn(inf) are just set to 0.
  !
  ! NOTE! The ranges are required to be well-formed.
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in)           :: xRanges(:)
    real(single), intent(in), optional :: dxArg
    integer,      intent(in), optional :: numSamplesArg
    real(single)                       :: approxIntegral

    interface
      function yFn(x)
        use types

        real(single), intent(in) :: x
        real(single) :: yFn
      end function yFn
    end interface

    integer :: numRanges, maxConceivableNumSamples, numSamples, rangeNum
    real(single) ::&
      sizeLowestRange, sizeHighestRange, totalXRange, dx, du
    real(double) :: upperUVal, lowerUVal
    real(single), allocatable :: xVals(:), yVals(:)
    type(RangeT), allocatable :: newRanges(:)

    !-------------------------------------------------------------------
    ! Start

    approxIntegral = 0.0 ! default

    numRanges = size(xRanges)
  if (numRanges <= 0) return

    allocate(newRanges(numRanges))

    !...................................................................
    ! Settle the value of dx:

      select case(xRanges(1)%lower%type)
      case(UNDEFINED)
        select case(xRanges(numRanges)%upper%type)
        case(UNDEFINED)
          if (numRanges > 1) then
            if (xRanges(1)%upper%value < 0.0) then
              sizeLowestRange = 1.0 + xRanges(1)%upper%value&
                / (1.0 - xRanges(1)%upper%value)
            else
              sizeLowestRange = 1.0 + xRanges(1)%upper%value
            end if

            if (xRanges(numRanges)%lower%value >= 0.0) then
              sizeHighestRange = 1.0 - xRanges(numRanges)%lower%value&
                / (1.0 + xRanges(numRanges)%lower%value)
            else
              sizeHighestRange = 1.0 - xRanges(numRanges)%lower%value
            end if

            totalXRange = sizeLowestRange + sizeHighestRange&
              + sum(xRanges(2:numRanges-1)%upper%value&
              - xRanges(2:numRanges-1)%lower%value)

          else ! numRanges = 1, both bounds undefined.
            totalXRange = 2.0
          end if

        case(INCLUSIVE,EXCLUSIVE)
          if (numRanges > 1) then
            if (xRanges(1)%upper%value < 0.0) then
              sizeLowestRange = 1.0 + xRanges(1)%upper%value&
                / (1.0 - xRanges(1)%upper%value)
            else
              sizeLowestRange = 1.0 + xRanges(1)%upper%value
            end if

            totalXRange = sizeLowestRange&
              + sum(xRanges(2:numRanges)%upper%value&
              - xRanges(2:numRanges)%lower%value)

          else ! numRanges = 1, lower bound undefined but upper defined.
            if (xRanges(1)%upper%value < 0.0) then
              totalXRange = 1.0 + xRanges(1)%upper%value&
                / (1.0 - xRanges(1)%upper%value)
            else
              totalXRange = 1.0 + xRanges(1)%upper%value
            end if
          end if

        case default
call error('',"")
        end select

      case(INCLUSIVE,EXCLUSIVE)
        select case(xRanges(numRanges)%upper%type)
        case(UNDEFINED)
          if (numRanges > 1) then
            if (xRanges(numRanges)%lower%value >= 0.0) then
              sizeHighestRange = 1.0 - xRanges(numRanges)%lower%value&
                / (1.0 + xRanges(numRanges)%lower%value)
            else
              sizeHighestRange = 1.0 - xRanges(numRanges)%lower%value
            end if

            totalXRange = sizeHighestRange&
              + sum(xRanges(1:numRanges-1)%upper%value&
              - xRanges(1:numRanges-1)%lower%value)

          else ! numRanges = 1, lower bound defined, upper bound undefined.
            if (xRanges(numRanges)%lower%value >= 0.0) then
              totalXRange = 1.0 - xRanges(numRanges)%lower%value&
                / (1.0 + xRanges(numRanges)%lower%value)
            else
              totalXRange = 1.0 - xRanges(numRanges)%lower%value
            end if
          end if

        case(INCLUSIVE,EXCLUSIVE)
          totalXRange = sum(xRanges(1:numRanges)%upper%value&
            - xRanges(1:numRanges)%lower%value)
        case default
call error('',"")
        end select
      case default
call error('',"")
      end select
!*** all the above arithmetics need to be checked out!

    if (present(dxArg)) then
      dx = dxArg
    else if (present(numSamplesArg)) then
      dx = totalXRange / real(numSamplesArg)
    else ! neither is present.
      dx = totalXRange / 10000.0
    end if

    du = dx

    maxConceivableNumSamples = ceiling(totalXRange / dx) + 2 * numRanges
    allocate(xVals(maxConceivableNumSamples), yVals(maxConceivableNumSamples))

    !...................................................................
    ! Now generate the x and y values. Note that the ranges have to be
    ! adjusted so that those in the x-coordinate which have one or more
    ! UNDEFINED bounds are changed to u-coordinate.

    numSamples = 0

    if (xRanges(1)%lower%type == UNDEFINED) then
      newRanges(1)%lower%value = -1.0
      newRanges(1)%lower%type  = INCLUSIVE

      if (xRanges(1)%upper%type == UNDEFINED) then
        newRanges(1)%upper%value = 1.0
        newRanges(1)%upper%type  = INCLUSIVE

        call getXYUndefLower(0.0d0, du, yFn, numSamples, xVals, yVals)

        numSamples = numSamples + 1 ! want to overwrite the value at x=0,
        ! otherwise we will end up with two x's = 0.

        call getXYUndefUpper(0.0d0, du, yFn, numSamples, xVals, yVals)

        ! Stop here because, if xRanges is well-formed (as it is required
        ! to be), there can only be one range if both its bounds are UNDEFINED.

      else ! xRanges(1)%upper%type \= UNDEFINED
        if (xRanges(1)%upper%value < 0.0) then
          upperUVal = xRanges(1)%upper%value / (1.0 - xRanges(1)%upper%value)
          newRanges(1)%upper%value = upperUVal
          newRanges(1)%upper%type  = xRanges(1)%upper%type

          call getXYUndefLower(upperUVal, du, yFn, numSamples, xVals, yVals)

        else ! xRanges(1)%upper%value >= 0.0
          newRanges(1)%upper = copyBound(xRanges(1)%upper)

          call getXYUndefLower(0.0d0, du, yFn, numSamples, xVals, yVals)

          numSamples = numSamples + 1 ! want to overwrite the value at x=0,
          ! otherwise we will end up with two x's = 0.

          call getXYBothDefined(0.0d0, xRanges(1)%upper%value, dx, yFn&
            , numSamples, xVals, yVals)
        end if ! xRanges(1)%upper%value < 0.0

        ! Continue with the remaining ranges. If xRanges is well-formed
        ! (as it is required to be), all ranges up to the one before the
        ! last must have defined bounds. We'll do these first.

        do rangeNum = 2, numRanges - 1
          newRanges(rangeNum) = copyRange(xRanges(rangeNum))

          call getXYBothDefined(xRanges(rangeNum)%lower%value&
            , xRanges(rangeNum)%upper%value, dx, yFn, numSamples&
            , xVals, yVals)
        end do

        ! Now we deal with the last range. Its lower bound cannot be UNDEFINED.

        if (xRanges(numRanges)%upper%type == UNDEFINED) then
          if (xRanges(numRanges)%lower%value < 0.0) then
            newRanges(numRanges)%lower = copyBound(xRanges(numRanges)%lower)

            call getXYBothDefined(xRanges(numRanges)%lower%value, 0.0d0&
              , dx, yFn, numSamples, xVals, yVals)

            numSamples = numSamples + 1 ! want to overwrite the value at x=0,
            ! otherwise we will end up with two x's = 0.

            call getXYUndefUpper(0.0d0, du, yFn, numSamples, xVals, yVals)

          else ! xRanges(numRanges)%lower%value >= 0.0
            lowerUVal = xRanges(numRanges)%lower%value / (1.0&
              + xRanges(numRanges)%lower%value)

            newRanges(numRanges)%lower%value = lowerUVal
            newRanges(numRanges)%lower%type  = xRanges(numRanges)%lower%type

            call getXYUndefUpper(lowerUVal, du, yFn, numSamples, xVals, yVals)
          end if ! xRanges(numRanges)%lower%value < 0.0

        else ! xRanges(numRanges)%upper%type \= UNDEFINED
          newRanges(numRanges) = copyRange(xRanges(numRanges))

          call getXYBothDefined(xRanges(numRanges)%lower%value&
            , xRanges(numRanges)%upper%value, dx, yFn, numSamples&
            , xVals, yVals)
        end if ! xRanges(numRanges)%upper%type == UNDEFINED

      end if ! xRanges(1)%upper%type == UNDEFINED

    else ! xRanges(1)%lower%type \= UNDEFINED

      ! If xRanges is well-formed (as it is required to be), all ranges up
      ! to the one before the last must have defined bounds. We'll do these
      ! first.

      do rangeNum = 1, numRanges - 1
        newRanges(rangeNum) = copyRange(xRanges(rangeNum))

        call getXYBothDefined(xRanges(rangeNum)%lower%value&
          , xRanges(rangeNum)%upper%value, dx, yFn, numSamples&
          , xVals, yVals)
      end do

      ! Now we deal with the last range. Its lower bound cannot be UNDEFINED.

      if (xRanges(numRanges)%upper%type == UNDEFINED) then
        if (xRanges(numRanges)%lower%value < 0.0) then
          newRanges(numRanges)%lower = copyBound(xRanges(numRanges)%lower)

          call getXYBothDefined(xRanges(numRanges)%lower%value, 0.0d0, dx, yFn&
              , numSamples, xVals, yVals)

          numSamples = numSamples + 1 ! want to overwrite the value at x=0,
          ! otherwise we will end up with two x's = 0.

          call getXYUndefUpper(0.0d0, du, yFn, numSamples, xVals, yVals)

        else ! xRanges(numRanges)%lower%value >= 0.0
          lowerUVal = xRanges(numRanges)%lower%value / (1.0&
            + xRanges(numRanges)%lower%value)

          newRanges(numRanges)%lower%value = lowerUVal
          newRanges(numRanges)%lower%type  = xRanges(numRanges)%lower%type

          call getXYUndefUpper(lowerUVal, du, yFn, numSamples, xVals, yVals)
        end if ! xRanges(numRanges)%lower%value < 0.0

        newRanges(numRanges)%upper%value = 1.0
        newRanges(numRanges)%upper%type  = INCLUSIVE

      else ! xRanges(numRanges)%upper%type \= UNDEFINED
        newRanges(numRanges) = copyRange(xRanges(numRanges))

        call getXYBothDefined(xRanges(numRanges)%lower%value&
          , xRanges(numRanges)%upper%value, dx, yFn, numSamples&
          , xVals, yVals)
      end if ! xRanges(numRanges)%upper%type == UNDEFINED
    end if ! xRanges(1)%lower%type == UNDEFINED

    !...................................................................
    ! Finally, the function call which performs the integration:

    approxIntegral = integrateRangesVector(xVals(:numSamples)&
      , yVals(:numSamples), newRanges)

  end function integrateRangesFnVector

  !---------------------------------------------------------------------
  subroutine getXYUndefLower(upperUVal, du, yFn, numSamples, xVals, yVals)
  !---------------------------------------------------------------------
  ! This subroutine returns a sequence of samples both of the x coordinate and of a function yFn of that coordinate. This particular routine is designed for a situation in which the range in x is unbounded on the lower side. In this case the best approach is to make a change of variable u = x/(1-x) and return samples evenly spaced in u from u = -1 (=> x = -inf) to u = upperUVal. Note that the function yFn must be divided by (1+u)^2 in order for the sum of y values times du to approximate the unbounded integral of yFn dx.
  !
  ! The number of samples (stored in the variable numLocalSamples) is made equal to the u-range divided by du, rounded to the next highest integer, plus 1. This ensures that at least 2 samples are returned, 1 each for the two bounds of the range in u.
  !
  ! Note that numSamples must be initialized with some value by the calling routine (for the present routine, 0 is the only sensible value); also, xVals and yVals must be allocated enough space such that size(xVals) >= (initial numSamples)+numLocalSamples. The values stored in xVals(1:initial numSamples) and yVals(1:initial numSamples) are left untouched.
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)    :: upperUVal
    real(single), intent(in)    :: du
    integer,      intent(inout) :: numSamples
    real(single), intent(inout) :: xVals(:), yVals(size(xVals))

    interface
      function yFn(x)
        use types

        real(single), intent(in) :: x
        real(single) :: yFn
      end function yFn
    end interface

    integer :: numLocalSamples, xi
    real(single) :: u, x

    !-------------------------------------------------------------------
    ! Start

    numLocalSamples = 1 + ceiling((1.0 + upperUVal) / du)
if (size(xVals) < numSamples + numLocalSamples) call error('',"")

    numSamples = numSamples + 1
    xVals(numSamples) = -1.0
    yVals(numSamples) = 0.0 ! can't calculate yFn at x=-infinity!

    do xi = 1, numLocalSamples - 1
      u = fracVal(-1.0d0, upperUVal, xi / real(numLocalSamples-1, double))
      ! fracVal is in ssclib/ssc_misc.

      x = u / (1 + u)

      numSamples = numSamples + 1
      xVals(numSamples) = u
      yVals(numSamples) = yFn(x) / (1 + u) / (1 + u)
    end do

  end subroutine getXYUndefLower

  !---------------------------------------------------------------------
  subroutine getXYBothDefined(lowerXVal, upperXVal, dx, yFn, numSamples&
  , xVals, yVals)
  !---------------------------------------------------------------------
  ! This subroutine returns a sequence of evenly-spaced samples of x in the range lowerXVal to upperXVal and the corresponding sequence of values of a function yFn(x).
  !
  ! The number of samples (stored in the variable numLocalSamples) is made equal to the range divided by dx, rounded to the next highest integer, plus 1. This ensures that at least 2 samples are returned, 1 each for the two bounds of the range.
  !
  ! Note that numSamples must be initialized with some value by the calling routine; also, xVals and yVals must be allocated enough space such that size(xVals) >= (initial numSamples)+numLocalSamples. The values stored in xVals(1:initial numSamples) and yVals(1:initial numSamples) are left untouched.
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)    :: lowerXVal, upperXVal
    real(single), intent(in)    :: dx
    integer,      intent(inout) :: numSamples
    real(single), intent(inout) :: xVals(:), yVals(size(xVals))

    interface
      function yFn(x)
        use types

        real(single), intent(in) :: x
        real(single) :: yFn
      end function yFn
    end interface

    integer :: numLocalSamples, xi
    real(single) :: x

    !-------------------------------------------------------------------
    ! Start

    numLocalSamples = 1 + ceiling((upperXVal - lowerXVal) / dx)
if (size(xVals) < numSamples + numLocalSamples) call error('',"")

    do xi = 0, numLocalSamples - 1
      x = fracVal(lowerXVal, upperXVal, xi / real(numLocalSamples-1, double))
      ! fracVal is in ssclib/ssc_misc.

      numSamples = numSamples + 1
      xVals(numSamples) = x
      yVals(numSamples) = yFn(x)
    end do

  end subroutine getXYBothDefined

  !---------------------------------------------------------------------
  subroutine getXYUndefUpper(lowerUVal, du, yFn, numSamples, xVals, yVals)
  !---------------------------------------------------------------------
  ! This subroutine returns a sequence of samples both of the x coordinate and of a function yFn of that coordinate. This particular routine is designed for a situation in which the range in x is unbounded on the upper side. In this case the best approach is to make a change of variable u = x/(1+x) and return samples evenly spaced in u from u = lowerUVal to u = 1 (=> x = inf). Note that the function yFn must be divided by (1-u)^2 in order for the sum of y values times du to approximate the unbounded integral of yFn dx.
  !
  ! The number of samples (stored in the variable numLocalSamples) is made equal to the u-range divided by du, rounded to the next highest integer, plus 1. This ensures that at least 2 samples are returned, 1 each for the two bounds of the range in u.
  !
  ! Note that numSamples must be initialized with some value by the calling routine; also, xVals and yVals must be allocated enough space such that size(xVals) >= (initial numSamples)+numLocalSamples. The values stored in xVals(1:initial numSamples) and yVals(1:initial numSamples) are left untouched.
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)    :: lowerUVal
    real(single), intent(in)    :: du
    integer,      intent(inout) :: numSamples
    real(single), intent(inout) :: xVals(:), yVals(size(xVals))

    interface
      function yFn(x)
        use types

        real(single), intent(in) :: x
        real(single) :: yFn
      end function yFn
    end interface

    integer :: numLocalSamples, xi
    real(single) :: u, x

    !-------------------------------------------------------------------
    ! Start

    numLocalSamples = 1 + ceiling((1.0 - lowerUVal) / du)
if (size(xVals) < numSamples + numLocalSamples) call error('',"")

    do xi = 0, numLocalSamples - 2
      u = fracVal(lowerUVal, 1.0d0, xi / real(numLocalSamples-1, double))
      ! fracVal is in ssclib/ssc_misc.

      x = u / (1 - u)

      numSamples = numSamples + 1
      xVals(numSamples) = u
      yVals(numSamples) = yFn(x) / (1 - u) / (1 - u)
    end do

    numSamples = numSamples + 1
    xVals(numSamples) = 1.0
    yVals(numSamples) = 0.0 ! can't calculate yFn at x=infinity!

  end subroutine getXYUndefUpper

  !---------------------------------------------------------------------
  function integrateRangesSingleScalar(x, y, xLo, xHi) result(approxIntegral)
  !---------------------------------------------------------------------
  ! This function takes a set of samples of y as a function of x, plus a
  ! range of integration, and returns a trapezoidal approximation to the
  ! integral of y(x) dx within this range.
  !
  ! NOTE! The values in x() should be in increasing order.

    implicit none

    real(single), intent(in) :: x(:), y(size(x)), xLo, xHi
    real(single) :: approxIntegral

    type(RangeT) :: xRanges(1)

    !-------------------------------------------------------------------
    ! Start

    xRanges(1)%lower%type  = INCLUSIVE
    xRanges(1)%upper%type  = INCLUSIVE
    xRanges(1)%lower%value = xLo
    xRanges(1)%upper%value = xHi
    approxIntegral = integrateRangesVector(x, y, xRanges)

  end function integrateRangesSingleScalar

  !---------------------------------------------------------------------
  function integrateRangesScalar(x, y, xRange) result(approxIntegral)
  !---------------------------------------------------------------------
  ! This function takes a set of samples of y as a function of x, plus a
  ! range of integration, and returns a trapezoidal approximation to the
  ! integral of y(x) dx within this range.
  !
  ! NOTE! The values in x() should be in increasing order.

    implicit none

    real(single), intent(in) :: x(:), y(size(x))
    type(RangeT), intent(in) :: xRange
    real(single) :: approxIntegral

    type(RangeT) :: xRanges(1)

    !-------------------------------------------------------------------
    ! Start

    xRanges(1) = copyRange(xRange)
    approxIntegral = integrateRangesVector(x, y, xRanges)

  end function integrateRangesScalar

  !---------------------------------------------------------------------
  function integrateRangesVector(x, y, xRanges) result(approxIntegral)
  !---------------------------------------------------------------------
  ! This function takes a set of samples of y as a function of x, plus a set
  ! of ranges of integration, and returns a trapezoidal approximation to the
  ! integral of y(x) dx within these ranges. 
  !
  ! NOTE! (i) The values in x() should be in increasing order. (ii) The
  ! ranges may occur in any order, but are otherwise assumed to be 'well-
  ! -formed'. (iii) If any upper range bound is undefined, the upper x
  ! value is used instead for that range bound; likewise for undefined
  ! lower bounds.

!*** should change it so that the ranges are required to be well-formed. This would make the routine less general but makes the accepted properties of ranges simpler.

    implicit none

    real(single), intent(in) :: x(:), y(size(x))
    type(RangeT), intent(in) :: xRanges(:)
    real(single) :: approxIntegral

    integer(int16) :: numSamples, numRanges, rangeNum, maxI, minI
    real(single) ::&
      minX, maxX, term, rangeMaxX, rangeMinX, yAtRangeMaxX, yAtRangeMinX, summ

    !-------------------------------------------------------------------
    ! Start

    approxIntegral = 0.0 ! default
  if (size(x) <= 1 .or. size(xRanges) <= 0) return

    numSamples = size(x)
    minX = minval(x)
    maxX = maxval(x)

    numRanges = size(xRanges)
    do rangeNum = 1, numRanges
      select case(xRanges(rangeNum)%lower%type)
      case(UNDEFINED)
        select case(xRanges(rangeNum)%upper%type)
        case(UNDEFINED)
          summ = trapezoidSum(x, y)

        case(INCLUSIVE,EXCLUSIVE)
          rangeMaxX = xRanges(rangeNum)%upper%value
          if (rangeMaxX <= minX) then
            summ = 0.0

          else
            maxI = largestTrueIndex(logical(x <= rangeMaxX, bool))
            summ = trapezoidSum(x(:maxI), y(:maxI))

            if (maxI < numSamples) then ! add the fraction at the end:
              yAtRangeMaxX = fracVal(y(maxI), y(maxI + 1)&
                , ((rangeMaxX - x(maxI)) / (x(maxI + 1) - x(maxI))))
              term = 0.5 * (y(maxI) + yAtRangeMaxX)&
                * (rangeMaxX - x(maxI))
              summ = summ + term
            end if
          end if

        case default
call error('badUpperRangeType','')
        end select

      case(INCLUSIVE,EXCLUSIVE)
        select case(xRanges(rangeNum)%upper%type)
        case(UNDEFINED)
          rangeMinX = xRanges(rangeNum)%lower%value
          if (rangeMinX >= maxX) then
            summ = 0.0

          else
            minI = smallestTrueIndex(logical(x >= rangeMinX, bool))
            summ = trapezoidSum(x(minI:), y(minI:))

            if (minI > 1) then ! add the fraction at the end:
              yAtRangeMinX = fracVal(y(minI - 1), y(minI)&
                , ((x(minI) - rangeMinX) / (x(minI) - x(minI - 1))))
              term = 0.5 * (yAtRangeMinX + y(minI))&
                * (x(minI) - rangeMinX)
              summ = summ + term
            end if
          end if

        case(INCLUSIVE,EXCLUSIVE)
          rangeMinX = xRanges(rangeNum)%lower%value
          rangeMaxX = xRanges(rangeNum)%upper%value
          if (rangeMaxX <= minX .or. rangeMinX >= maxX) then
            summ = 0.0

          else
            minI = smallestTrueIndex(logical(x >= rangeMinX, bool))
            maxI = largestTrueIndex(logical(x <= rangeMaxX, bool))
            summ = trapezoidSum(x(minI:maxI), y(minI:maxI))
            if (minI > 1) then ! add the fraction at the end:
              yAtRangeMinX = fracVal(y(minI - 1), y(minI)&
                , ((x(minI) - rangeMinX) / (x(minI) - x(minI - 1))))
              term = 0.5 * (yAtRangeMinX + y(minI))&
                * (x(minI) - rangeMinX)
              summ = summ + term
            end if
            if (maxI < numSamples) then ! add the fraction at the end:
              yAtRangeMaxX = fracVal(y(maxI), y(maxI + 1)&
                , ((rangeMaxX - x(maxI)) / (x(maxI + 1) - x(maxI))))
              term = 0.5 * (y(maxI) + yAtRangeMaxX)&
                * (rangeMaxX - x(maxI))
              summ = summ + term
            end if
          end if

        case default
call error('badUpperRangeType','')
        end select

      case default
call error('badLowerRangeType','')
      end select

      approxIntegral = approxIntegral + summ
    end do

  end function integrateRangesVector

  !---------------------------------------------------------------------
  subroutine sortBounds(bounds, style)
  !---------------------------------------------------------------------

    implicit none

    type(BoundT), intent(inout) :: bounds(:)
    character(*), intent(in) :: style

    integer :: i, numBounds, locVec(1), start, finis
    integer, allocatable :: index(:)
    type(BoundT), allocatable :: localBounds(:)

    !-------------------------------------------------------------------
    ! Start

    numBounds = size(bounds)
    allocate(index(numBounds), localBounds(numBounds))

    if (style .eq. 'low') then
      ! First, move all the type = UNDEFINED down to the bottom:

      do i = 1, numBounds - 1
      if (bounds(i)%type .eq. UNDEFINED) cycle
      if (.not. any(bounds(i+1:)%type .eq. UNDEFINED)) exit

        locVec = minloc(bounds(i+1:)%value&
        , (bounds(i+1:)%type .eq. UNDEFINED))

        call swapBounds(bounds(i), bounds(locVec(1)))
      end do

      if (any(bounds%type .ne. UNDEFINED)) then ! sort them:
        do i = 1, numBounds
          if (bounds(i)%type .ne. UNDEFINED) then
            start = i
        exit
          end if
        end do

        call quickSortStatic(bounds(start:)%value, index(start:))
        do i = 1, start - 1
          localBounds(i) = bounds(i)
        end do
        do i = start, numBounds
          localBounds(i) = bounds(index(i))
        end do
      end if

    else ! style .eq. 'hi'
      ! First, move all the type = UNDEFINED up to the top:

      do i = numBounds, 2, -1
      if (bounds(i)%type .eq. UNDEFINED) cycle
      if (.not. any(bounds(:i-1)%type .eq. UNDEFINED)) exit

        locVec = minloc(bounds(:i-1)%value&
        , (bounds(:i-1)%type .eq. UNDEFINED))

        call swapBounds(bounds(i), bounds(locVec(1)))
      end do

      if (any(bounds%type .ne. UNDEFINED)) then ! sort them:
        do i = numBounds, 1, -1
          if (bounds(i)%type .ne. UNDEFINED) then
            finis = i
        exit
          end if
        end do

        call quickSortStatic(bounds(:finis)%value, index(:finis))
        do i = 1, finis
          localBounds(i) = bounds(index(i))
        end do
        do i = finis + 1, numBounds
          localBounds(i) = bounds(i)
        end do
      end if
    end if

  end subroutine sortBounds

  !---------------------------------------------------------------------
  subroutine swapBounds(boundA, boundB)
  !---------------------------------------------------------------------

    implicit none

    type(BoundT), intent(inout) :: boundA, boundB

    type(BoundT) :: tempBound

    !-------------------------------------------------------------------
    ! Start

    tempBound = boundA
    boundA = boundB
    boundB = tempBound

  end subroutine swapBounds

  !---------------------------------------------------------------------
  subroutine dumpRangesScalar(range, numIndentSpacesArg)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in)           :: range
    integer,      intent(in), optional :: numIndentSpacesArg

    integer :: numIndentSpaces = 0

    !-------------------------------------------------------------------
    ! Start

    if (present(numIndentSpacesArg)) numIndentSpaces = numIndentSpacesArg

    print *, repeat(' ', numIndentSpaces)//'Lower bound:'
    call dumpBoundsScalar(range%lower, numIndentSpaces+2)
    print *, repeat(' ', numIndentSpaces)//'Upper bound:'
    call dumpBoundsScalar(range%upper, numIndentSpaces+2)

  end subroutine dumpRangesScalar

  !---------------------------------------------------------------------
  subroutine dumpRangesScalar_20040506(range, indentStrArg)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in) :: range
    character(*), intent(in), optional :: indentStrArg

    integer :: numIndentSpaces = 0

    !-------------------------------------------------------------------
    ! Start

    if (present(indentStrArg)) numIndentSpaces = len(indentStrArg)

    print *, repeat(' ', numIndentSpaces)//'Lower bound:'
    call dumpBoundsScalar_20040506(range%lower,repeat(' ',numIndentSpaces)//'  ')
    print *, repeat(' ', numIndentSpaces)//'Upper bound:'
    call dumpBoundsScalar_20040506(range%upper, repeat(' ', numIndentSpaces)//'  ')

  end subroutine dumpRangesScalar_20040506

  !---------------------------------------------------------------------
  subroutine dumpRangesVector(ranges, numIndentSpacesArg)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in)           :: ranges(:)
    integer,      intent(in), optional :: numIndentSpacesArg

    integer(int32) :: rangeNum
    integer :: numIndentSpaces = 0

    !-------------------------------------------------------------------
    ! Start

    if (present(numIndentSpacesArg)) numIndentSpaces = numIndentSpacesArg

    do rangeNum = 1, size(ranges)
      print *, repeat(' ', numIndentSpaces)//'Range number =', rangeNum
      call dumpRangesScalar(ranges(rangeNum), numIndentSpaces+2)
    end do

  end subroutine dumpRangesVector

  !---------------------------------------------------------------------
  subroutine dumpRangesVector_20040506(ranges, indentStrArg)
  !---------------------------------------------------------------------

    implicit none

    type(RangeT), intent(in) :: ranges(:)
    character(*), intent(in), optional :: indentStrArg

    integer(int32) :: rangeNum
    integer :: numIndentSpaces = 0

    !-------------------------------------------------------------------
    ! Start

    if (present(indentStrArg)) numIndentSpaces = len(indentStrArg)

    do rangeNum = 1, size(ranges)
      print *, repeat(' ', numIndentSpaces)//'Range number =', rangeNum
      call dumpRangesScalar_20040506(ranges(rangeNum)&
        , repeat(' ', numIndentSpaces)//'  ')
    end do

  end subroutine dumpRangesVector_20040506

  !---------------------------------------------------------------------
  subroutine andRangesScalScalVec(rangeA, rangeB, andedRanges, doChecks)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate andedRanges after use.

! Not transferred.

    implicit none

    type(RangeT),  intent(in)           :: rangeA, rangeB
    type(RangeT),  pointer              :: andedRanges(:)
    logical(bool), intent(in), optional :: doChecks

    logical(bool) :: localDoChecks

    !-------------------------------------------------------------------
    ! Start

    if (present(doChecks)) then
      localDoChecks = doChecks
    else
      localDoChecks = .false.
    end if

    call andRangesVecVecVec((/rangeA/), (/rangeB/), andedRanges, localDoChecks)

  end subroutine andRangesScalScalVec

  !---------------------------------------------------------------------
  subroutine andRangesScalVecVec(rangeA, rangesB, andedRanges, doChecks)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate andedRanges after use.

! Not transferred.

    implicit none

    type(RangeT),  intent(in)           :: rangeA, rangesB(:)
    type(RangeT),  pointer              :: andedRanges(:)
    logical(bool), intent(in), optional :: doChecks

    logical(bool) :: localDoChecks

    !-------------------------------------------------------------------
    ! Start

    if (present(doChecks)) then
      localDoChecks = doChecks
    else
      localDoChecks = .false.
    end if

    call andRangesVecVecVec((/rangeA/), rangesB, andedRanges, localDoChecks)

  end subroutine andRangesScalVecVec

  !---------------------------------------------------------------------
  subroutine andRangesVecScalVec(rangesA, rangeB, andedRanges, doChecks)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate andedRanges after use.

! Not transferred.

    implicit none

    type(RangeT),  intent(in)           :: rangesA(:), rangeB
    type(RangeT),  pointer              :: andedRanges(:)
    logical(bool), intent(in), optional :: doChecks

    logical(bool) :: localDoChecks

    !-------------------------------------------------------------------
    ! Start

    if (present(doChecks)) then
      localDoChecks = doChecks
    else
      localDoChecks = .false.
    end if

    call andRangesVecVecVec(rangesA, (/rangeB/), andedRanges, localDoChecks)

  end subroutine andRangesVecScalVec

  !---------------------------------------------------------------------
  subroutine andRangesVecVecVec(rangesA, rangesB, andedRanges, doChecks)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate andedRanges after use.

! Transferred to ranges_aux:andRangesVec

    implicit none

    type(RangeT),  intent(in)           :: rangesA(:), rangesB(:)
    type(RangeT),  pointer              :: andedRanges(:)
    logical(bool), intent(in), optional :: doChecks

    type(RangeT), allocatable :: tempRanges(:)
    type(RangeT), pointer ::&
      localRangesA(:), localRangesB(:)
    integer :: numRangesA, numRangesB, andedRangeNum, startB, i, j

    !-------------------------------------------------------------------
    ! Start

    allocate(tempRanges(size(rangesA) + size(rangesB)))
    allocate(localRangesA(size(rangesA)))
    localRangesA = rangesA
    allocate(localRangesB(size(rangesB)))
    localRangesB = rangesB

    if (present(doChecks)) then
      if (doChecks) then
        ! Enforce correct ordering and overlap on ranges:
        call correctRanges(localRangesA)
        call correctRanges(localRangesB)
      end if
    end if

    numRangesA = size(localRangesA)
    numRangesB = size(localRangesB)

    if (numRangesA .le. 0 .or. numRangesB .le. 0) then
      deallocate(localRangesA, localRangesB)
      allocate(andedRanges(0))
  return
    end if

    ! Check to see if either range includes (or both ranges include) the
    ! whole of the real number axis:

    if (any(localRangesA%lower%type .eq. UNDEFINED&
    .and. localRangesA%upper%type .eq. UNDEFINED)) then
      if (any(localRangesB%lower%type .eq. UNDEFINED&
      .and. localRangesB%upper%type .eq. UNDEFINED)) then
        deallocate(localRangesA, localRangesB)
        allocate(andedRanges(1))
        andedRanges(1)%lower%type = UNDEFINED
        andedRanges(1)%upper%type = UNDEFINED
  return

      else
        allocate(andedRanges(numRangesB))
        andedRanges = localRangesB
        deallocate(localRangesA, localRangesB)
  return
      end if

    else if (any(rangesB%lower%type .eq. UNDEFINED&
    .and. rangesB%upper%type .eq. UNDEFINED)) then
      allocate(andedRanges(numRangesA))
      andedRanges = localRangesA
      deallocate(localRangesA, localRangesB)
  return
    end if

    ! Now compute the AND. This is done as follows. We go through one of the
    ! two sequences of intervals (it doesn't matter which) one by one. For
    ! each interval, we go through a subset of the other sequence and check
    ! for overlap. Action is then taken depending on the result of this check.

    andedRangeNum = 0
    startB = 1
!*** work out startA and maybe finisA also.
    do i = 1, numRangesA
      BLOOP: do j = startB, numRangesB
        select case(checkRangeOverlap(localRangesA(i), localRangesB(j)))
        case(BISTOOHIGH) ! The B range (and all the following ones) are too
        ! high in value to overlap this A range. Quit looking at Bs and
        ! try the next A.

          startB = max(1, j - 1) ! because this must be the first B tested
          ! which has gone over. The previous one must therefore either
          ! have overlapped the A range or been below it (or, there were
          ! no previous Bs, this is the very first). In either case, the
          ! previous B (if there was one) is a convenient start point next
          ! time around.
      exit BLOOP

        case(BISTOOLOW) ! The B range is too low in value to overlap this A
        ! range. Try the next B.
      cycle BLOOP
        end select

        ! Otherwise the two ranges must overlap. Let us deal
        ! with that then.

        andedRangeNum = andedRangeNum + 1
!        tempRanges(andedRangeNum) = andRangePair(localRangesA(i), localRangesB(j))
call andRangesScalScalScal(localRangesA(i), localRangesB(j)&
  , tempRanges(andedRangeNum))
      end do BLOOP
    end do

    deallocate(localRangesA, localRangesB)

    allocate(andedRanges(andedRangeNum))
    andedRanges = tempRanges(:andedRangeNum)

  end subroutine andRangesVecVecVec

  !---------------------------------------------------------------------
  subroutine andRangesScalScalScal(rangeA, rangeB, andedRange)
  !---------------------------------------------------------------------

! Transferred to ranges_aux:andRangesScal

    implicit none

    type(RangeT), intent(in)  :: rangeA, rangeB
    type(RangeT), intent(out) :: andedRange

    !-------------------------------------------------------------------
    ! Start

!    if (checkRangeOverlap(rangeA, rangeB) .ne. BOVERLAPSA) then
if ( checkRangeOverlap(rangeA, rangeB) .eq. BISTOOHIGH&
.or. checkRangeOverlap(rangeA, rangeB) .eq. BISTOOLOW) then
call error('noRangeOverlap',"")
    end if

    ! First calculate the value and type of the start of the ANDed range:
    if (rangeA%lower%type .eq. UNDEFINED) then
      if (rangeB%lower%type .eq. UNDEFINED) then
        andedRange%lower%type = UNDEFINED
        ! Value doesn't matter.

      else
        andedRange%lower  = rangeB%lower
      end if

    else ! rangeA%lower%type .ne. UNDEFINED
      if (rangeB%lower%type .eq. UNDEFINED) then
        andedRange%lower  = rangeA%lower 
      else ! neither rangeA%lower%type nor rangeB%lower%type is UNDEFINED.
        if      (rangeA%lower%value .gt. rangeB%lower%value) then
          andedRange%lower  = rangeA%lower 
        else if (rangeB%lower%value .gt. rangeA%lower%value) then
          andedRange%lower  = rangeB%lower 
        else ! rangeA%lower%value = rangeB%lower%value.
          andedRange%lower%value = rangeA%lower%value ! ist egal.

          if (rangeA%lower%type .eq. EXCLUSIVE&
          .or. rangeB%lower%type .eq. EXCLUSIVE) then ! we must exclude the
          ! boundary point.
            andedRange%lower%type = EXCLUSIVE
          else
            andedRange%lower%type = INCLUSIVE
          end if
        end if
      end if
    end if

    ! Now calculate the value and type of the end of the ANDed range:
    if (rangeA%upper%type .eq. UNDEFINED) then
      if (rangeB%upper%type .eq. UNDEFINED) then
        andedRange%upper%type = UNDEFINED
        ! Value doesn't matter.

      else
        andedRange%upper  = rangeB%upper 
      end if

    else
      if (rangeB%upper%type .eq. UNDEFINED) then
        andedRange%upper  = rangeA%upper 
      else ! neither rangeA%upper%type nor rangeB%upper%type is UNDEFINED.
        if      (rangeA%upper%value .lt. rangeB%upper%value) then
          andedRange%upper  = rangeA%upper 
        else if (rangeB%upper%value .lt. rangeA%upper%value) then
          andedRange%upper  = rangeB%upper 
        else ! rangeA%upper%value = rangeB%upper%value.
          andedRange%upper%value = rangeA%upper%value ! ist egal.

          if (rangeA%upper%type .eq. EXCLUSIVE&
          .or. rangeB%upper%type .eq. EXCLUSIVE) then ! we must exclude the
          ! boundary point.
            andedRange%upper%type = EXCLUSIVE
          else
            andedRange%upper%type = INCLUSIVE
          end if
        end if
      end if
    end if

  end subroutine andRangesScalScalScal

  !---------------------------------------------------------------------
  subroutine andRangesScalScal(inRange, andedRange)
  !---------------------------------------------------------------------

! Transferred to ranges_aux:accumAndRangesScalScal

    implicit none

    type(RangeT), intent(in)    :: inRange
    type(RangeT), intent(inout) :: andedRange

    type(RangeT) :: tempRange

    !-------------------------------------------------------------------
    ! Start

    call andRangesScalScalScal(inRange, andedRange, tempRange)
    andedRange = tempRange

  end subroutine andRangesScalScal

  !---------------------------------------------------------------------
  subroutine andRangesScalVec(inRange, andedRanges, doChecks)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate andedRanges after use.

! Transferred to ranges_aux:accumAndRangesScalVec

    implicit none

    type(RangeT),  intent(in)           :: inRange
    type(RangeT),  pointer              :: andedRanges(:)
    logical(bool), intent(in), optional :: doChecks

    type(RangeT), pointer :: tempRanges(:)
    logical(bool) :: localDoChecks

    !-------------------------------------------------------------------
    ! Start

    if (present(doChecks)) then
      localDoChecks = doChecks
    else
      localDoChecks = .false.
    end if

    call andRangesVecVecVec((/inRange/), andedRanges, tempRanges, localDoChecks)
    deallocate(andedRanges)
!    allocate(andedRanges(size(tempRanges)))
!    andedRanges = tempRanges
!    dealocate(tempRanges)
andedRanges => tempRanges

  end subroutine andRangesScalVec

  !---------------------------------------------------------------------
  subroutine andRangesVecVec(inRanges, andedRanges, doChecks)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate andedRanges after use.

! Transferred to ranges_aux:accumAndRangesVecVec

    implicit none

    type(RangeT),  intent(in)           :: inRanges(:)
    type(RangeT),  pointer              :: andedRanges(:)
    logical(bool), intent(in), optional :: doChecks

    type(RangeT), pointer :: tempRanges(:)
    logical(bool) :: localDoChecks

    !-------------------------------------------------------------------
    ! Start

    if (present(doChecks)) then
      localDoChecks = doChecks
    else
      localDoChecks = .false.
    end if

    call andRangesVecVecVec(inRanges, andedRanges, tempRanges, localDoChecks)
    deallocate(andedRanges)
!    allocate(andedRanges(size(tempRanges)))
!    andedRanges = tempRanges
!    dealocate(tempRanges)
andedRanges => tempRanges

  end subroutine andRangesVecVec





  !---------------------------------------------------------------------
  subroutine orRangesScalScalVec(rangeA, rangeB, oredRanges)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate oredRanges after use.

! Transferred to ranges_aux:orRangesScal

    implicit none

    type(RangeT),  intent(in) :: rangeA, rangeB
    type(RangeT),  pointer    :: oredRanges(:)

    !-------------------------------------------------------------------
    ! Start

    call orRangesVecVecVec((/rangeA/), (/rangeB/), oredRanges)

  end subroutine orRangesScalScalVec

  !---------------------------------------------------------------------
  subroutine orRangesScalVecVec(rangeA, rangesB, oredRanges)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate oredRanges after use.

! Not transferred.

    implicit none

    type(RangeT),  intent(in) :: rangeA, rangesB(:)
    type(RangeT),  pointer    :: oredRanges(:)

    !-------------------------------------------------------------------
    ! Start

    call orRangesVecVecVec((/rangeA/), rangesB, oredRanges)

  end subroutine orRangesScalVecVec

  !---------------------------------------------------------------------
  subroutine orRangesVecScalVec(rangesA, rangeB, oredRanges)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate oredRanges after use.

! Not transferred.

    implicit none

    type(RangeT),  intent(in) :: rangesA(:), rangeB
    type(RangeT),  pointer    :: oredRanges(:)

    !-------------------------------------------------------------------
    ! Start

    call orRangesVecVecVec(rangesA, (/rangeB/), oredRanges)

  end subroutine orRangesVecScalVec

  !---------------------------------------------------------------------
  subroutine orRangesVecVecVec(rangesA, rangesB, oredRanges)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate oredRanges after use.

! Transferred to ranges_aux:orRangesVec

    implicit none

    type(RangeT),  intent(in) :: rangesA(:), rangesB(:)
    type(RangeT),  pointer    :: oredRanges(:)

    !-------------------------------------------------------------------
    ! Start

    ! Simplest way to do this is to simply join the two range chains
    ! together and send them to correctRanges.

    allocate(oredRanges(size(rangesA) + size(rangesB)))
    oredRanges(:size(rangesA))&
      = rangesA
    oredRanges(size(rangesA)+1:size(rangesA)+size(rangesB))&
      = rangesB

    call correctRanges(oredRanges)

  end subroutine orRangesVecVecVec

  !---------------------------------------------------------------------
  subroutine orRangesScalScalScal(rangeA, rangeB, oredRange)
  !---------------------------------------------------------------------

! Not transferred.

    implicit none

    type(RangeT), intent(in)  :: rangeA, rangeB
    type(RangeT), intent(out) :: oredRange

    !-------------------------------------------------------------------
    ! Start

!    if (checkRangeOverlap(rangeA, rangeB) .ne. BOVERLAPSA) then
if ( checkRangeOverlap(rangeA, rangeB) .eq. BISTOOHIGH&
.or. checkRangeOverlap(rangeA, rangeB) .eq. BISTOOLOW) then
call error('noRangeOverlap',"")
    end if

    ! First calculate the value and type of the start of the ORed range:
    if (rangeA%lower%type .eq. UNDEFINED .or. rangeB%lower%type .eq. UNDEFINED)&
    then
      oredRange%lower%type = UNDEFINED
      ! Value doesn't matter.

    else ! neither rangeA%lower%type nor rangeB%lower%type is UNDEFINED.
      if      (rangeA%lower%value .lt. rangeB%lower%value) then
        oredRange%lower  = rangeA%lower
      else if (rangeB%lower%value .lt. rangeA%lower%value) then
        oredRange%lower  = rangeB%lower 
      else ! rangeA%lower%value = rangeB%lower%value.
        oredRange%lower%value = rangeA%lower%value ! ist egal.

        if (rangeA%lower%type .eq. INCLUSIVE&
        .or. rangeB%lower%type .eq. INCLUSIVE) then ! we must include the
        ! boundary point.
          oredRange%lower%type = INCLUSIVE
        else
          oredRange%lower%type = INCLUSIVE
        end if
      end if
    end if

    ! Now calculate the value and type of the end of the ORed range:
    if (rangeA%upper%type .eq. UNDEFINED .or. rangeB%upper%type .eq. UNDEFINED)&
    then
      oredRange%upper%type = UNDEFINED
      ! Value doesn't matter.

    else ! neither rangeA%upper%type nor rangeB%upper%type is UNDEFINED.
      if      (rangeA%upper%value .gt. rangeB%upper%value) then
        oredRange%upper  = rangeA%upper
      else if (rangeB%upper%value .gt. rangeA%upper%value) then
        oredRange%upper  = rangeB%upper 
      else ! rangeA%upper%value = rangeB%upper%value.
        oredRange%upper%value = rangeA%upper%value ! ist egal.

        if (rangeA%upper%type .eq. INCLUSIVE&
        .or. rangeB%upper%type .eq. INCLUSIVE) then ! we must include the
        ! boundary point.
          oredRange%upper%type = INCLUSIVE
        else
          oredRange%upper%type = EXCLUSIVE
        end if
      end if
    end if

  end subroutine orRangesScalScalScal

  !---------------------------------------------------------------------
  subroutine orRangesScalVec(inRange, oredRanges)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate oredRanges after use.

! Transferred to ranges_aux:accumOrRangesScal

    implicit none

    type(RangeT),  intent(in) :: inRange
    type(RangeT),  pointer    :: oredRanges(:)

    type(RangeT), pointer :: tempRanges(:)

    !-------------------------------------------------------------------
    ! Start

    call orRangesVecVecVec((/inRange/), oredRanges, tempRanges)
    deallocate(oredRanges)
!    allocate(oredRanges(size(tempRanges)))
!    oredRanges = tempRanges
!    dealocate(tempRanges)
oredRanges => tempRanges

  end subroutine orRangesScalVec

  !---------------------------------------------------------------------
  subroutine orRangesVecVec(inRanges, oredRanges)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate oredRanges after use.

! Transferred to ranges_aux:accumOrRangesVec

    implicit none

    type(RangeT),  intent(in) :: inRanges(:)
    type(RangeT),  pointer    :: oredRanges(:)

    type(RangeT), pointer :: tempRanges(:)

    !-------------------------------------------------------------------
    ! Start

    call orRangesVecVecVec(inRanges, oredRanges, tempRanges)
    deallocate(oredRanges)
!    allocate(oredRanges(size(tempRanges)))
!    oredRanges = tempRanges
!    dealocate(tempRanges)
oredRanges => tempRanges

  end subroutine orRangesVecVec

  !---------------------------------------------------------------------
  subroutine orRangesScalScal(inRange, oredRange)
  !---------------------------------------------------------------------

! Not transferred.

    implicit none

    type(RangeT), intent(in)    :: inRange
    type(RangeT), intent(inout) :: oredRange

    type(RangeT) :: tempRange

    !-------------------------------------------------------------------
    ! Start

    call orRangesScalScalScal(inRange, oredRange, tempRange)
    oredRange = tempRange

  end subroutine orRangesScalScal

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module dss_ranges_aux
