!
! Module: math_utils
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some assorted routines, mainly concerned with transcendental functions and non-linear transforms.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module math_utils

  use types

  use constants
!  use array_utils	! for getBinNumber
!  use ssc_misc		! for fracVal.

  implicit none 

  private 

  public incompleteGammaP
  public lnIncompleteGammaP
  public incompleteGammaQ
  public lnIncompleteGammaQ
  private incompleteGammaPbySeries
  private incompleteGammaQbyFraction
public incompleteGammaQByAsymtotic
  public lnGamma
  public factorial
  public erf
  private erfApprox0
  private erfApprox1
  public erfc
  private erfcApprox0
!  public calcAitoff
!  public bilinearSample
!  public bilinearValue
!  public interpolateInPolarCoord

contains 

  !---------------------------------------------------------------------
  subroutine incompleteGammaP(a, x, p, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function P(a,x), defined by
  !
  !		                 x
  !		           1    /
  !		P(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               0
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the following routine.
  !
  ! This routine was cribbed from gammp() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in)  :: a, x
    real(single), intent(out) :: p
    integer, intent(out), optional :: status

    integer :: localStatus
    real(single) :: q

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    p = 0.0

    if (x < 0.0 .or. a < 0.0 .or. (x == 0.0 .and. a == 0.0)) then
      if (present(status)) status = -1
  return
    end if

    if (x == 0.0) then
      p = 0.0
  return
    else if (a == 0.0) then
      p = 1.0
  return
    end if

    if (x < a + 1) then ! use the series representation:
      call incompleteGammaPbySeries(a, x, p, localStatus)
    else ! use the continued-fraction representation:
      call incompleteGammaQbyFraction(a, x, q, localStatus)
      p = 1.0 - q
    end if

    if (present(status)) status = localStatus

  end subroutine incompleteGammaP

  !---------------------------------------------------------------------
  subroutine incompleteGammaQ(a, x, q, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function Q(a,x), defined by
  !
  !		                inf
  !		           1    /
  !		Q(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               x
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the previous routine. Note also that the
  ! _unnormalised_ incomplete gamma T(a)Q(a,x) is also available from the
  ! slatec function gamic(a,x).
  !
  ! This routine was cribbed from gammq() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in)  :: a, x
    real(single), intent(out) :: q
    integer, intent(out), optional :: status

    integer :: localStatus
    real(single) :: p

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    p = 0.0

    if (x < 0.0 .or. a < 0.0 .or. (x == 0.0 .and. a == 0.0)) then
      if (present(status)) status = -1
  return
    end if

    if (x == 0.0) then
      p = 1.0
  return
    else if (a == 0.0) then
      p = 0.0
  return
    end if

    if (x < a + 1) then ! use the series representation:
      call incompleteGammaPbySeries(a, x, p, localStatus)
      q = 1.0 - p
    else ! use the continued-fraction representation:
      call incompleteGammaQbyFraction(a, x, q, localStatus)
    end if

    if (present(status)) status = localStatus

  end subroutine incompleteGammaQ

  !---------------------------------------------------------------------
  subroutine incompleteGammaPbySeries(a, x, p, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function P(a,x), defined by
  !
  !		                 x
  !		           1    /
  !		P(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               0
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the following routine.
  !
  ! The function is here evaluated from a series.
  !
  ! This routine was cribbed from gser() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in)  :: a, x
    real(single), intent(out) :: p
    integer, intent(out) :: status

    integer, parameter :: maxNumIterations = 100
    real(single) :: convergeCriterion, ap, summ, del, lnGammaVal
    integer :: i
    logical :: notConverged

    !-------------------------------------------------------------------
    ! Start

    status = 0
    convergeCriterion = 3.0 ** (-precision(x))

    p = 0.0

    if (x .lt. 0.0) then
      status = -1
  return
    else if (x .eq. 0.0) then
  return
    end if

    ap = a
    summ = 1.0 / a
    del = summ
    notConverged = .true. ! default
    do i = 1, maxNumIterations
      ap = ap + 1.0
      del = del * x / ap
      summ = summ + del

      if (abs(del) .lt. abs(summ) * convergeCriterion) then
        notConverged = .false.
    exit
      end if
    end do

    if (notConverged) then
      status = -2
  return
    end if

    if (a .eq. 0.5) then
      lnGammaVal = LnPi * 0.5
    else
      lnGammaVal = lnGamma(a)
    end if

    p = summ * exp(-x + a * log(x) - lnGammaVal)

  end subroutine incompleteGammaPbySeries

  !---------------------------------------------------------------------
  subroutine lnIncompleteGammaP(a, x, lnP)!!!, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function P(a,x), defined by
  !
  !		                 x
  !		           1    /
  !		P(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               0
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the following routine.
  !
  ! The function is here evaluated from a series.
  !
  ! This routine was cribbed from gser() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in)            :: a, x
    real(single), intent(out)           :: lnP
!!!    integer,      intent(out), optional :: status
integer :: status

    integer, parameter :: maxNumIterations = 100
    real(single) :: convergeCriterion, ap, summ, del, lnGammaVal
    integer :: i
    logical :: notConverged

    !-------------------------------------------------------------------
    ! Start

    status = 0
    convergeCriterion = 3.0 ** (-precision(x))

    lnP = 0.0

    if (x .le. 0.0) then
      status = -1
  return
    else if (x .eq. 0.0) then
  return
    end if

    ap = a
    summ = 1.0 / a
    del = summ
    notConverged = .true. ! default
    do i = 1, maxNumIterations
      ap = ap + 1.0
      del = del * x / ap
      summ = summ + del

      if (abs(del) .lt. abs(summ) * convergeCriterion) then
        notConverged = .false.
    exit
      end if
    end do

    if (notConverged) then
      status = -2
  return
    end if

    if (a .eq. 0.5) then
      lnGammaVal = LnPi * 0.5
    else
      lnGammaVal = lnGamma(a)
    end if

!    p = summ * exp(-x + a * log(x) - lnGammaVal)
    lnP = log(summ) - x + a * log(x) - lnGammaVal

  end subroutine lnIncompleteGammaP

  !---------------------------------------------------------------------
  subroutine incompleteGammaQbyFraction(a, x, q, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function Q(a,x), defined by
  !
  !		                inf
  !		           1    /
  !		Q(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               x
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the previous routine. Note also that the
  ! _unnormalised_ incomplete gamma T(a)Q(a,x) is also available from the
  ! slatec routine gamic(a.x).
  !
  ! The function is evaluated by a modified Lentz's continued fraction.
  !
  ! This routine was cribbed from gcf() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in)  :: a, x
    real(single), intent(out) :: q
    integer, intent(out) :: status

    integer, parameter :: maxNumIterations = 100
    real(single) :: convergeCriterion, verySmallReal, b, c, d, h, an, del, lnGammaVal
    integer :: i
    logical :: notConverged

    !-------------------------------------------------------------------
    ! Start

    status = 0
    convergeCriterion = 3.0 ** (-precision(x))
    verySmallReal = epsilon(x)

    b = x + 1.0 - a
    c = 1.0 / verySmallReal
    d = 1.0 / b
    h = d
    notConverged = .true. ! default
    do i = 1, maxNumIterations
      an = -i * (i - a)
      b = b + 2.0
      d = an * d + b
      if (abs(d) .lt. verySmallReal) d = verySmallReal
      c = b + an / c
      if (abs(c) .lt. verySmallReal) c = verySmallReal
      d = 1.0 / d
      del = d * c
      h = h * del
      if(abs(del - 1.0) .lt. convergeCriterion) then
        notConverged = .false.
    exit
      end if
    end do

    if (notConverged) then
      status = -2
  return
    end if

    if (a .eq. 0.5) then
      lnGammaVal = LnPi * 0.5
    else
      lnGammaVal = lnGamma(a)
    end if

    q = h * exp(-x + a * log(x) - lnGammaVal)

  end subroutine incompleteGammaQbyFraction

  !---------------------------------------------------------------------
  subroutine lnIncompleteGammaQ(a, x, lnQ)!!!, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function Q(a,x), defined by
  !
  !		                inf
  !		           1    /
  !		Q(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               x
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the previous routine. Note also that the
  ! _unnormalised_ incomplete gamma T(a)Q(a,x) is also available from the
  ! slatec routine gamic(a.x).
  !
  ! The function is evaluated by a modified Lentz's continued fraction.
  !
  ! This routine was cribbed from gcf() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in)  :: a, x
    real(single), intent(out) :: lnQ
!!!    integer,      intent(out), optional :: status
integer :: status

    integer, parameter :: maxNumIterations = 100
    real(single) :: convergeCriterion, verySmallReal, b, c, d, h, an, del, lnGammaVal
    integer :: i
    logical :: notConverged

    !-------------------------------------------------------------------
    ! Start

    status = 0
    convergeCriterion = 3.0 ** (-precision(x))
    verySmallReal = epsilon(x)

    b = x + 1.0 - a
    c = 1.0 / verySmallReal
    d = 1.0 / b
    h = d
    notConverged = .true. ! default
    do i = 1, maxNumIterations
      an = -i * (i - a)
      b = b + 2.0
      d = an * d + b
      if (abs(d) .lt. verySmallReal) d = verySmallReal
      c = b + an / c
      if (abs(c) .lt. verySmallReal) c = verySmallReal
      d = 1.0 / d
      del = d * c
      h = h * del
      if(abs(del - 1.0) .lt. convergeCriterion) then
        notConverged = .false.
    exit
      end if
    end do

    if (notConverged) then
      status = -2
  return
    end if

    if (a .eq. 0.5) then
      lnGammaVal = LnPi * 0.5
    else
      lnGammaVal = lnGamma(a)
    end if

!    q = h * exp(-x + a * log(x) - lnGammaVal)
    lnQ = log(h) - x + a * log(x) - lnGammaVal

  end subroutine lnIncompleteGammaQ

  !---------------------------------------------------------------------
  subroutine incompleteGammaQByAsymtotic(a, x, q, status)
  !---------------------------------------------------------------------
  ! Returns the (normalised) incomplete gamma function Q(a,x), defined by
  !
  !		                inf
  !		           1    /
  !		Q(a,x) = ------ | dt exp(-t) t^(a-1),
  !		          T(a)  /
  !		               x
  !
  ! where T(a) is the complete gamma function. Note that Q + P = 1, where
  ! P(a,x) is defined in the previous routine. Note also that the
  ! _unnormalised_ incomplete gamma T(a)Q(a,x) is also available from the
  ! slatec function gamic(a,x).
  !
  ! This routine was cribbed from gammq() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in)  :: a, x
    real(single), intent(out) :: q
    integer, intent(out) :: status

    integer, parameter :: maxNumIterations = 100
    real(single) :: convergeCriterion, ap, summ, del, oldDel, lnGammaVal
    integer :: i
    logical :: notConverged

    !-------------------------------------------------------------------
    ! Start

    status = 0
    convergeCriterion = 3.0 ** (-precision(x))
print *, 'convergeCriterion:', convergeCriterion

    q = 0.0

    if (x .lt. 0.0) then
      status = -1
  return
    else if (x .eq. 0.0) then
  return
    end if

    ap = a
    summ = 1.0
    del = summ
    oldDel = del
    notConverged = .true. ! default
    do i = 1, maxNumIterations
      ap = ap - 1.0
      del = del * ap / x
      summ = summ + del
print *, 'del, summ:', del, summ

      if (abs(del) < abs(summ) * convergeCriterion) then
        notConverged = .false.
    exit
      end if

    if (abs(del) > abs(oldDel)) exit

      oldDel = del
    end do

    if (notConverged) then
      status = -2
!!!  return
!*** for the time being calculate a prob anyway.
    end if

    if (a .eq. 0.5) then
      lnGammaVal = LnPi * 0.5
    else
      lnGammaVal = lnGamma(a)
    end if

    q = summ * exp(-x + (a - 1) * log(x) - lnGammaVal)
print *, 'q:', q

  end subroutine incompleteGammaQByAsymtotic

  !---------------------------------------------------------------------
  real function lnGamma(a)
  !---------------------------------------------------------------------
  ! Returns the natural log of the absolute value of the complete gamma
  ! function T(a), which is defined by
  !
  !		       inf
  !		       /
  !		T(a) = | dt exp(-t) t^(a-1).
  !		       /
  !		      0
  !
  ! The function is evaluated by a Lanczos approximation.
  !
  ! The ln|T(a)| function can also be evaluated using the slatec function
  ! alngam(a).
  !
  ! This routine was cribbed from gammln() in Numerical Methods for Fortran,
  ! W H Press et al.

    implicit none

    real(single), intent(in) :: a

    integer :: j
    real(double) :: ser, stp, tmp, x, y, cof(6)

    save cof, stp
    data cof, stp /&
      76.18009172947146d0,&
     -86.50532032941677d0,&
      24.01409824083091d0,&
      -1.231739572450155d0,&
       0.1208650973866179d-2,&
      -0.5395239384953d-5,&
       2.5066282746310005d0/

    !-------------------------------------------------------------------
    ! Start

    x = a
    y = x
    tmp = x + 5.5d0
    tmp = (x + 0.5d0) * log(tmp) - tmp
    ser = 1.000000000190015d0
    do j = 1, 6
      y = y + 1.0d0
      ser = ser + cof(j) / y
    end do
    lnGamma = tmp + log(stp * ser / x)

  end function lnGamma

  !---------------------------------------------------------------------
  integer function factorial(i)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: i

    integer :: j

    !-------------------------------------------------------------------
    ! Start

    factorial = 1
    do j = 2, i
      factorial = factorial * j
    end do

  end function factorial

  !---------------------------------------------------------------------
  real function erf(x, style)
  !---------------------------------------------------------------------
  ! The error function
  !
  !	                     x
  !	             2      /
  !	erf(x) = ---------- | dt exp(-t^2).
  !	          sqrt(pi)  /
  !	                   0
  !

    implicit none

    real(single), intent(in) :: x
    integer, intent(in), optional :: style

    integer :: localStyle

    !-------------------------------------------------------------------
    ! Start

    if (present(style)) then
      localStyle = style
    else
      localStyle = 0
    end if

    select case(localStyle)
    case(0)
      call incompleteGammaP(0.5, x*x, erf)
    case(1)
      erf = erfApprox0(x)
    case(2)
      erf = erfApprox1(x)
    case default
    end select

  end function erf

  !---------------------------------------------------------------------
  real function erfc(x, style)
  !---------------------------------------------------------------------
  ! The complementary error function
  !
  !	                                   inf
  !	                           2      /
  !	erfc(x) = 1 - erf(x) = ---------- | dt exp(-t^2).
  !	                        sqrt(pi)  /
  !	                                 x
  !
  ! Note that, for values obeying a gaussian distribution of mean <y> and
  ! standard deviation s, the probability of a given y value (or greater)
  ! occurring by chance is erfc(abs(y - <y>)/s/sqrt(2))

    implicit none

    real(single), intent(in) :: x
    integer, intent(in), optional :: style

    integer :: localStyle

    !-------------------------------------------------------------------
    ! Start

    if (present(style)) then
      localStyle = style
    else
      localStyle = 0
    end if

    select case(localStyle)
    case(0)
      call incompleteGammaQ(0.5, x*x, erfc)
    case(1)
      erfc = 1.0 - erfApprox0(x)
    case(2)
      erfc = 1.0 - erfApprox1(x)
    case default
    end select

  end function erfc

  !---------------------------------------------------------------------
  function erfApprox0(x) result(erf)
  !---------------------------------------------------------------------
  ! Taken from Numerical Methods for Fortran.

    implicit none

    real(single), intent(in) :: x

    real(single) :: t, z, erf
    real(single), parameter ::&
      p0 = -1.26551223,&
      p1 =  1.00002368,&
      p2 =  0.37409196,&
      p3 =  0.09678418,&
      p4 = -0.18628806,&
      p5 =  0.27886807,&
      p6 = -1.13520398,&
      p7 =  1.48851587,&
      p8 = -0.82215223,&
      p9 =  0.17087277

    !-------------------------------------------------------------------
    ! Start

    z = abs(x)
    t = 1.0 / (1.0 + 0.5 * z)
    erf = t * exp(-z * z + p0 + t * (p1 + t * (p2&
        + t * (p3 + t * (p4 + t * (p5 + t * (p6&
        + t * (p7 + t * (p8 + t * p9)))))))))

    if (x .lt. 0.0) erf = 2.0 - erf

  end function erfApprox0

  !---------------------------------------------------------------------
  function erfApprox1(x) result(erf)
  !---------------------------------------------------------------------
  ! This approx uses the formula of Busen K M, Am J Phys 49, 886 (1981).

    implicit none

    real(single), intent(in) :: x
    real(single) :: erf, sign

    !-------------------------------------------------------------------
    ! Start

    sign = 1.0
    if (x < 0.0) sign = -1.0

    if (x > 1.0) then
      erf = sqrt(1.0 - exp(-4.0 * x**1.88 / Pi))
    else
      erf = sqrt(1.0 - exp(-4.0 * x * x / Pi))
    end if

    erf = erf * sign

  end function erfApprox1

  !---------------------------------------------------------------------
  function erfcApprox0(x) result(erfc)
  !---------------------------------------------------------------------
  !x    input	Deviate, range -infinity to +infinity.
  !erfc output	Area under Gaussian from X to +infinity.
  !-
  ! Author	Clive Page	1983 July 5
  ! Standard Fortran-77 version 1984 Sept 25, CGP.
  ! From algorithm AS66 by I.D.Hill, Appl Stats  22,424 (1973).
  ! Adapted to f90 by Ian Stewart 8 Aug 2000

    implicit none 

    real(single), intent(in) :: x
    real(single) :: erfc, y, z

    !-------------------------------------------------------------------
    ! Start

    z = abs(x)
    if(z .gt. 13.0) then
      erfc = 0.0

    else
      y = 0.5 * z * z
      if(z .le. 1.28) then
        erfc = 0.5 - z * (0.398942280444 - 0.399903438504 * y /&
        (y + 5.75885480458 - 29.8213557808 /&
        (y + 2.62433121679 + 48.6959930692 /&
        (y + 5.92885724438))))

      else
        erfc = 0.398942280385 * exp(-y) /&
        (z - 3.8052E-8      + 1.00000615302 /&
        (z + 3.98064794E-4  + 1.98615381364 /&
        (z - 0.151689116635 + 5.29330324926 /&
        (z + 4.8385912808   - 15.1508972451 /&
        (z + 0.742380924027 + 30.789933034  /&
        (z + 3.99019417011))))))
      end if
    end if

    if(x .lt. 0.0) erfc = 1.0 - erfc

  end function erfcApprox0

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module math_utils
