!
! Module: minimizations
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some assorted function minimization routines.
! 
! Comments: TODO:
!*** change y and y(x) data types in newtonsMethod and hybridNewtonsMethod to real(single)
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module minimizations

  use types

!  use constants
  use array_utils	! for getBinNumber, smallestTrueIndex, largestTrueIndex,
			! trapezoidSum, minLoc1d, maxLoc1d.
  use ssc_misc		! for fracVal, swap.

  implicit none 

  private 

  public newtonsMethod
  public hybridNewtonsMethod
  public riddersMethod
  public simplexMinimization

contains 

  !---------------------------------------------------------------------
  subroutine newtonsMethod(startX, y, dy, solutionX, status)
  !---------------------------------------------------------------------
  ! This subroutine solves the equation y(x) = 0 by Newton's Method. The
  ! user must construct two functions in the calling module, one of which
  ! returns y and the other dy/dx for a given x. These functions (and their
  ! single argument x) must be of type real. The user should pass their
  ! names in the call to newtonsMethod.

    implicit none

    real(single), intent(in) :: startX
    real(single), intent(out) :: solutionX
    integer, intent(out) :: status

!    interface
!      real(single) function y(x)
!        real(single), intent(in) :: x
!      end function y
!      real(single) function dy(x)
!        real(single), intent(in) :: x
!      end function dy
!    end interface
interface
  real function y(x)
    real, intent(in) :: x
  end function y
  real function dy(x)
    real, intent(in) :: x
  end function dy
end interface

    real(single) :: yValue, yDeriv, x, oldXstep, xStep, myPrecision

    !-------------------------------------------------------------------
    ! Start

    status = 0 ! default

    myPrecision = 10.0 ** (-precision(x))

    yValue = y(startX)
    yDeriv = dy(startX)
    if (yDeriv .eq. 0.0) then
      status = -1
  return
    end if
    Xstep = -Yvalue / yDeriv
    if (abs(startX) .gt. 0.0) then
      if (abs(Xstep / startX) .le. myPrecision) then
        solutionX = startX + Xstep
  return
      end if
    end if
    x = startX + Xstep
    oldXstep = Xstep

    do
      yValue = y(x)
      yDeriv = dy(x)
      if (yDeriv .eq. 0.0) then
        status = -1
  return
      end if

      Xstep = -yValue / yDeriv

      if (abs(x) .gt. 0.0) then
        if (abs(Xstep / x) .le. myPrecision) then
          solutionX = x + Xstep
  return
        end if
      end if

      if (abs(Xstep) .gt. abs(oldXstep)) then
        status = -2
  return
      end if

      x = x + Xstep
      oldXstep = Xstep
    end do

  end subroutine newtonsMethod

  !---------------------------------------------------------------------
  subroutine hybridNewtonsMethod(startX0, startX1, y, dy, solutionX&
  , maxNumIterations, convergenceCriterion, status)
  !---------------------------------------------------------------------
  ! Adapted from the hybrid bisection/Newton-Raphson method in Numerical
  ! Methods.
  !
  ! This subroutine solves the equation y(x) = 0 by Newton's Method. The
  ! user must construct two functions in the calling module, one of which
  ! returns y and the other dy/dx for a given x. These functions (and their
  ! single argument x) must be of type real. The user should pass their
  ! names in the call to hybridNewtonsMethod.
  !
  ! The starting values startX0 and startX1 should bracket the zero point
  ! but it is not necessary that they be in any order.

    implicit none

    real(single), intent(in)            :: startX0, startX1
    real(single), intent(out)           :: solutionX
    integer,      intent(in),  optional :: maxNumIterations
    real(single), intent(in),  optional :: convergenceCriterion
    integer,      intent(out), optional :: status

!    interface
!      real(single) function y(x)
!        real(single), intent(in) :: x
!      end function y
!      real(single) function dy(x)
!        real(single), intent(in) :: x
!      end function dy
!    end interface
interface
  real function y(x)
use types
    real, intent(in) :: x
  end function y
  real function dy(x)
    real, intent(in) :: x
  end function dy
end interface

    real(single) ::&
      yValue, yDeriv, localConvergenceCriterion, startY0, startY1,&
      xLo, xHi, dx, dxOld, oldSolutionX
    integer :: localStatus, localMaxNumIter, j

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    solutionX = 0.0 ! Don't like to leave it with no value, even if the routine
    ! can't calculate it.

    if (present(maxNumIterations)) then
      localMaxNumIter = maxNumIterations
    else
      localMaxNumIter = 100
!*** can do better than this.
    end if

    startY0 = y(startX0)
    startY1 = y(startX1)

    if (startY0 == 0.0) then
      solutionX = startX0
  return
    end if
    if (startY1 == 0.0) then
      solutionX = startX1
  return
    end if

    if (startX0 == startX1) then
      if (present(status)) status = -1 ! no space between bounds.
  return
    end if

    if (startY0 > 0.0 .and. startY1 > 0.0&
    .or. startY0 < 0.0 .and. startY1 < 0.0) then
      if (present(status)) status = -2 ! startX0 and
      ! startX1 don't bracket the root.
  return
    end if

    ! Orient bounds such that y(xLo) < 0:
    if (startY0 < 0.0) then
      xLo = startX0
      xHi = startX1
    else
      xLo = startX1
      xHi = startX0
    end if

    solutionX = 0.5 * (xLo + xHi)
    dxOld = xHi - xLo
    dx = dxOld

    if (present(convergenceCriterion)) then
      localConvergenceCriterion = abs(convergenceCriterion)
    else
      localConvergenceCriterion = 0.5 * abs(xLo + xHi)&
      * (10.0 ** (1-precision(dx)))
    end if

    yValue = y(solutionX)
    yDeriv = dy(solutionX)
    do j = 1, localMaxNumIter
      dxOld = dx

      if (((solutionX - xHi) * yDeriv - yValue)&
      * ((solutionX - xLo) * yDeriv - yValue) > 0.0&
      .or. abs(2.0 * yValue) > abs(dxOld * yDeriv)) then
        dx = 0.5 * (xHi - xLo)
        solutionX = xLo + dx
    if (solutionX == xLo) exit

      else
        dx = yValue / yDeriv
        oldSolutionX = solutionX
        solutionX = solutionX - dx
    if (solutionX == oldSolutionX) exit
      end if

    if (abs(dx) < localConvergenceCriterion) exit

      yValue = y(solutionX)
      yDeriv = dy(solutionX)

      if (yValue < 0.0) then
        xLo = solutionX
      else
        xHi = solutionX
      end if
    end do
    if (j >= localMaxNumIter) then
      localStatus = ibset(localStatus, 0)
    end if

    if (present(status)) status = localStatus

  end subroutine hybridNewtonsMethod

  !---------------------------------------------------------------------
  subroutine riddersMethod(startX0, startX1, y, solutionX&
  , maxNumIterations, convergenceCriterion, dx, status)
  !---------------------------------------------------------------------
  ! This subroutine has been adapted from an algorithm described in
  ! Numerical Methods for Fortran. I changed it slightly in that the
  ! convergence test compares the difference between the present and
  ! last values of xNew, rather than the difference between xHi and xLo
  ! as in the original algorithm, to the value of convergenceCriterion.
  !
  ! This subroutine solves the equation y(x) = 0 by Ridders' Method. The
  ! user must construct the function y(x) in the calling module. This
  ! function (and its single argument x) must be of type real. The user
  ! should pass the function name in the call to riddersMethod.
  !
  ! Note that y(startX0) and y(startX1) must be of opposite sign.

    implicit none

    real(single), intent(in)            :: startX0, startX1
    real(single), intent(out)           :: solutionX
    integer,      intent(in),  optional :: maxNumIterations
    real(single), intent(in),  optional :: convergenceCriterion
    real(single), intent(out), optional :: dx
    integer,      intent(out), optional :: status

    interface
      function y(x)
        use types

        real(single), intent(in) :: x
        real(single) :: y
      end function y
    end interface

    integer ::&
      localStatus, j
    real(single) ::&
      startY0, startY1, xLo, yLo, xMid, yMid, xHi, yHi, xOld, xNew, yNew,&
      root, expQ, fracY, localConvergenceCriterion

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default
    if (present(status)) status = localStatus
    solutionX = 0.0 ! Don't like to leave it with no value, even if the routine
    ! can't calculate it.

    dx = 0.0 ! default

    startY0 = y(startX0)
    startY1 = y(startX1)

    if (startY0 == 0) then
      solutionX = startX0
  return
    end if
    if (startY1 == 0) then
      solutionX = startX1
  return
    end if

    if (startX0 == startX1) then
      if (present(status)) status = -1 ! no space between startX0 and startX1.
  return
    end if

    if (startY0 > 0.0 .and. startY1 > 0.0&
    .or. startY0 < 0.0 .and. startY1 < 0.0) then
      if (present(status)) status = -2
  return
    end if

    ! Orient bounds such that y(xLo) < 0:
    if (startY0 < 0.0) then
      xLo = startX0
      yLo = startY0
      xHi = startX1
      yHi = startY1
    else
      xLo = startX1
      yLo = startY1
      xHi = startX0
      yHi = startY0
    end if

    dx = xHi - xLo

    xOld = xLo ! could have also chosen xHi; at this point it is
    ! indeterminate which of these two values is the 'latest estimate'.

    if (present(convergenceCriterion)) then
      localConvergenceCriterion = abs(convergenceCriterion)
    else
      localConvergenceCriterion = 0.5 * abs(startX0 + startX1)&
      * (10.0 ** (1 - precision(dx)))
    end if

    j = 1
    do
      xMid = 0.5 * (xLo + xHi)
      yMid = y(xMid)
      root = sqrt(yMid**2 - yLo * yHi)

      if (root == 0) then ! The argument of the square root can
      ! only equal zero if one at least of yLo and yHi equals zero (and
      ! if yMid is also equal to zero). The routine is supposed to screen out
      ! such cases before this point is reached. This status value indicates
      ! a bug.
        if (present(status)) status = -3
  return
      end if

      ! Solve quadratic equation to get expQ:

      expQ = (yMid + root) / yHi ! That yHi > 0 is checked implicitly
      ! elsewhere in the loop.

      ! Apply false-position method to y(x)*exp(2Q(x-xLo)/(xHi-xLo)):

      fracY = -yLo / (yHi * expQ * expQ - yLo)
      xNew = fracVal(xLo, xHi, fracY)
      yNew = y(xNew)

      if (yNew == 0.0) then
        solutionX = xNew
    exit
      end if

      ! Now decide new x bounds:

      if (yNew < 0.0) then
        xLo = xNew
        yLo = yNew
      else
        xHi = xNew
        yHi = yNew
      end if

      ! Check convergence:

      dx = xNew - xOld
      if (abs(dx) < localConvergenceCriterion) then
        solutionX = xNew
    exit
      end if

      j = j + 1
      if (present(maxNumIterations)) then
        if (j > maxNumIterations) then
          localStatus = ibset(localStatus, 0)
          solutionX = xNew
    exit
        end if
      end if

      xOld = xNew
    end do

    if (present(status)) status = localStatus

  end subroutine riddersMethod

  !---------------------------------------------------------------------
  subroutine simplexMinimization(funk, ftol, simplex, y, numFunctionEvals, status)
  !---------------------------------------------------------------------
  ! This implements a multidimensional minimization of the function funk(x), where x(1:numDimensions) is a vector in numDimensions dimensions, by the downhill simplex method of Nelder and Mead. The matrix p(1:N+1,1:N), where N is the number of dimensions, contains the simplex. Each of the N+1 rows of p is an N-dimensional vector: these represent the vertices of the simplex. The calling routine must supply starting values of p, which is therefore of type inout. Also input is the vector y(1:N+1), the components of which must be pre-initialized to the values of funk evaluated at the N+1 vertices (rows) of p; and ftol the fractional convergence tolerance to be achieved in the function value. On output, p and y will have been reset to numDimensions+1 new points, all within ftol of a minimum function value, and iter gives the number of function evaluations taken.
  !
  ! Adapted from Numerical Recipes in Fortran
!**** authors, press, year

    implicit none

    interface
      function funk(x)
        use types

        real(single), intent(in) :: x(:)
        real(single)             :: funk
      end function funk
    end interface

    real(single),   intent(in)            :: ftol
    real(single),   intent(inout)         :: simplex(:,:)
    real(single),   intent(out)           :: y(size(simplex, 1))
    integer(int32), intent(out)           :: numFunctionEvals
    integer,        intent(out), optional :: status

    integer :: numDimensions, n, ilo, ihi, inhi, i
    real(single) :: rtol, ytry, ysave
    real(single), parameter :: TINY = 1.0e-10
    integer, parameter :: maxNumIterations=5000

    !-------------------------------------------------------------------
    ! Start

    numDimensions = size(simplex, 2)

!*** default values for y?

    if (size(simplex, 1) .ne. numDimensions + 1) then
      if (present(status)) status = -1
  return
    end if

    if (numDimensions < 2) then
      if (present(status)) status = -2
  return
    end if

    ! Calculate initial y:
    do i = 1, numDimensions + 1
      y(i) = funk(simplex(i, :)) ! not counted in numFunctionEvals.
    end do

    numFunctionEvals = 0

    do ! start of iteration loop

      ! Determine which point of the simplex is the highest (ie worst), the
      ! next highest, and the lowest (ie best):

      call calcBestWorst(y, ilo, inhi, ihi)

      ! Check for convergence:

      rtol = 2.0 * abs(y(ihi) - y(ilo)) / (abs(y(ihi)) + abs(y(ilo)) + TINY)

      if (rtol < ftol) then ! if returning, put best point and value in slot 1:
        call swap(y(1), y(ilo)) ! in ssc_misc

        do n = 1, numDimensions
          call swap(simplex(1, n), simplex(ilo, n)) ! in ssc_misc
        end do

        if (present(status)) status = 0
  return
      end if

      if (numFunctionEvals >= maxNumIterations) then
!*** do anything with data?
        if (present(status)) status = ibset(status, 0)
  return
      end if

      ! Reflect the position of the point with the highest y value
      ! through the opposite face of the simplex and calculate y there:

      call moveHiPointTrial(simplex, y, funk, ihi, -1.0, ytry)
      numFunctionEvals = numFunctionEvals + 1

      if (ytry <= y(ilo)) then ! gives a result better than the best point,
      ! so try an additional extrapolation by a factor 2:

        call moveHiPointTrial(simplex, y, funk, ihi, 2.0, ytry)
        numFunctionEvals = numFunctionEvals + 1

      else if (ytry >= y(inhi)) then ! the reflected point is worse than
      ! the second-highest, so look for an intermediate lower point, ie do
      ! a 1-dimensional contraction:

        ysave = y(ihi)
        call moveHiPointTrial(simplex, y, funk, ihi, 0.5, ytry)
        numFunctionEvals = numFunctionEvals + 1

        if (ytry >= ysave) then ! can't seem to get rid of that high point.
        ! Better contract around the lowest (best) point:

          do i = 1, numDimensions + 1
          if (i == ilo) cycle
            simplex(i, :) = 0.5 * (simplex(i, :) + simplex(ilo, :)) ! vector op
            y(i) = funk(simplex(i, :))
          end do
          numFunctionEvals = numFunctionEvals + numDimensions ! Keep track
          ! of function iterations.
        end if
      end if
    end do

  end subroutine simplexMinimization

  !---------------------------------------------------------------------
  subroutine calcBestWorst(y, ilo, inhi, ihi)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: y(:)
    integer,      intent(out) :: ilo, ihi, inhi

    logical(bool), allocatable :: mask(:)

    !-------------------------------------------------------------------
    ! Start

    allocate(mask(size(y)))

    ilo  = minLoc1d(y)		! in array_utils
    mask = .true.
    mask(ilo) = .false.
    ihi  = maxLoc1d(y, mask)	! in array_utils
    mask(ihi) = .false.
    inhi = maxLoc1d(y, mask)	! in array_utils

  end subroutine calcBestWorst

  !---------------------------------------------------------------------
  subroutine moveHiPointTrial(simplex, y, funk, ihi, fac, ytry)
  !---------------------------------------------------------------------
  ! Moves the point with the highest function value away from the mean point by the factor fac, tries it the function at this new location, and replaces the high point if the new point is better.

    implicit none

    integer,      intent(in)    :: ihi
    real(single), intent(in)    :: fac
    real(single), intent(inout) :: simplex(:,:), y(size(simplex,1))
    real(single), intent(out)   :: ytry

    interface
      function funk(x)
        use types

        real(single), intent(in) :: x(:)
        real(single)             :: funk
      end function funk
    end interface

    real(single), allocatable :: ptry(:), meanPoint(:)
    integer :: numDimensions, numPoints, dimNum

    !-------------------------------------------------------------------
    ! Start

    numDimensions = size(simplex,2)
    allocate(ptry(numDimensions), meanPoint(numDimensions))

    numPoints = numDimensions + 1
    do dimNum = 1, numDimensions
      meanPoint(dimNum) = sum(simplex(:, dimNum))
    end do
    meanPoint = meanPoint / real(numPoints)

    ptry = ((1.0 - fac) * numPoints * meanPoint&
      + (numPoints * fac - 1.0) * simplex(ihi, :))&
      / real(numDimensions) ! vector op

    ! This expression is derived as follows. Suppose we have a simplex of N+1 points p_i in a space of dimension N. The position of the mean point 'a' is given by a = sum(i=1,N+1)(p_i). Now select one of the simplex points, say point p_j. (The interest in the present routine is to select that which currently gives the highest value of the function funk.) 'Moving the point through the opposite face of the simplex' means to slide it like a bead on a wire along a straight line connecting its initial position and that of the mean point. To specify where we want it to end up, it is most convenient to specify the ratio fac, which is given by
    !
    !	       |p_j_new - a_new|
    !	fac = -------------------,
    !	       |p_j_old - a_old|
    !
    ! or in other words
    !
    !	p_j_new - a_new = fac * p_j_old - fac * a_old.			(1)
    !
    ! To arrive at the formula for p_j_new we should also consider that the new average can be expressed in another way as:
    !
    !	                 p_j_old     p_j_new
    !	a_new = a_old - --------- + ---------.				(2)
    !	                   N+1         N+1
    !
    ! Inserting (2) into (1) gives the desired expression.

    ytry = funk(ptry)
    if (ytry < y(ihi)) then
      y(ihi) = ytry
!      psum = psum - p(ihi, :) + ptry	! vector op
      simplex(ihi, :) = ptry	! vector op
    end if

  end subroutine moveHiPointTrial

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module minimizations
